/***************************************************************************
                          bootroot2cd.cpp  -  description
                             -------------------
    begin                : Tue Sep 26 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <getopt.h>

#include "bootroot2cd.h"

extern char *optarg;
extern int optind;
extern int opterr;

static struct option const long_options[] =
{
  {"askdisk", no_argument, NULL, 'a'},
  {"loaddisk", no_argument, NULL, 'l'},
  {"device", required_argument, NULL, 'd'},
  {"help", required_argument, NULL, 'h'},
  {NULL, 0, NULL, 0}
};

static char optstring[]="ald:h";

#define ROOTFLAGS_LOADRAMDISK 16384
#define ROOTFLAGS_ASKRAMDISK 32768

// =======================================================
int main(int argc, char *argv[])
{
  cout << "======= bootable CD image creator ==========" << endl;
	
	// ------- open files -------
	FILE *fBoot, *fRoot, *fCd;
	unsigned long int dwBootSize;
	unsigned long int dwRootSize;
	unsigned short int nBootBlocksCount;
	int nRdevRootFlags;
	int nCurByte;
	int nOptCh;
	
	// Default values
	unsigned int nFlags = ROOTFLAGS_LOADRAMDISK;
	unsigned int nRdevRootDevice = 0; // 0: the root device is the "Boot device", 0x37ee = ??
	
	// read command line
	while ((nOptCh = getopt_long (argc, argv, optstring, long_options, NULL)) != EOF)
	{
		switch(nOptCh)
		{
			case 'h': // help
				usage();
				return EXIT_SUCCESS;	
				break;
								
			case 'a': // ask for root disk
				nFlags |= ROOTFLAGS_ASKRAMDISK;
				break;
								
			case 'l': // do not load ram disk
				nFlags &= ~ROOTFLAGS_LOADRAMDISK;
				break;
			
			case 'd': // root device
				nRdevRootDevice = atol(optarg);
				break;
			
		}
	}
	
	if (argc - optind != 3)
  {	
		usage();
		return EXIT_FAILURE;
	}
	
	errno = 0;
	fCd = fopen(argv[optind+0], "wb+");
	if (!fCd)
	{
		printf("Can't create file %s (%s)\n", argv[optind+0], strerror(errno));	
		return EXIT_FAILURE;
	}
	
	errno = 0;
	fBoot = fopen(argv[optind+1], "rb");
	if (!fBoot)
	{
		printf("Can't read file %s (%s)\n", argv[optind+1], strerror(errno));	
		return EXIT_FAILURE;
	}
	
	errno = 0;
	fRoot = fopen(argv[optind+2], "rb");
	if (!fRoot)
	{
		printf("Can't read file %s (%s)\n", argv[optind+2], strerror(errno));	
		return EXIT_FAILURE;
	}
	
	// ------- get size of boot & root disks -------
	
	fseek(fBoot, 0, SEEK_END);
	dwBootSize = ftell(fBoot);
	fseek(fBoot, 0, SEEK_SET);
	
	fseek(fRoot, 0, SEEK_END);
	dwRootSize = ftell(fRoot);
	fseek(fRoot, 0, SEEK_SET);
	
	printf("boot image = %lu bytes\nroot image = %lu bytes\n", dwBootSize, dwRootSize);
	
	// ------- copy data -------

	// copy boot image data
	while (!feof(fBoot))
	{ nCurByte = fgetc(fBoot);
		fputc(nCurByte, fCd);
	}
	
	// write zero bytes to finish last 512 block
	while ((ftell(fCd)+1) % 1024 != 0)
	{	fputc(0, fCd);
	}
	
	// offset of the RAM disk
	nBootBlocksCount = (ftell(fCd)+1) / 1024;
	
	// copy root disk
	while (!feof(fRoot))
	{ nCurByte = fgetc(fRoot);
		fputc(nCurByte, fCd);
	}
		
	// make size of the image be 2880 KB
	while ((ftell(fCd)) < (2880 * 1024))
	{	fputc(0, fCd);
	}

	// --- rdev: set ram disk size
	fseek(fCd, 508, SEEK_SET); // 508 = offset of the
	fwrite(&nRdevRootDevice, 2, 1, fCd);
	printf("Root device = %d = 0x%x\n", nRdevRootDevice, nRdevRootDevice);
	
	// --- rdev: set rootflags
	fseek(fCd, 504, SEEK_SET); // 498 = flags
	nRdevRootFlags = nBootBlocksCount | nFlags; // "load ramdisk", "ask rootdisk" and "ram disk offset"
	fwrite(&nRdevRootFlags, 2, 1, fCd);
	printf("RAM disk image offset = block %d\n", nBootBlocksCount);
	if (nRdevRootFlags & ROOTFLAGS_LOADRAMDISK) // 2^14
		printf("load root image: yes\n");
	else
		printf("load root image: no\n");
	if (nRdevRootFlags & ROOTFLAGS_ASKRAMDISK) // 2^15
		printf("ask for root disk: yes\n");
	else
		printf("ask for root disk: no\n");
	
	// ------- close files -------
	fclose(fBoot);
	fclose(fRoot);
	fclose(fCd);
		
	return EXIT_SUCCESS;
}

// =======================================================
void usage()
{
  cout << "=========== bootable CD image creator usage ============" << endl << endl;
  cout << "This utility is used to create a boot image you will burn on a CD-R" << endl;
  cout << "to create a bootable CD with the Linux kernel and Partition Image at" << endl;
  cout << "startup. This utility requires the two image files (root & boot disks)" << endl;
  cout << "provided on the official Partition Image web site." << endl;
  cout << "Usage: bootroot2cd <boot-image-to-create> <boot floppy image> <root floppy image>" << endl << endl;
  cout << "ex: bootroot2cd boot-image.raw partimage-0.3.3-bootdisk-1.raw partimage-0.3.3-rootdisk-1.raw" << endl << endl;
  cout << "Options:" << endl;
  cout << "  -a,  --askdisk: ask root disk (default = no)" << endl;
  cout << "  -l,  --loaddisk: do not load root disk, (default = yes)" << endl;
  cout << "  -dX, --device=X: number of the root device to load (default = 0x37ee)" << endl;
  cout << "  -h,  --help: show help" << endl;
  cout << endl;
}