/***************************************************************************
                          cext2part.cpp  -  description
                             -------------------
    begin                : Mon May 22 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cext2part.h"
#include "partimage.h"
#include "imagefile.h"
#include "misc.h"
#include "gui_text.h"

// =======================================================
CExt2Part::CExt2Part(char *szDevice, FILE *fDeviceFile)
{
	fprintf(m_fDebug, "begin of CExt2Part::CExt2Part()\n"); fflush(m_fDebug);  
	m_fDeviceFile = fDeviceFile;
	strncpy(m_szDevice, szDevice, MAX_DEVICENAMELEN);
	fprintf(m_fDebug, "end of CExt2Part::CExt2Part()\n"); fflush(m_fDebug);  
}

// =======================================================
CExt2Part::~CExt2Part()
{
}

// =======================================================
int CExt2Part::saveToImage(CImage *image, COptions *options, CSaveWindow *gui)
{
	int nRes;	
	DWORD i; // for(;;)
	errcode_t err;
	DWORD dwBitmapSize;
	char cBootBuffer[EXT2_BOOTSECT_LEN+10];
	char szCommand[1024];

	DWORD dwNbWrittenBlocks = 0L;
	image->m_dwCurrentVolumeSize = 0L;
	image->m_dwVolumeNumber = 0L;
	m_guiSave = gui;
	
	m_image = image;
	
	// 0. open file system
	err = ext2fs_open (m_szDevice, 0, 0, 0, unix_io_manager, &m_fs);
	if (err)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't open partition with ext2fs_open(). Maybe partition is not an ext2 file system."));
		return -1;
	}
	
	// 1. check partition has an ext2 file system
	if (m_fs->super->s_magic != EXT2_SUPER_MAGIC)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("%s is not an ext2fs file system partition"), m_szDevice);
		return -1;
	}
	
	// 2. check the file system with e2fsck
	if (options->bCheckBeforeSaving == true)
	{
	  newtPushHelpLine(i18n("checking the file system with e2fsck"));

	  sprintf(szCommand, "/sbin/e2fsck -nf -C 0 %s > /dev/null 2>/dev/null", m_szDevice);
		nRes = system(szCommand);
		if (nRes != 0)
		{	
			nRes = newtWinChoice(i18n("Warning"), i18n("Continue"), i18n("Cancel"), i18n("There are errors on the file system, or cannot run e2fsck"));
			if (nRes == 2)
				return -1;
		}
	}
		
	// 6. read and print important informations	
	m_dwBlockSize = EXT2_BLOCK_SIZE(m_fs->super);
	
	m_dwUsedBlocks = m_fs->super->s_blocks_count - m_fs->super->s_free_blocks_count;
	
	// ---- 1. Write header
	CExt2Header headExt2;
	memset(&headExt2, 0, sizeof(CExt2Header));
	
	strcpy(headExt2.szAppFormat, CURRENT_EXT2_FORMAT);
	strcpy(headExt2.szLabel, m_fs->super->s_volume_name);
	headExt2.dwBlockSize = EXT2_BLOCK_SIZE(m_fs->super);
	headExt2.dwBootSize = EXT2_BOOTSECT_LEN;
	headExt2.dwFirstBlock = m_fs->super->s_first_data_block;
	headExt2.dwExt2Flags = 0;
	headExt2.dwNbBootCopies = NB_BOOT_COPIES_IN_IMAGE;
	headExt2.dwNbGroups = m_fs->group_desc_count;
	headExt2.dwTotalBlocksPerGroup = m_fs->super->s_blocks_per_group;
	headExt2.dwTotalBlocksNb = m_fs->super->s_blocks_count;
	headExt2.dwBitmapSize = (m_fs->super->s_blocks_count / 8)+1; // more than the size
	headExt2.dwFreeBlocksNb = m_fs->super->s_free_blocks_count;
	headExt2.dwUsedBlocksNb = m_fs->super->s_blocks_count - m_fs->super->s_free_blocks_count;

	// printf informations about partition		
	m_dwGroupsCount = headExt2.dwNbGroups;
	m_dwTotalBlocksCount = headExt2.dwTotalBlocksNb;
	m_dwBitmapSize = headExt2.dwBitmapSize;
	m_dwFirstBlock =  headExt2.dwFirstBlock;
	m_dwFreeBlocksCount = headExt2.dwFreeBlocksNb;
	m_dwUsedBlocksCount = headExt2.dwUsedBlocksNb;
	strncpy(m_szLabel, headExt2.szLabel, 128);
	
	printfInformations();
	
	// write header to image file
	nRes = m_image->write(&headExt2, sizeof(CExt2Header), *options, true);
	if (nRes != sizeof(CExt2Header))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write ext2 header"));
		ext2fs_close (m_fs);
		return -1;
	}

	// ---- 3. copy boot sector
	nRes = m_image->writeMagic(MAGIC_BEGIN_BOOT, *options);
	if (nRes == -1)
		return -1;
	
	nRes = fread(cBootBuffer, EXT2_BOOTSECT_LEN, 1, m_fDeviceFile);
	if (nRes != 1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read boot sector"));
		ext2fs_close (m_fs);
		return -1;
	}
	
	for (i=0; i < NB_BOOT_COPIES_IN_IMAGE; i++)
	{	
		nRes = m_image->write(cBootBuffer, EXT2_BOOTSECT_LEN, *options, true);
		if (nRes != EXT2_BOOTSECT_LEN)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write boot sector"));
			ext2fs_close (m_fs);
			return -1;
		}
	}
	
	// ---- 4. read bitmap
	err = ext2fs_read_block_bitmap(m_fs);
	if (err)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read block bitmaps\n"));
		ext2fs_close (m_fs);
		return -1;
	}
	
	// ---- 5. copy blocks bitmap
	nRes = m_image->writeMagic(MAGIC_BEGIN_BITMAP, *options);
	if (nRes == -1)
		return -1;
	
	dwBitmapSize = (m_fs->super->s_blocks_count / 8)+1; // more than the size
	
	nRes = m_image->write(m_fs->block_map->bitmap, dwBitmapSize, *options, true);
	if (nRes != dwBitmapSize)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write blocks bitmap"));
		ext2fs_close (m_fs);
		return -1;
	}	
	
	// ---- 6. Write all groups
	nRes = m_image->writeMagic(MAGIC_BEGIN_DATABLOCKS, *options);
	if (nRes == -1)
		return -1;
	
	// go to the beginning of the first group
	fseek(m_fDeviceFile, (long) (m_dwFirstBlock * m_dwBlockSize), SEEK_SET);

	newtPushHelpLine(i18n("copying used data blocks"));
	fprintf(m_fDebug, "begin of writeGroupInImageFile(..............)\n"); fflush(m_fDebug);  
	for (i=0; i < m_fs->group_desc_count; i++)
	{
		nRes = writeGroupInImageFile(i, &dwNbWrittenBlocks, options);
		if (nRes == -1)
		{	ext2fs_close (m_fs);
			return -1;	
		}
	}	
	fprintf(m_fDebug, "end of writeGroupInImageFile(..............)\n"); fflush(m_fDebug);  
	
	m_guiSave->setProgressValue(dwNbWrittenBlocks, dwNbWrittenBlocks);
	ext2fs_close (m_fs);
	
	return 0;
}

// =======================================================
int CExt2Part::writeGroupInImageFile(DWORD dwGroupNum, DWORD *dwNbWrittenBlocks, COptions *options)
{
	DWORD i;
	char cBlockBuffer[EXT2_MAX_BLOCK_SIZE]; // Check for the size
	int nRes;
	DWORD dwBlocksInThisGroup;
	char *cBitmap;
	
	// Number of blocks in this group (can be less if this is the last block
	if (m_dwFirstBlock + ((dwGroupNum+1) * m_fs->super->s_blocks_per_group) <= m_fs->super->s_blocks_count)
		dwBlocksInThisGroup = m_fs->super->s_blocks_per_group;
	else
		dwBlocksInThisGroup = m_fs->super->s_blocks_count - (m_dwFirstBlock + ((dwGroupNum) * m_fs->super->s_blocks_per_group));

	// read descriptor
	cBitmap = m_fs->block_map->bitmap + ((dwGroupNum * m_fs->super->s_blocks_per_group) / 8);
		
	// copy all used blocks
	for (i=0; i < dwBlocksInThisGroup; i++)
	{
		// if block is USED
		if (ext2fs_test_bit(i, cBitmap))
		{
			errno = 0;
			nRes = fread(cBlockBuffer, m_dwBlockSize, 1, m_fDeviceFile);
			if (nRes != 1)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read block %lu of group %lu. %s"), i, dwGroupNum, strerror(errno));
				return -1;
			}
			
			nRes = m_image->write(cBlockBuffer, m_dwBlockSize, *options, true);
			if (nRes != m_dwBlockSize)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write block %lu of group %lu"), i, dwGroupNum);
				return -1;
			}	
			
			(*dwNbWrittenBlocks)++;
		}
		else // if block is FREE
		{
		    // we must skip the free block with fseek()
		    fseek(m_fDeviceFile, (long) m_dwBlockSize, SEEK_CUR);
	    	
		}
		
		// draw progress
		if (i%15 == 0)
			m_guiSave->setProgressValue((*dwNbWrittenBlocks), m_dwUsedBlocks);
	}
	return 0;
}

// =======================================================
int CExt2Part::restoreFromImage(CImage *image, COptions *options, CRestoreWindow *gui)
{
	int nRes;
	char cBlockBuffer[EXT2_MAX_BLOCK_SIZE]; // Check for the size
	char cNullBuffer[EXT2_MAX_BLOCK_SIZE]; // Check for the size
	DWORD i; // for(;;)
	DWORD dwRestoredBlocks;
	char *cBitmap;
	char cBootBuffer[EXT2_BOOTSECT_LEN+10];
	//char cBootBuffer2[EXT2_BOOTSECT_LEN+10];
	
	image->m_dwCurrentVolumeSize = 0L;
	image->m_dwVolumeNumber = 0L;
	m_guiRestore = gui;
	
	m_image = image;
	
	newtPushHelpLine(i18n("reading partition informations from image"));
	
	// ---- read header
	CExt2Header headExt2;
	
	nRes = m_image->read(&headExt2, sizeof(CExt2Header), image->m_bIsSplitted, *options, true); // false = maybe
	if (nRes != sizeof(CExt2Header))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read main header"));
		return -1;
	}
	
	m_dwGroupsCount = headExt2.dwNbGroups;
	m_dwTotalBlocksCount = headExt2.dwTotalBlocksNb;
	m_dwBitmapSize = headExt2.dwBitmapSize;
	m_dwFirstBlock =  headExt2.dwFirstBlock;
	m_dwFreeBlocksCount = headExt2.dwFreeBlocksNb;
	m_dwUsedBlocksCount = headExt2.dwUsedBlocksNb;
	m_dwBlockSize = headExt2.dwBlockSize;
	strncpy(m_szLabel, headExt2.szLabel, 128);

	// ---- printf informations on the original partition
	printfInformations();
	
	nRes = newtWinChoice(i18n("Confirmation"), i18n("Yes"), i18n("No"), i18n("Do you really want to restore %s from the image file %s ?"), m_szDevice, image->m_szImageFilename);
	if (nRes == 2)
		return -1;
	
	// ---- copy boot sector
	newtPushHelpLine(i18n("copying boot sector"));
	
	nRes = m_image->readAndCheckMagic(MAGIC_BEGIN_BOOT, *options);
	if (nRes == -1)
		return -1;
	
	// TODO: Compare the copies of the boot
	
	for (i=0; i < NB_BOOT_COPIES_IN_IMAGE; i++)
	{	
		nRes = m_image->read(cBootBuffer, EXT2_BOOTSECT_LEN, image->m_bIsSplitted, *options, true);
		if (nRes != EXT2_BOOTSECT_LEN)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read boot sector"));
			return -1;
		}	
	}
	
	fseek(m_fDeviceFile, 0L, SEEK_SET); // go to the begin of the partition
	nRes = fwrite (cBootBuffer, EXT2_BOOTSECT_LEN, 1, m_fDeviceFile);
	if (nRes != 1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read boot sector"));
		return -1;
	}
	
	// ---- read bitmap
	newtPushHelpLine(i18n("copying blocks bitmap"));
	
	nRes = m_image->readAndCheckMagic(MAGIC_BEGIN_BITMAP, *options);
	if (nRes == -1)
		return -1;
	
	cBitmap = new char [headExt2.dwBitmapSize];
	if (cBitmap == 0)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read bitmap from image file"));
		return -1;
	}
		
	nRes = m_image->read(cBitmap, headExt2.dwBitmapSize, image->m_bIsSplitted, *options, true);
	if (nRes != headExt2.dwBitmapSize)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read bitmap from image file"));
		return -1;
	}
	
	// ---- restore all blocks
	newtPushHelpLine(i18n("copying used data blocks"));
	
	nRes = m_image->readAndCheckMagic(MAGIC_BEGIN_DATABLOCKS, *options);
	if (nRes == -1)
		return -1;
	
	// go to the beginning of the first block
	fseek(m_fDeviceFile, (long) (headExt2.dwFirstBlock * headExt2.dwBlockSize), SEEK_SET);
	dwRestoredBlocks = 0;
	
	for (i=headExt2.dwFirstBlock; i < headExt2.dwTotalBlocksNb; i++)
	{
		// if block is USED
		if (ext2fs_test_bit(i-headExt2.dwFirstBlock, cBitmap))
		{	
			dwRestoredBlocks++;
			
			nRes = m_image->read(cBlockBuffer, headExt2.dwBlockSize, image->m_bIsSplitted, *options, true);
			if (nRes != headExt2.dwBlockSize)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read block %ld/%ld"), i, headExt2.dwTotalBlocksNb);
				return -1;
			}
			
			nRes = fwrite(cBlockBuffer, headExt2.dwBlockSize, 1, m_fDeviceFile);
			if (nRes != 1)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write block %ld"), i);
				return -1;
			}
		}
		else // if block is FREE
		{	
			if (options->bEraseWithNull == true)
			{	nRes = fwrite(cNullBuffer, headExt2.dwBlockSize, 1, m_fDeviceFile);
				if (nRes != 1)
				{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write null block %ld"), i);
					return -1;
				}
				dwRestoredBlocks++;
			}
			else // skip free block with fseek()
			{	fseek(m_fDeviceFile, headExt2.dwBlockSize, SEEK_CUR);
			
			}
		
		}
		
		// draw progress
		if (i%20 == 0)
			m_guiRestore->setProgressValue(dwRestoredBlocks, headExt2.dwUsedBlocksNb, headExt2.dwTotalBlocksNb, options->bEraseWithNull);
	
	}
	
	m_guiRestore->setProgressValue(dwRestoredBlocks, headExt2.dwUsedBlocksNb, headExt2.dwTotalBlocksNb, options->bEraseWithNull);
		
	// ---- end
	delete []cBitmap;
	
	return 0;
}

// =======================================================
int CExt2Part::printfInformations()
{
	char szTemp[2048];
	char szTemp2[2048];
	char szText[8192];
	
	sprintf(szText,  i18n("Number of groups..............%lu\n"), m_dwGroupsCount);
	sprintf(szTemp2, i18n("Number of blocks..............%lu\n"), m_dwTotalBlocksCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Bitmap size...................%s (%lu bytes)\n"), formatSize(m_dwBitmapSize, szTemp), m_dwBitmapSize);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("First block...................%lu\n"), m_dwFirstBlock);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Block size....................%lu bytes\n"), m_dwBlockSize);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Free blocks count.............%lu\n"), m_dwFreeBlocksCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Used blocks count.............%lu\n"), m_dwUsedBlocksCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Space usage:..................%lu %%\n"), (int) (100 * m_dwUsedBlocksCount)/m_dwTotalBlocksCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Used size.....................%s\n"), formatSize(m_dwUsedBlocksCount * m_dwBlockSize, szTemp));
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Free size.....................%s\n"), formatSize(m_dwFreeBlocksCount * m_dwBlockSize, szTemp));
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Label:........................%s\n"), m_szLabel);
	strcat(szText, szTemp2);

	fprintf(m_fDebug, "\n====================\n%s====================\n\n", szText); fflush(m_fDebug);
	newtWinMessage(i18n("Ext2fs informations"), i18n("Ok"), szText);
						
	return 0;
}
			
