/***************************************************************************
                          cext2part.h  -  description
                             -------------------
    begin                : Mon May 22 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CEXT2PART_H
#define CEXT2PART_H

#include "partimage.h"
#include "common.h"

extern "C"
{
#include "ext2fs/ext2fs.h" // ext2fs library
}

struct CImage;
struct CVolumeHeader;
struct CMainFooter;
struct CMainHeader;
struct COptions;
class CSaveWindow;
class CRestoreWindow;

#define EXT2_BOOTSECT_LEN     1024
#define CURRENT_EXT2_FORMAT            "0.3" // format version of the ext2fs image

// ================================================
class CExt2Part
{
public:
	CExt2Part(char *szDevice, FILE *fDeviceFile);
	~CExt2Part();
	int printfInformations();
	
	// save options
	int saveToImage(CImage *image, COptions *options, CSaveWindow *gui);
	int writeGroupInImageFile(DWORD dwGroupNum, DWORD *dwNbWrittenBlocks, COptions *options);

	// restore options
	int restoreFromImage(CImage *image, COptions *options, CRestoreWindow *gui);

private:
	ext2_filsys	m_fs;
	CImage *m_image;
	CSaveWindow *m_guiSave;
	CRestoreWindow *m_guiRestore;
	
	char m_szDevice[MAX_DEVICENAMELEN+1]; // "/dev/hdc5"
	char m_szImageFilename[MAXPATHLEN+1]; // ex: /home/user/filename.partimg.gz.002 (if splitted)
	char m_szOriginalFilename[MAXPATHLEN+1]; // ex: filename.partimg.gz
	char m_szPath[MAXPATHLEN+1]; // ex: /home/user/
	
	// informations
	DWORD m_dwGroupsCount;
	DWORD m_dwTotalBlocksCount;
	DWORD m_dwBitmapSize;
	DWORD m_dwFirstBlock;
	DWORD m_dwFreeBlocksCount;
	DWORD m_dwUsedBlocksCount;
	char m_szLabel[129];
	
	// volumes
	DWORD m_dwCurrentVolumeSize;
	DWORD m_dwVolumeNumber;
		
	DWORD m_dwBlockSize;
	DWORD m_dwUsedBlocks;
	
	// File descriptors
	FILE *m_fDeviceFile;

	//gzFile m_gzImageFile;
	//FILE *m_fImageFile;
	//BZFILE *m_bzImageFile;
};

// ================================================
struct CExt2Header // size must be 16384 (adjust the reserved data)
{
	char szAppFormat[20]; // Version of the format of the application file
	DWORD dwExt2Flags;
	
	DWORD dwNbBootCopies; // Number of copies of the boot sector in the image file
	
	DWORD dwBlockSize;
	DWORD dwBootSize;
	DWORD dwBitmapSize;
	
	DWORD dwTotalBlocksPerGroup;
	DWORD dwTotalBlocksNb;
	DWORD dwFreeBlocksNb;
	DWORD dwUsedBlocksNb;
	
	DWORD dwFirstBlock; // After boot
	DWORD dwNbGroups; // Number of groups (with desc, bitmaps, data, ...)
	
	char szLabel[128];

	unsigned char cReserved[16192]; // Adjust to fit with total header size
};

#endif //CEXT2PART_H
