/***************************************************************************
                          cfatpart.cpp  -  description
                             -------------------
    begin                : Thu Jul 13 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>

#include "cfatpart.h"
#include "partimage.h"
#include "imagefile.h"
#include "misc.h"
#include "gui_text.h"

#define fat32to28(A) (A & 0x0FFFFFFF)

// =======================================================
CFatPart::CFatPart(char *szDevice, FILE *fDeviceFile)
{
	fprintf(m_fDebug, "begin of CFatPart::CFatPart()\n"); fflush(m_fDebug);  
	m_fDeviceFile = fDeviceFile;
	strncpy(m_szDevice, szDevice, MAX_DEVICENAMELEN);
	fprintf(m_fDebug, "end of CFatPart::CFatPart()\n"); fflush(m_fDebug);  
}

// =======================================================
CFatPart::~CFatPart()
{
}

// =======================================================
// -1: error
// 0: success
int CFatPart::checkFatDamagedClusters(DWORD *dwDamagedClusters, DWORD *dwUsedClusters, DWORD *dwFreeClusters)
{	
	int nRes;
	DWORD dwReservedBytes;
	DWORD i;
	UINT cFat16Entry;
	DWORD dwFat32Entry;
	
	// init
	dwReservedBytes = m_nReservedSectorsCount * m_nBytesPerSector;
	if (dwUsedClusters) *dwUsedClusters = 0L;
	if (dwDamagedClusters) *dwDamagedClusters = 0L;
	if (dwFreeClusters) *dwFreeClusters = 0L;
	
	// skip reserved data and two first FAT entries
	if (m_dwFileSystem == FS_FAT16)
		fseek(m_fDeviceFile, dwReservedBytes+2*2, SEEK_SET);
	else if (m_dwFileSystem == FS_FAT32) 
		fseek(m_fDeviceFile, dwReservedBytes+2*4, SEEK_SET);
	else
		return -1;
	
	// for each FAT entry...
	for (i=0; i < m_dwClustersCount; i++)
	{
		// If FAT16
		if (m_dwFileSystem == FS_FAT16)
		{	nRes = fread(&cFat16Entry, sizeof(UINT), 1, m_fDeviceFile);
			if (nRes != 1)
				return -1;
				
			if (cFat16Entry == 0xFFF7) // bad FAT16 cluster
			{
				if (dwDamagedClusters)
					(*dwDamagedClusters)++;
				m_bitmap.setBit(i, false);
			}
			else if (cFat16Entry == 0x0000) // free
			{
				if (dwFreeClusters)
					(*dwFreeClusters)++;
				m_bitmap.setBit(i, false);	
			}
			else // used
			{
				if (dwUsedClusters)
					(*dwUsedClusters)++;
				m_bitmap.setBit(i, true);
			}
		}
	
		// If FAT32
		else if (m_dwFileSystem == FS_FAT32)
		{	nRes = fread(&dwFat32Entry, sizeof(DWORD), 1, m_fDeviceFile);
			if (nRes != 1)
				return -1;
				
			/*if (i < 50)
				printf("FAT entry (%10lu=%.8X) = %.8X\n", i, i, dwFat32Entry);*/
			
			if (fat32to28(dwFat32Entry) == 0x0FFFFFF7) // bad FAT32 cluster
			{	
			    if (dwDamagedClusters)
				(*dwDamagedClusters)++;
			    m_bitmap.setBit(i, false);    
			}
			else if (fat32to28(dwFat32Entry) == 0x00000000) // free
			{	
			    if (dwFreeClusters)
				(*dwFreeClusters)++;
			    m_bitmap.setBit(i, false);    
			}
			else // used
			{
			    if (dwUsedClusters)
				(*dwUsedClusters)++;
			    m_bitmap.setBit(i, true);
			}
		}
		
		// If other file system
		else
		{	return -1;	
		}

	}
	
	return 0; // success
}


// =======================================================
// -1: error
// false: no differences between FATs
// true: differences between FATs
int CFatPart::compareFatCopies()
{	
	int nResult;
	int nRes;
	DWORD dwReservedBytes;
	DWORD i;
	UINT cFat16Entry1;
	UINT cFat16Entry2;
	DWORD dwFat32Entry1;
	DWORD dwFat32Entry2;
	
	// init
	nResult = 0;
	dwReservedBytes = m_nReservedSectorsCount * m_nBytesPerSector;
	
	// for each FAT entry...
	for (i=0; i < m_dwClustersCount; i++)
	{
		// If FAT16
		if (m_dwFileSystem == FS_FAT16)
		{	
			// read first FAT entry
			nRes = fseek(m_fDeviceFile, dwReservedBytes+(2*i), SEEK_SET);
			if (nRes == -1)
				return -1;
				
			nRes = fread(&cFat16Entry1, sizeof(UINT), 1, m_fDeviceFile);
			if (nRes != 1)
				return -1;
				
			// read second FAT entry
			nRes = fseek(m_fDeviceFile, dwReservedBytes+(m_dwSectorsPerFAT*m_nBytesPerSector)+(2*i), SEEK_SET);
			if (nRes == -1)
				return -1;
				
			nRes = fread(&cFat16Entry2, sizeof(UINT), 1, m_fDeviceFile);
			if (nRes != 1)
				return -1;
				
			if (cFat16Entry1 != cFat16Entry2)
				return true;
		}
	
		// If FAT32
		else if (m_dwFileSystem == FS_FAT32)
		{	
			// read first FAT entry
			nRes = fseek(m_fDeviceFile, dwReservedBytes+(4*i), SEEK_SET);
			if (nRes == -1)
				return -1;
				
			nRes = fread(&dwFat32Entry1, sizeof(DWORD), 1, m_fDeviceFile);
			if (nRes != 1)
				return -1;
				
			// read first FAT entry
			nRes = fseek(m_fDeviceFile, dwReservedBytes+(m_dwSectorsPerFAT*m_nBytesPerSector)+(4*i), SEEK_SET);
			if (nRes == -1)
				return -1;
				
			nRes = fread(&dwFat32Entry2, sizeof(DWORD), 1, m_fDeviceFile);
			if (nRes != 1)
				return -1;
				
			if (fat32to28(dwFat32Entry1) != fat32to28(dwFat32Entry2)) // bad FAT32 cluster
				return true;
		}
		
		// If other file system
		else
		{	return -1;	
		}

	}

	return 0;
}

// =======================================================
int CFatPart::saveUsedClustersToImageFile(COptions *options)
{
	int nRes;
	DWORD dwReservedBytes;
	DWORD i;
	DWORD dwDataBegin;
	DWORD dwBytesPerCluster;
	char cBuffer[MAX_FAT_CLUSTER_SIZE+1];
	
	// init
	dwReservedBytes = m_nReservedSectorsCount * m_nBytesPerSector;
	dwDataBegin = (m_nReservedSectorsCount + (m_cNumberOfFATs * m_dwSectorsPerFAT) + m_dwRootDirSectors) * m_nBytesPerSector;
	dwBytesPerCluster = m_nBytesPerSector * m_cSectorsPerCluster;
	
	DWORD dwSavedClusters = 0L;
	
	// go to the beginning of the data clusters
	nRes = fseek(m_fDeviceFile, (long) dwDataBegin, SEEK_SET);
	//printf("data begin=%ld\n",  (long) dwDataBegin);
	if (nRes == -1)
	    return -1;
	
	// for each FAT entry...
	for (i=0; i < m_dwClustersCount; i++)
	{
		// If FAT16 or FAT32
		if ((m_dwFileSystem == FS_FAT16) || (m_dwFileSystem == FS_FAT32))
		{	
			// if cluster is used...
			if (m_bitmap.isBitSet(i)) // not damaged and not free
			{
				//printf ("%.4ld  ", i);
				dwSavedClusters++;
				
				errno = 0;
				nRes = fread(cBuffer, dwBytesPerCluster, 1, m_fDeviceFile);
				if (nRes != 1)
				{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read block %ld from image (%s)"),i, strerror(errno));
					return -1;
				}
				
				//printf ("dwSavedClusters=%ld\n", dwSavedClusters);
				errno = 0;
				nRes = m_image->write(cBuffer, dwBytesPerCluster, *options, true);
				if (nRes != dwBytesPerCluster)
				{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write block %ld to image. (%s)"), i, strerror(errno));
					return -1;
				}
			}
			else // block free --> skip
			{
				nRes = fseek(m_fDeviceFile, dwBytesPerCluster, SEEK_CUR);
				if (nRes == -1)
					return -1;	
			}
		}
	
		// If other file system
		else
		{	return -1;	
		}
		
		// draw progress
		if (dwSavedClusters%15 == 0)
			m_guiSave->setProgressValue(dwSavedClusters, m_dwUsedClusters);
	}
			
	m_guiSave->setProgressValue(m_dwUsedClusters, m_dwUsedClusters);
	
	return 0;
}

// =======================================================
int CFatPart::restoreUsedClustersToDisk(COptions *options)
{
	int nRes;
	DWORD dwReservedBytes;
	DWORD i;
	DWORD dwDataBegin;
	DWORD dwBytesPerCluster;
	BYTE cBuffer[MAX_FAT_CLUSTER_SIZE+1];
	BYTE cZeroBuffer[MAX_FAT_CLUSTER_SIZE+1];
	
	// init
	dwReservedBytes = m_nReservedSectorsCount * m_nBytesPerSector;
	dwDataBegin = (m_nReservedSectorsCount + (m_cNumberOfFATs * m_dwSectorsPerFAT) + m_dwRootDirSectors) * m_nBytesPerSector;
	dwBytesPerCluster = m_nBytesPerSector * m_cSectorsPerCluster;
	memset(cZeroBuffer, 0, MAX_FAT_CLUSTER_SIZE);
	
	DWORD dwSavedClusters = 0L;
	
	// go to the beginning of the data clusters
	nRes = fseek(m_fDeviceFile, (long) dwDataBegin, SEEK_SET);
	//printf("data begin=%ld\n",  (long) dwDataBegin);
	if (nRes == -1)
	    return -1;
	
	// for each FAT entry...
	for (i=0; i < m_dwClustersCount; i++)
	{
		//printf ("%.4ld  ", i);
	
		// If FAT16 or FAT32
		if ((m_dwFileSystem == FS_FAT16) || (m_dwFileSystem == FS_FAT32))
		{	
			if (m_bitmap.isBitSet(i)) // not damaged and not free
			{
				dwSavedClusters++;
					
				nRes = m_image->read(cBuffer, dwBytesPerCluster, m_image->m_bIsSplitted, *options, true);
				if (nRes != dwBytesPerCluster)
				{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read data block %ld from image file"), i);
					return -1;
				}	
				
				errno = 0;
				nRes = fwrite(cBuffer, dwBytesPerCluster, 1, m_fDeviceFile);
				if (nRes != 1)
				{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't restore data block %ld from image file. (%s) (%d)"), i, strerror(errno));
					return -1;
				}		
			}
			else // if cluster is free
			{
				if (options->bEraseWithNull == true) // if free and we must write 0 bytes on unused clusters
				{
					dwSavedClusters++;
					
					errno = 0;
					nRes = fwrite(cZeroBuffer, dwBytesPerCluster, 1, m_fDeviceFile);
					if (nRes != 1)
					{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write zero block %lu (%s) (%d)"), i, strerror(errno));
						return -1;
					}		
			  }
			  else // don't write zero bytes
			  {
					nRes = fseek(m_fDeviceFile, dwBytesPerCluster, SEEK_CUR);
					if (nRes == -1)
				    return -1;
			   }			
			}
		}
	
		// If other file system
		else
		{	
			return -1;	
		}

		// draw progress
		if (dwSavedClusters % 20 == 0)
			m_guiRestore->setProgressValue(dwSavedClusters, m_dwUsedClusters, m_dwClustersCount, options->bEraseWithNull);
	}
	m_guiRestore->setProgressValue(dwSavedClusters, m_dwUsedClusters, m_dwClustersCount, options->bEraseWithNull);
	
	return false;
}

// =======================================================
int CFatPart::printfInformations()
{
	char szTemp[2048];
	char szTemp2[2048];
	char szText[8192];
	
	sprintf (szText, "File system:..................");
	switch(m_dwFileSystem)
	{	case FS_FAT12:
			strcat(szText, "FAT12\n");
			break;
			
		case FS_FAT16:
			strcat(szText, "FAT16\n");
			break;
		
		case FS_FAT32:
			strcat(szText, "FAT32\n");
			break;
	}

	sprintf(szTemp2, i18n("Sector Size...................%u\n"), m_nBytesPerSector);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Sectors per cluster...........%u\n"), m_cSectorsPerCluster);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Reserved Sectors Count........%u\n"), m_nReservedSectorsCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Root directory sectors........%lu\n"), m_dwRootDirSectors);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Number Of FATs................%u\n"), m_cNumberOfFATs);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Total Sectors Count...........%lu\n"), m_dwTotalSectorsCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Sectors per FAT...............%lu\n"), m_dwSectorsPerFAT);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Number of clusters:...........%lu\n"), m_dwClustersCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Used clusters.................%lu\n"), m_dwUsedClusters);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Free clusters.................%lu\n"), m_dwFreeClusters);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Damaged clusters..............%lu\n"), m_dwDamagedClusters);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Used space....................%lu %%\n"), (100 * m_dwUsedClusters)/m_dwClustersCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Used size.....................%s\n"), formatSize(m_dwUsedClusters * m_cSectorsPerCluster * m_nBytesPerSector, szTemp));
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Free size.....................%s\n"), formatSize(m_dwFreeClusters * m_cSectorsPerCluster * m_nBytesPerSector, szTemp));
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Label:........................%s\n"), m_szLabel);
	strcat(szText, szTemp2);

	fprintf(m_fDebug, "\n====================\n%s====================\n\n", szText); fflush(m_fDebug);
	newtWinMessage(i18n("FAT informations"), i18n("Ok"), szText);
						
	return 0;
}			

// =======================================================
int CFatPart::readInformations()
{
	BYTE cBoot[MAX_BOOT_SIZE];
	UINT nTemp;
	DWORD dwTemp;
	BYTE cTemp;
	int nRes;
	
	// 1. read and print important informations
	fseek(m_fDeviceFile, 0, SEEK_SET);
	nRes = fread(&cBoot, MAX_BOOT_SIZE, 1, m_fDeviceFile);
	if (nRes != 1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read boot sector"));
		return -1;
	}
		
	// 2. calculate infos
	memcpy(&m_nBytesPerSector, cBoot+11, 2);
	memcpy(&m_cSectorsPerCluster, cBoot+13, 1);
	memcpy(&m_nReservedSectorsCount, cBoot+14, 2);
	memcpy(&m_cNumberOfFATs, cBoot+16, 1);
	memcpy(&m_nRootEntriesCount, cBoot+17, 2);
	
	// check data
	
	if ((m_nBytesPerSector % 512 != 0) || (m_nBytesPerSector < 512))
		return -1;
	if (m_cSectorsPerCluster == 0)
		return -1;
	if (m_nReservedSectorsCount == 0)
		return -1;
	if (m_cNumberOfFATs == 0)
		return -1;
	
	// Label
	memcpy(&cTemp, cBoot+38, 1);
	if (cTemp == 0x29) // there is a label
		memcpy(m_szLabel, cBoot+43, 11);
	else
		memset(m_szLabel, 0, 11);
	
	// Total number of sectors
	memcpy(&nTemp, cBoot+19, 2);
	memcpy(&dwTemp, cBoot+32, 4);
	if (nTemp)
		m_dwTotalSectorsCount = (DWORD)nTemp;
	else
		m_dwTotalSectorsCount = dwTemp;
		
	// Count of sectors occupied by one FAT
	memcpy(&nTemp, cBoot+22, 2);
	memcpy(&dwTemp, cBoot+36, 4);
	if (nTemp)
		m_dwSectorsPerFAT = (DWORD)nTemp;
	else
		m_dwSectorsPerFAT = dwTemp;		
	
	// calculate the FAT Type with the cluster count (FAT12 or FAT16 or FAT32)
	m_dwRootDirSectors = ((m_nRootEntriesCount * 32) + (m_nBytesPerSector-1)) / m_nBytesPerSector;
		
	m_dwDataSectors = m_dwTotalSectorsCount - (m_nReservedSectorsCount + (m_cNumberOfFATs * m_dwSectorsPerFAT) + m_dwRootDirSectors);
	
	if (m_cSectorsPerCluster == 0)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("m_cSectorsPerCluster can't be null. Check the partition to save is a FAT one.") );
		return -1;
	}
	
	m_dwClustersCount = m_dwDataSectors / m_cSectorsPerCluster;
	
	if (m_dwClustersCount < 4085)
	{	
		m_dwFileSystem = FS_FAT12;
		m_guiSave->showDetectedFS("FAT12");
	}
	else if (m_dwClustersCount < 65525)
	{
		m_dwFileSystem = FS_FAT16;
		m_guiSave->showDetectedFS("FAT16");
	}
	else
	{
		m_dwFileSystem = FS_FAT32;
		m_guiSave->showDetectedFS("FAT32");
	}
		
	return 0;
}

// =======================================================
int CFatPart::saveToImage(CImage *image, COptions *options, CSaveWindow *gui)
{
	image->m_dwCurrentVolumeSize = 0L;
	image->m_dwVolumeNumber = 0L;
	char cBuffer[65536];
	int nRes;	
	DWORD i, j; // for(;;)
	
	m_image = image;
	m_guiSave = gui;

	// ---- 0. Initializations
	
	// A. read FAT informations
	fprintf(m_fDebug, "begin of readInformations()\n"); fflush(m_fDebug);  
	nRes = readInformations();
	fprintf(m_fDebug, "end of readInformations()\n"); fflush(m_fDebug);  
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read partition boot sector, or not a FAT partition. Check the file system is a FAT."));
		return -1;
	}	

	// B. check partition has a FAT file system	
	if (m_dwFileSystem != FS_FAT16 && m_dwFileSystem != FS_FAT32)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("%s is not a valid FAT16 or FAT32 partition. Cannot save it."), m_szDevice);
		return -1;
	}	
		
	m_bitmap.init(m_dwClustersCount);
	
	// C. check there are no damaged clusters, and ask what to do it true
	// and create the bitmap from the FAT
	newtPushHelpLine(i18n("reading and checking FAT entries. Please wait..."));
	fprintf(m_fDebug, "begin of checkFatDamagedClusters()\n"); fflush(m_fDebug);  
	nRes = checkFatDamagedClusters(&m_dwDamagedClusters, &m_dwUsedClusters, &m_dwFreeClusters);
	fprintf(m_fDebug, "end of checkFatDamagedClusters()\n"); fflush(m_fDebug);  
	if (nRes == true)
	{	
		nRes = newtWinChoice(i18n("Warning"), i18n("Ignore"), i18n("Cancel"), i18n("There are damaged blocks on the partition."));	
		if (nRes == 2)
			return -1;
	}
	
	// D. check all fat copies have no differences
	newtPushHelpLine(i18n("comparing copies of the FATs. Please wait..."));
	if (m_cNumberOfFATs >= 2)
	{	
		fprintf(m_fDebug, "begin of compareFatCopies()\n"); fflush(m_fDebug);  
		nRes = compareFatCopies();
		fprintf(m_fDebug, "end of compareFatCopies()\n"); fflush(m_fDebug);  
		if (nRes == -1)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Unable to compare FAT tables copies"));
			return -1;
		}
		else if (nRes == true) // there are differences
		{	
			nRes = newtWinChoice(i18n("Warning"), i18n("Ignore"), i18n("Cancel"), i18n("There are differences between the two copies of the FAT tables on the partition."));	
			if (nRes == 2)
				return -1;
		}
		else if (nRes == false) // there are no differences
		{	
		}
	}
	
	// E. print informations	
	newtPushHelpLine(i18n("informations about file system"));
	printfInformations();
	
	// ---- 1. Write FAT header
	CFatHeader headFat;
	memset(&headFat, 0, sizeof(CFatHeader));
	
	strcpy(headFat.szAppFormat, CURRENT_FAT_FORMAT);
	headFat.dwFileSystem = m_dwFileSystem;
	headFat.nBytesPerSector = m_nBytesPerSector;
	headFat.cSectorsPerCluster = m_cSectorsPerCluster;
	headFat.nReservedSectorsCount = m_nReservedSectorsCount;
	headFat.cNumberOfFATs = m_cNumberOfFATs;
	headFat.dwTotalSectorsCount = m_dwTotalSectorsCount;
	headFat.dwSectorsPerFAT = m_dwSectorsPerFAT;
	headFat.dwClustersCount = m_dwClustersCount;
	headFat.dwUsedClusters = m_dwUsedClusters;
	headFat.dwDamagedClusters = m_dwDamagedClusters;
	headFat.dwFreeClusters = m_dwFreeClusters;
	headFat.dwRootDirSectors = m_dwRootDirSectors;

	memcpy(headFat.szLabel, m_szLabel, 11);
	
	errno = 0;
	nRes = m_image->write(&headFat, sizeof(CFatHeader), *options, true);
	if (nRes != sizeof(CFatHeader))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write fat16 header: (%s)"), strerror(errno));
		return -1;
	}

	// ---- 3. copy reserved sectors (boot in general)
	nRes = m_image->writeMagic(MAGIC_BEGIN_RESERVED, *options);
	if (nRes == -1)
		return -1;
		
	DWORD dwReservedBytes = m_nReservedSectorsCount * m_nBytesPerSector;
	
	fseek(m_fDeviceFile, 0, SEEK_SET);
	nRes = fread(cBuffer, dwReservedBytes, 1, m_fDeviceFile);
	if (nRes != 1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read boot sector"));
		return -1;
	}
	
	for (i=0; i < NB_BOOT_COPIES_IN_IMAGE; i++)
	{	
		nRes = m_image->write(cBuffer, dwReservedBytes, *options, true);
		if (nRes != dwReservedBytes)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write boot sector"));
			return -1;
		}
	}
	
	// ---- 4. write two copies of the FATs
	newtPushHelpLine(i18n("copying FAT tables. Please wait..."));
	
	nRes = m_image->writeMagic(MAGIC_BEGIN_FAT, *options);
	if (nRes == -1)
		return -1;
	
	fseek(m_fDeviceFile, dwReservedBytes, SEEK_SET);
	
	for (i=0; i < NB_FAT_COPIES_IN_IMAGE; i++)
	{
		// go at the beginning of the first FAT table
		fseek(m_fDeviceFile, dwReservedBytes, SEEK_SET);
		
		for (j=0; j < m_dwSectorsPerFAT; j++)
		{
			nRes = fread(cBuffer, m_nBytesPerSector, 1,  m_fDeviceFile);
			if (nRes != 1)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read FAT tables"));
				return -1;
			}
			nRes = m_image->write(cBuffer, m_nBytesPerSector, *options, true);
			if (nRes != m_nBytesPerSector)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write FAT tables in the image file"));
				return -1;
			}
		}
	}
		
	// ---- 4. Write root dir sectors
	if (m_dwFileSystem == FS_FAT16)
	{
		nRes = m_image->writeMagic(MAGIC_BEGIN_ROOTDIR, *options);
		if (nRes == -1)
			return -1;
		
		DWORD dwDataLen = m_nBytesPerSector * m_dwRootDirSectors;
		DWORD dwBeginRootDir = (m_nReservedSectorsCount + (m_cNumberOfFATs * m_dwSectorsPerFAT)) * m_nBytesPerSector;
	
		fseek(m_fDeviceFile, dwBeginRootDir, SEEK_SET); // after FAT tables
		
		// read a sector of the FAT table
		nRes = fread(cBuffer, dwDataLen, 1, m_fDeviceFile);
		if (nRes != 1)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read root directory"));
			return -1;
		}	
				
		// write the root directory to the image file
		errno = 0;
		nRes = m_image->write(cBuffer, dwDataLen, *options, true);
		if (nRes != dwDataLen)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write root directory (%s)"), strerror(errno));
			return -1;
		}	
	}
	
	// ---- 5. Write all used clusters
	newtPushHelpLine(i18n("copying used data blocks"));
	
	nRes = m_image->writeMagic(MAGIC_BEGIN_DATABLOCKS, *options);
	if (nRes == -1)
		return -1;
		
	fprintf(m_fDebug, "begin of saveUsedClustersToImageFile()\n"); fflush(m_fDebug);  
	nRes = saveUsedClustersToImageFile(options);
	fprintf(m_fDebug, "end of saveUsedClustersToImageFile()\n"); fflush(m_fDebug);  
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Unable to copy used blocks to the image file"));
		return -1;
	}
	
	return 0;
}

// =======================================================
int CFatPart::restoreFromImage(CImage *image, COptions *options, CRestoreWindow *gui)
{
	CFatHeader headFat;
	char cBuffer[65536];
	int nRes;
	DWORD i, j;

	// init
	m_image = image;
	m_guiRestore = gui;

	newtPushHelpLine(i18n("reading partition informations"));
	
	// 1. ---- Read FAT header	
	memset(&headFat, 0, sizeof(CFatHeader));
	nRes = m_image->read(&headFat, sizeof(CFatHeader), m_image->m_bIsSplitted, *options, true);
	if (nRes != sizeof(CFatHeader))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read FAT header from the image file"));
		return -1;
	}
	
	// check the current format is supported
	if (strncmp(headFat.szAppFormat, CURRENT_FAT_FORMAT, strlen(CURRENT_FAT_FORMAT)) < 0)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("The partition was saved with a newer version of Partition Image. Please download the last version."));
		return -1;
	}	
	
	m_dwFileSystem = headFat.dwFileSystem;
	m_nBytesPerSector = headFat.nBytesPerSector;
	m_cSectorsPerCluster = headFat.cSectorsPerCluster;
	m_nReservedSectorsCount = headFat.nReservedSectorsCount;
	m_cNumberOfFATs = headFat.cNumberOfFATs;
	m_dwTotalSectorsCount = headFat.dwTotalSectorsCount;
	m_dwSectorsPerFAT = headFat.dwSectorsPerFAT;
	m_dwClustersCount = headFat.dwClustersCount;
	m_dwUsedClusters = headFat.dwUsedClusters;
	m_dwDamagedClusters = headFat.dwDamagedClusters;
	m_dwFreeClusters = headFat.dwFreeClusters;
	m_dwRootDirSectors = headFat.dwRootDirSectors;
	memcpy(m_szLabel, headFat.szLabel, 11);
	
	// show FAT file system type
	switch (m_dwFileSystem)
	{	case  FS_FAT12:
			m_guiRestore->showFS("FAT12");
			break;
			
		case  FS_FAT16:
			m_guiRestore->showFS("FAT16");
			break;
			
		case  FS_FAT32:
			m_guiRestore->showFS("FAT32");
			break;
			
	};
		
	// 2. ---- Print informations and ask confirmation to restore
	printfInformations();
	
	// ask confirmation
	nRes = newtWinChoice(i18n("Confirmation"), i18n("Yes"), i18n("No"), i18n("Do you really want to restore %s from the image file %s ?"), m_szDevice, image->m_szImageFilename);
	if (nRes == 2)
		return -1;
	
	// 3. ---- Restore reserved informations
	newtPushHelpLine(i18n("copying reserved sectors (boot)"));
	
	nRes = m_image->readAndCheckMagic(MAGIC_BEGIN_RESERVED, *options);
	if (nRes == -1)
		return -1;
	
	DWORD dwReservedBytes = m_nReservedSectorsCount * m_nBytesPerSector;
	
	for (i=0; i < NB_BOOT_COPIES_IN_IMAGE; i++)
	{
		nRes = m_image->read(cBuffer, dwReservedBytes, m_image->m_bIsSplitted, *options, true);
		if (nRes != dwReservedBytes)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read boot sector from image file"));
			return -1;
		}	
	}
	
	// TODO: compare copies of the reserved data
	
	fseek(m_fDeviceFile, 0, SEEK_SET);
	nRes = fwrite(cBuffer, dwReservedBytes, 1, m_fDeviceFile);
	if (nRes != 1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write boot sector"));
		return -1;
	}
	
	// ---- 4. restore FAT tables
	newtPushHelpLine(i18n("copying FAT tables. Please wait..."));
	
	nRes = m_image->readAndCheckMagic(MAGIC_BEGIN_FAT, *options);
	if (nRes == -1)
		return -1;
	
	fseek(m_fDeviceFile, dwReservedBytes, SEEK_SET); // beginning of the first FAT table
	
	for (i=0; i < m_cNumberOfFATs; i++)
	{
		for (j=0; j < m_dwSectorsPerFAT; j++)
		{
			// read a sector of the FAT table in image file
			nRes = m_image->read(cBuffer, m_nBytesPerSector, m_image->m_bIsSplitted, *options, true);
			if (nRes != m_nBytesPerSector)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read FAT table from image file"));
				return -1;
			}	
			// write the sector to the disk
			errno = 0;
			nRes = fwrite(cBuffer, m_nBytesPerSector, 1, m_fDeviceFile);
			if (nRes != 1)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write FAT table on partition. (%s)"), strerror(errno));
				return -1;
			}
		}
	}

	// ---- 5. create bitmap from FAT tables
	newtPushHelpLine(i18n("reading FAT tables. Please wait..."));
	m_bitmap.init(m_dwClustersCount);

	fprintf(m_fDebug, "begin of checkFatDamagedClusters()\n"); fflush(m_fDebug);  
	nRes = checkFatDamagedClusters(NULL, NULL, NULL);
	fprintf(m_fDebug, "end of checkFatDamagedClusters()\n"); fflush(m_fDebug);  
	if (nRes == -1)
		return -1;

	// ---- 6. restore root directory
	if (m_dwFileSystem == FS_FAT16)
	{
		nRes = m_image->readAndCheckMagic(MAGIC_BEGIN_ROOTDIR, *options);
		if (nRes == -1)
			return -1;
	
		DWORD dwDataLen = m_nBytesPerSector * m_dwRootDirSectors;
		DWORD dwBeginRootDir = (m_nReservedSectorsCount + (m_cNumberOfFATs * m_dwSectorsPerFAT)) * m_nBytesPerSector;
	
		nRes = m_image->read(cBuffer, dwDataLen, m_image->m_bIsSplitted, *options, true);
		if (nRes != dwDataLen)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read root directory from image file"));
			return -1;
		}	
			
		fseek(m_fDeviceFile, dwBeginRootDir, SEEK_SET); // beginning of the first FAT table
		errno = 0;
		nRes = fwrite(cBuffer, dwDataLen, 1, m_fDeviceFile);
		if (nRes != 1)
		{
			newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write root directory (%s)"), strerror(errno));
			return -1;
		}
	}
	
	// ---- 7. Write all used clusters
	newtPushHelpLine(i18n("copying used data blocks"));
	
	nRes = m_image->readAndCheckMagic(MAGIC_BEGIN_DATABLOCKS, *options);
	if (nRes == -1)
		return -1;
	
	errno = 0;
	fprintf(m_fDebug, "begin of restoreUsedClustersToDisk()\n"); fflush(m_fDebug);  
	nRes = restoreUsedClustersToDisk(options);
	fprintf(m_fDebug, "end of restoreUsedClustersToDisk()\n"); fflush(m_fDebug);  
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Unable to restore used blocks to disk. (%s)"), strerror(errno));
		return -1;
	}
		
	return 0;
}

