/***************************************************************************
                          cfatpart.h  -  description
                             -------------------
    begin                : Thu Jul 13 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CFATPART_H
#define CFATPART_H


#include "partimage.h"
#include "common.h"
#include "misc.h"

struct CImage;
struct CVolumeHeader;
struct CMainFooter;
struct CMainHeader;
struct COptions;
class CSaveWindow;
class CRestoreWindow;

#define CURRENT_FAT_FORMAT					  "0.2" // format version of the fat image
#define MAX_BOOT_SIZE									4096
#define NB_FAT_COPIES_IN_IMAGE        2
#define MAX_FAT_CLUSTER_SIZE					65536

// ================================================
struct CFatHeader // size must be 16384 (adjust the reserved data)
{
	char szAppFormat[20]; // Version of the format of the application file
	DWORD dwFatFlags;
	char szLabel[64];
	DWORD dwFileSystem; // FS_FAT12 or FS_FAT16 or FS_FAT32
	
	DWORD dwNbBootCopies; // Number of copies of the boot sector in the image file
	DWORD dwNbFatCopies;
	
	UINT nBytesPerSector;
	BYTE cSectorsPerCluster;
	UINT nReservedSectorsCount;
	BYTE cNumberOfFATs;
	DWORD dwTotalSectorsCount;
	UINT nRootEntriesCount;
	DWORD dwClustersCount;
	DWORD dwRootDirSectors;
	DWORD dwRootEntriesCount;
	DWORD dwSectorsPerFAT;
	DWORD dwDataSectors;
	DWORD dwUsedClusters;
	DWORD dwDamagedClusters;
	DWORD dwFreeClusters;
	
	unsigned char cReserved[16236]; // Adjust to fit with total header size
};

// ================================================
class CFatPart
{
public:CFatPart(char *szDevice, FILE *fDeviceFile);
	~CFatPart();
	
	// save options
	int saveToImage(CImage *image, COptions *options, CSaveWindow *gui);

	// restore options
	int restoreFromImage(CImage *image, COptions *options, CRestoreWindow *gui);
	
private:
	int printfInformations();
	int readInformations();
	int checkFatDamagedClusters(DWORD *dwDamagedClusters, DWORD *dwUsedClusters, DWORD *dwFreeClusters);
	int compareFatCopies();
	int saveUsedClustersToImageFile(COptions *options);
	int restoreUsedClustersToDisk(COptions *options);

private:
	CImage *m_image;
	CBlocksUseBitmap m_bitmap;
	CSaveWindow *m_guiSave;
	CRestoreWindow *m_guiRestore;

	char m_szDevice[MAX_DEVICENAMELEN+1]; // "/dev/hdc5"
	char m_szImageFilename[MAXPATHLEN+1]; // ex: /home/user/filename.partimg.gz.002 (if splitted)
	char m_szOriginalFilename[MAXPATHLEN+1]; // ex: filename.partimg.gz
	char m_szPath[MAXPATHLEN+1]; // ex: /home/user/
	
	// informations
	UINT m_nBytesPerSector;
	BYTE m_cSectorsPerCluster;
	UINT m_nReservedSectorsCount;
	BYTE m_cNumberOfFATs;
	DWORD m_dwTotalSectorsCount;
	UINT m_nRootEntriesCount;
	DWORD m_dwClustersCount;
	DWORD m_dwRootDirSectors;
	DWORD m_dwRootEntriesCount;
	DWORD m_dwSectorsPerFAT;
	DWORD m_dwDataSectors;
	DWORD m_dwFileSystem;
	DWORD m_dwUsedClusters;
	DWORD m_dwDamagedClusters;
	DWORD m_dwFreeClusters;
	char m_szLabel[11];
	
	// volumes
	DWORD m_dwCurrentVolumeSize;
	DWORD m_dwVolumeNumber;
		
	// File descriptors
	FILE *m_fDeviceFile;

	//gzFile m_gzImageFile;
	//FILE *m_fImageFile;
	//BZFILE *m_bzImageFile;
};

#endif // CFATPART_H
