/***************************************************************************
                          chpfspart.cpp  -  description
                             -------------------
    begin                : Thu Jul 13 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>

#include <linux/types.h>

#include "chpfspart.h"
#include "partimage.h"
#include "imagefile.h"
#include "misc.h"
#include "gui_text.h"

#include "cntfspart.h"

// =======================================================
CHpfsPart::CHpfsPart(char *szDevice, FILE *fDeviceFile)
{
	fprintf(m_fDebug, "begin of CHpfsPart::CHpfsPart()\n"); fflush(m_fDebug);
	m_fDeviceFile = fDeviceFile;
	strncpy(m_szDevice, szDevice, MAX_DEVICENAMELEN);
	fprintf(m_fDebug, "end of CHpfsPart::CHpfsPart()\n"); fflush(m_fDebug);
}

// =======================================================
CHpfsPart::~CHpfsPart()
{
}

// =======================================================
int CHpfsPart::saveUsedBlocksToImageFile(COptions *options)
{
	int nRes;
	DWORD i;
	char cBuffer[1024];
	
	DWORD dwSavedBlocks = 0L;

	// go to the beginning of the disk
	nRes = fseek(m_fDeviceFile, 0L, SEEK_SET);
	if (nRes == -1)
	    return -1;
	
	for (i=0; i < m_dwBlocksCount; i++)
	{
		// if block is used...
		if (m_bitmap.isBitSet(i)) // not damaged and not free
		{
			dwSavedBlocks++;
			//fprintf(m_fDebug, "%lu used\n", i); fflush(m_fDebug);
	
			errno = 0;
			nRes = fread(cBuffer, m_dwBlockSize, 1, m_fDeviceFile);
			if (nRes != 1)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read block %ld from partition (%s)"), i, strerror(errno));
				return -1;
			}
				
			errno = 0;
			nRes = m_image->write(cBuffer, m_dwBlockSize, *options, true);
			if (nRes != m_dwBlockSize)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write block %ld to image (%s)"), i, strerror(errno));
				return -1;
			}
		}
		else // block free --> skip
		{
			//fprintf(m_fDebug, "%lu free\n", i); fflush(m_fDebug);
			nRes = fseek(m_fDeviceFile, m_dwBlockSize, SEEK_CUR);
			if (nRes == -1)
				return -1;	
		}
		
		// draw progress
		if (dwSavedBlocks%20 == 0)
			m_guiSave->setProgressValue(dwSavedBlocks, m_dwUsedBlocks);
	}
		
	m_guiSave->setProgressValue(m_dwUsedBlocks, m_dwUsedBlocks);

	return 0;
}

// =======================================================
int CHpfsPart::copyBitmap()
{
	BYTE cBitmapList[4096+1];
	BYTE cBuffer[2048+1];
	DWORD dwOffset;
	DWORD dwPos;
	DWORD dwByte;
	DWORD *dwBuffer;
	QWORD qwBit;
	int i, j;
	int nRes;
	
	nRes = bigFseek(m_fDeviceFile, (m_dwBitmapPointer * m_dwBlockSize)); // SEEK_SET forced
	if (nRes == -1)
		return -1;
			
	nRes = fread(cBitmapList, 4096, 1, m_fDeviceFile);
	if (nRes != 1)
		return -1;
	
	fprintf(m_fDebug, "\n\n======== begin of Bitmap sectors list ============\n"); fflush(m_fDebug);
	
	// calculates bitmap blocks count
	m_dwBitmapQuadBlocksCount = 0L;
	do
	{
		// ---- Read entry of the list
		memcpy(&dwOffset, cBitmapList+(m_dwBitmapQuadBlocksCount*sizeof(DWORD)), sizeof(DWORD));
		if (dwOffset)
		{	fprintf(m_fDebug, "%lu, ", dwOffset); fflush(m_fDebug);
			m_dwBitmapQuadBlocksCount++;
		}

	} while ((m_dwBitmapQuadBlocksCount < 512) && (dwOffset));
	// The 4KB bitmap sectors list can contain 512 entries. Error if partition > 4 GB...
	
	fprintf(m_fDebug, "\n======== end of Bitmap sectors list ============\n\n\n"); fflush(m_fDebug);
		
	nRes = m_bitmap.init(m_dwBitmapQuadBlocksCount * 4 * 512 * 8);
	if (nRes == -1)
		return -1;
	
	dwBuffer = (DWORD *)cBuffer;
	qwBit = 0L;
	
	for (dwPos=0; dwPos < m_dwBitmapQuadBlocksCount; dwPos++)
	{
		// ---- Read entry of the list
		memcpy(&dwOffset, cBitmapList+(dwPos*sizeof(DWORD)), sizeof(DWORD));
		
		// ---- Read bitmap block
		nRes = bigFseek(m_fDeviceFile, (dwOffset * m_dwBlockSize));
		if (nRes == -1)
			return -1;
	
		nRes = fread(cBuffer, 2048, 1, m_fDeviceFile); // 2048 bytes = QuadBitmapBlock
		if (nRes != 1)
			return -1;
			
		// analyse buffer
		for (i=0; i < 2048 / sizeof(DWORD); i++)
		{	
			dwByte = dwBuffer[i];
				
			for (j=0; j < sizeof(DWORD) * 8; j++)
			{
				if (dwByte & 1) // block is free
					m_bitmap.setBit(qwBit, false);
				else // block is used
					m_bitmap.setBit(qwBit, true);
			
				dwByte>>=1;
				qwBit++;
			}
		}
	}
	
	// Free/Used blocks count
	m_dwUsedBlocks = 0L;
	for (dwPos=0; dwPos	< m_dwBlocksCount; dwPos++)
	{
		if (m_bitmap.isBitSet(dwPos))
			m_dwUsedBlocks++;
	}	

  return 0;
}			

// =======================================================
int CHpfsPart::printfInformations()
{
	char szTemp[2048];
	char szTemp2[2048];
	char szText[8192];
	
	sprintf(szText,  i18n("Block size....................%lu bytes\n"), m_dwBlockSize);
	sprintf(szTemp2, i18n("Total blocks count............%lu\n"), m_dwBlocksCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Used blocks count.............%lu\n"), m_dwUsedBlocks);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Free blocks count.............%lu\n"), m_dwBlocksCount-m_dwUsedBlocks);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Bitmap blocks count...........%lu\n"), 4*m_dwBitmapQuadBlocksCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Space usage:..................%lu %%\n"), (100 * m_dwUsedBlocks)/m_dwBlocksCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Used size.....................%s\n"), formatSize(m_dwUsedBlocks * m_dwBlockSize, szTemp));
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Free size.....................%s\n"), formatSize((m_dwBlocksCount-m_dwUsedBlocks) * m_dwBlockSize, szTemp));
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Label:........................%s\n"), m_szLabel);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("HPFS version:.................%d\n"), m_cHpfsVersion);
	strcat(szText, szTemp2);

	fprintf(m_fDebug, "\n====================\n%s====================\n\n", szText); fflush(m_fDebug);
	newtWinMessage(i18n("HPFS informations"), i18n("Ok"), szText);
						
  return 0;
}			

// =======================================================
int CHpfsPart::readInformations()
{
	int nRes;
	hpfs_boot_block boot;
	hpfs_super_block super;
	hpfs_spare_block spare;

	// read BOOT block
	nRes = fseek(m_fDeviceFile, 0L, SEEK_SET);
	if (nRes == -1)
		return -1;
	
	nRes = fread(&boot, sizeof(hpfs_boot_block), 1, m_fDeviceFile);
	if (nRes != 1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read boot block"));
		return -1;
	}
	
	// read SUPER block
	nRes = fseek(m_fDeviceFile, (long)(16 * 512), SEEK_SET);
	if (nRes == -1)
		return -1;
	
	nRes = fread(&super, sizeof(hpfs_super_block), 1, m_fDeviceFile);
	if (nRes != 1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read super block"));
		return -1;
	}
	
	// read SPARE block
	nRes = fseek(m_fDeviceFile, (long)(17 * 512), SEEK_SET);
	if (nRes == -1)
		return -1;
	
	nRes = fread(&spare, sizeof(hpfs_spare_block), 1, m_fDeviceFile);
	if (nRes != 1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read spare block"));
		return -1;
	}
	
	// check magic numbers
	if (/*(boot.magic != 0xaa55) || */(super.magic != 0xf995e849) || (spare.magic != 0xf9911849))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Invalid HPFS magic numbers. Not a valid HPFS partition."));
		return -1;
	}
	
	// analyze informations
	m_dwBlockSize = (DWORD) (boot.bytes_per_sector[1] * 256 + boot.bytes_per_sector[0]);
	m_dwBlocksCount = (DWORD) boot.n_sectors_l - 1; // TODO: why -1 ? (if not, error at the end of the volume)
	m_cHpfsVersion = super.version;
	memcpy(m_szLabel, boot.vol_label, 11);
	m_dwBitmapPointer = super.bitmaps;
	fprintf(m_fDebug, "m_dwBitmapPointer = %lu\n", m_dwBitmapPointer); fflush(m_fDebug);
	
	// check data
	if ((m_dwBlockSize % 512 != 0) || (m_dwBlockSize < 512))
		return -1;

	// read bitmap informations		
	fprintf(m_fDebug, "begin of copyBitmap()\n"); fflush(m_fDebug);
	copyBitmap();
	fprintf(m_fDebug, "end of copyBitmap()\n"); fflush(m_fDebug);

	return 0;
}

// =======================================================
int CHpfsPart::saveToImage(CImage *image, COptions *options, CSaveWindow *gui)
{
	image->m_dwCurrentVolumeSize = 0L;
	image->m_dwVolumeNumber = 0L;
	m_guiSave = gui;
	int nRes;	
	
	m_image = image;

	// ---- 0. Initializations
	
	// A. read informations
	newtPushHelpLine(i18n("reading partition properties"));
	fprintf(m_fDebug, "begin of readInformations()\n"); fflush(m_fDebug);
	nRes = readInformations();
	fprintf(m_fDebug, "end of readInformations()\n"); fflush(m_fDebug);
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read partition super block."));
		return -1;
	}	
	
	// B. print informations	
	newtPushHelpLine(i18n("reading partition informations"));
	printfInformations();
	
	// ---- 1. Write Hpfs header
	newtPushHelpLine(i18n("writing header"));
	CHpfsHeader headHpfs;
	memset(&headHpfs, 0, sizeof(CHpfsHeader));
	
	strcpy(headHpfs.szAppFormat, CURRENT_HPFS_FORMAT);
	
	headHpfs.dwBlockSize = m_dwBlockSize;
	headHpfs.dwBlocksCount = m_dwBlocksCount;
	headHpfs.cHpfsVersion = m_cHpfsVersion;
	headHpfs.dwBitmapPointer = m_dwBitmapPointer;
	memcpy(headHpfs.szLabel, m_szLabel, 11);
	headHpfs.dwBitmapQuadBlocksCount = m_dwBitmapQuadBlocksCount;
	headHpfs.dwUsedBlocks = m_dwUsedBlocks;
	
	errno = 0;
	nRes = m_image->write(&headHpfs, sizeof(CHpfsHeader), *options, true);
	if (nRes != sizeof(CHpfsHeader))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write hpfs header: (%s)"), strerror(errno));
		return -1;
	}

	// B. Copy bitmap to image file
	nRes = m_image->writeMagic(MAGIC_BEGIN_BITMAP, *options);
	if (nRes == -1)
		return -1;
	
	DWORD dwBitmapSize = (DWORD)(m_dwBitmapQuadBlocksCount * 4 * 512);
	nRes = m_image->write(m_bitmap.data(), dwBitmapSize, *options, true);
	if (nRes != dwBitmapSize)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write blocks bitmap"));
		return -1;
	}	

	// write magic string
	nRes = m_image->writeMagic(MAGIC_BEGIN_DATABLOCKS, *options);
	if (nRes == -1)
		return -1;

	// ---- 5. Write all used blocks
	newtPushHelpLine(i18n("copying used data blocks"));
	fprintf(m_fDebug, "begin of saveUsedBlocksToImageFile()\n"); fflush(m_fDebug);
	nRes = saveUsedBlocksToImageFile(options);
	fprintf(m_fDebug, "end of saveUsedBlocksToImageFile()\n"); fflush(m_fDebug);
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Unable to copy used blocks to the image file"));
		return -1;
	}
	
	return 0;
}

// =======================================================
int CHpfsPart::restoreFromImage(CImage *image, COptions *options, CRestoreWindow *gui)
{
	CHpfsHeader headHpfs;
	int nRes;

	// init
	m_image = image;
	m_guiRestore = gui;

	newtPushHelpLine(i18n("reading partition informations"));
	
	// 1. ---- Read Hpfs header	
	memset(&headHpfs, 0, sizeof(CHpfsHeader));
	nRes = m_image->read(&headHpfs, sizeof(CHpfsHeader), m_image->m_bIsSplitted, *options, true);
	if (nRes != sizeof(CHpfsHeader))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read hpfs header from the image file"));
		return -1;
	}
	
	// check the current format is supported
	if (strncmp(headHpfs.szAppFormat, CURRENT_HPFS_FORMAT, strlen(CURRENT_HPFS_FORMAT)) < 0)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("The partition was saved with a newer version of Partition Image. Please download the last version."));
		return -1;
	}	
	
	m_dwBlockSize = headHpfs.dwBlockSize;
	m_dwBlocksCount = headHpfs.dwBlocksCount;
	m_cHpfsVersion = headHpfs.cHpfsVersion ;
	m_dwBitmapPointer = headHpfs.dwBitmapPointer;
	memcpy(m_szLabel, headHpfs.szLabel, 11);
	m_dwBitmapQuadBlocksCount = headHpfs.dwBitmapQuadBlocksCount;
	m_dwUsedBlocks = headHpfs.dwUsedBlocks;

	
	// 2. ---- Print informations and ask confirmation to restore
	printfInformations();
	
	// ask confirmation
	nRes = newtWinChoice(i18n("Confirmation"), i18n("Yes"), i18n("No"), i18n("Do you really want to restore %s from the image file %s ?"), m_szDevice, image->m_szImageFilename);
	if (nRes == 2)
		return -1;
	
	// ---- 5. read bitmap
	nRes = m_image->readAndCheckMagic(MAGIC_BEGIN_BITMAP, *options);
	if (nRes == -1)
		return -1;
	
	// init bitmap size
	m_bitmap.init(m_dwBitmapQuadBlocksCount * 4 * 512 * 8);

	// copy bitmap blocks to bitmap in memory
	DWORD dwBitmapSize = (DWORD)(m_dwBitmapQuadBlocksCount * 4 * 512);
	nRes = m_image->read(m_bitmap.data(), (long)dwBitmapSize , m_image->m_bIsSplitted, *options, true);
	if (nRes != dwBitmapSize)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read bitmap from image file"));
		return -1;
	}
	
	// read and check magic string
	nRes = m_image->readAndCheckMagic(MAGIC_BEGIN_DATABLOCKS, *options);
	if (nRes == -1)
		return -1;
	
	// ---- 7. Restore all used blocks
	errno = 0;
	newtPushHelpLine(i18n("copying used data blocks"));
	fprintf(m_fDebug, "begin of restoreUsedBlocksToDisk()\n"); fflush(m_fDebug);  
	nRes = restoreUsedBlocksToDisk(options);
	fprintf(m_fDebug, "end of restoreUsedBlocksToDisk()\n"); fflush(m_fDebug);  
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Unable to restore used blocks to disk. (%s)"), strerror(errno));
		return -1;
	}

	return 0;
}

// =======================================================
int CHpfsPart::restoreUsedBlocksToDisk(COptions *options)
{
	int nRes;
	DWORD i;
	char cBuffer[1024];
	char cZeroBuffer[1024];
	
	DWORD dwSavedBlocks = 0L;
	memset(cZeroBuffer, 0, 1023);

	// go to the beginning of the disk
	nRes = fseek(m_fDeviceFile, 0L, SEEK_SET);
	if (nRes == -1)
	    return -1;
	
	for (i=0; i < m_dwBlocksCount; i++)
	{	
		if (m_bitmap.isBitSet(i)) // block used
		{
			dwSavedBlocks++;
				
		  nRes = m_image->read(cBuffer, m_dwBlockSize, m_image->m_bIsSplitted, *options, true);
			if (nRes != m_dwBlockSize)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read data block %ld from image file"), i);
				return -1;
			}	
				
			errno = 0;
			nRes = fwrite(cBuffer, m_dwBlockSize, 1, m_fDeviceFile);
			if (nRes != 1)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't restore data block %ld from image file. (%s) (%d)"), i, strerror(errno));
				return -1;
			}				
		}
		else // block free --> skip
		{
			if (options->bEraseWithNull == true) // if free and we must write 0 bytes on unused clusters
			{
				errno = 0;
				nRes = fwrite(cZeroBuffer, m_dwBlockSize, 1, m_fDeviceFile);
				if (nRes != 1)
				{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write zero block %lu (%s) (%d)"), i, strerror(errno));
					return -1;
				}				
				dwSavedBlocks++;
			}
			else // don't write zero bytes
			{
				nRes = fseek(m_fDeviceFile, m_dwBlockSize, SEEK_CUR);
				if (nRes == -1)
					return -1;	
			}			
		}
		
		// draw progress
		if (dwSavedBlocks%20 == 0)
			m_guiRestore->setProgressValue(dwSavedBlocks, m_dwUsedBlocks, m_dwBlocksCount, options->bEraseWithNull);
	}
		
	m_guiRestore->setProgressValue(dwSavedBlocks, m_dwUsedBlocks, m_dwBlocksCount, options->bEraseWithNull);
	
	return 0;
}

