/***************************************************************************
                          cntfspart.cpp  -  description
                             -------------------
    begin                : Mon May 22 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cntfspart.h"
#include "partimage.h"
#include "imagefile.h"
#include "misc.h"
#include "gui_text.h"

// =======================================================
CNtfsPart::CNtfsPart(char *szDevice, FILE *fDeviceFile)
{
	fprintf(m_fDebug, "begin of CNtfsPart::CNtfsPart()\n"); fflush(m_fDebug);  
	m_fDeviceFile = fDeviceFile;
	strncpy(m_szDevice, szDevice, MAX_DEVICENAMELEN);
	fprintf(m_fDebug, "end of CNtfsPart::CNtfsPart()\n"); fflush(m_fDebug);  
}

// =======================================================
CNtfsPart::~CNtfsPart()
{
}

// =======================================================
int CNtfsPart::readBootSector()
{
	int nRes;
	const int nMaxBootSectSize = 16384;
	BYTE cBoot[nMaxBootSectSize+1];

	nRes = fread(cBoot, nMaxBootSectSize, 1, m_fDeviceFile);
	if (nRes != 1)
		return -1;

	// 1. check partition has an ntfs file system
	if (memcmp(cBoot+3, "NTFS", 4) != 0)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("%s is not an NTFS file system partition"), m_szDevice);
		return -1;
	}
	
	/*
	// OLD METHOD
	memcpy(&m_nBytesPerSector, cBoot+11, 2);
	memcpy(&m_cSectorsPerCluster, cBoot+13, 1);
	memcpy(&m_qwTotalSectorsCount, cBoot+40, 8);
	memcpy(&m_qwLCNOfMftDataAttrib, cBoot+48, 8);
	memcpy(&m_dwFileRecordSize, cBoot+64, 4);
  */
	
	fprintf(m_fDebug, "\n======== begin NTFS BOOT INFOS ========\n"); fflush(m_fDebug);

  // NEW METHOD
	char cClustersPerMftRecord;
  	
  m_nBytesPerSector = NTFS_GETU16(cBoot+0xB);
	m_cSectorsPerCluster = NTFS_GETU8(cBoot+0xD);
	m_qwTotalSectorsCount = NTFS_GETU64(cBoot+0x28);
  m_qwLCNOfMftDataAttrib = NTFS_GETU64(cBoot+0x30);
	cClustersPerMftRecord = NTFS_GETS8(cBoot+0x40);

	// check informations validity
	if (m_nBytesPerSector % 512 != 0)
		return -1;
    
	// Calculate clusters and misc informations
	m_dwClusterSize = (DWORD) (m_nBytesPerSector * m_cSectorsPerCluster);
	m_dwFileRecordSize *= m_dwClusterSize;
	m_qwTotalClustersCount = (QWORD) (m_qwTotalSectorsCount / m_cSectorsPerCluster);

	if (cClustersPerMftRecord > 0)
		m_dwFileRecordSize = cClustersPerMftRecord * m_dwClusterSize;
	else
		m_dwFileRecordSize = 1 << (-cClustersPerMftRecord);

	fprintf(m_fDebug, "m_nBytesPerSector = %d\n", m_nBytesPerSector); fflush(m_fDebug);
	fprintf(m_fDebug, "m_cSectorsPerCluster = %d\n", m_cSectorsPerCluster); fflush(m_fDebug);
	fprintf(m_fDebug, "m_qwTotalSectorsCount = %llu\n", m_qwTotalSectorsCount); fflush(m_fDebug);
	fprintf(m_fDebug, "m_qwLCNOfMftDataAttrib = %llu\n", m_qwLCNOfMftDataAttrib); fflush(m_fDebug);
	fprintf(m_fDebug, "cClustersPerMftRecord = %d\n", cClustersPerMftRecord); fflush(m_fDebug);
	fprintf(m_fDebug, "Calculated m_dwFileRecordSize = %lu\n", m_dwFileRecordSize); fflush(m_fDebug);
	
	fprintf(m_fDebug, "======== end NTFS BOOT INFOS ========\n\n"); fflush(m_fDebug);
	
	// read label
	fprintf(m_fDebug, "begin of readVolumeLabel()\n"); fflush(m_fDebug);  
	nRes = readVolumeLabel();
	fprintf(m_fDebug, "end of readVolumeLabel(), return = %d\n", nRes); fflush(m_fDebug);
	
	return 0;
}

// =======================================================
void printfBuffer(BYTE *cBuffer, int nSize)
{
	DWORD i;
	printf ("\n=========================================\n");
	for (i=0; i < nSize; i++)
		if (cBuffer[i] > 33)
			printf ("%c", cBuffer[i]);
		else
			printf(".");
	    	
	printf ("\n=========================================\n");
}

// =======================================================
void printBuf(FILE *port, char *string, int len)
{
	char *szDest;
	szDest = new char[4096];
	memset(szDest,0,4095);
	int i;
	for (i=0; i < len; i++)
		szDest[i] = (string[i] < 32) ? '.' : string[i];
	fwrite(szDest, len, 1, port);
	delete [] szDest;
}

// =======================================================
int CNtfsPart::saveUsedBlocksToImageFile(COptions *options)
{
	int nRes;
	QWORD i;
	char cBuffer[MAX_NTFS_BLOCK_SIZE+1];
	
	QWORD qwSavedClusters = 0L;

	// go to the beginning of the disk
	nRes = fseek(m_fDeviceFile, 0L, SEEK_SET);
	if (nRes == -1)
	    return -1;
	
	for (i=0; i < m_qwTotalClustersCount; i++)
	{
		// if block is used...
		if (m_bitmap.isBitSet(i)) // not damaged and not free
		{
			qwSavedClusters++;
				
			errno = 0;
			nRes = fread(cBuffer, m_dwClusterSize, 1, m_fDeviceFile);
			if (nRes != 1)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read block %ld from image (%s)"), i, strerror(errno));
				return -1;
			}
				
			errno = 0;
			nRes = m_image->write(cBuffer, m_dwClusterSize, *options, true);
			if (nRes != m_dwClusterSize)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write block %ld to image (%s)"), i, strerror(errno));
				return -1;
			}
		}
		else // block free --> skip
		{
			nRes = fseek(m_fDeviceFile, m_dwClusterSize, SEEK_CUR);
			if (nRes == -1)
				return -1;	
		}
		
		// draw progress
		if (qwSavedClusters%20 == 0)
			m_guiSave->setProgressValue(qwSavedClusters, m_qwUsedClustersCount);
	}
		
	m_guiSave->setProgressValue(m_qwUsedClustersCount, m_qwUsedClustersCount);
	
	return 0;
}

// =======================================================
int CNtfsPart::restoreUsedBlocksToDisk(COptions *options)
{
	int nRes;
	QWORD i;
	char cBuffer[MAX_NTFS_BLOCK_SIZE+1];
	char cZeroBuffer[MAX_NTFS_BLOCK_SIZE+1];
	
	QWORD qwSavedClusters = 0L;
	memset(cZeroBuffer, 0, MAX_NTFS_BLOCK_SIZE);

	// go to the beginning of the disk
	nRes = fseek(m_fDeviceFile, 0L, SEEK_SET);
	if (nRes == -1)
	    return -1;
	
	for (i=0; i < m_qwTotalClustersCount; i++)
	{	
		if (m_bitmap.isBitSet(i)) // block used
		{
			qwSavedClusters++;
				
		  nRes = m_image->read(cBuffer, m_dwClusterSize, m_image->m_bIsSplitted, *options, true);
			if (nRes != m_dwClusterSize)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read data block %ld from image file"), i);
				return -1;
			}	
				
			errno = 0;
			nRes = fwrite(cBuffer, m_dwClusterSize, 1, m_fDeviceFile);
			if (nRes != 1)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't restore data block %ld from image file. (%s) (%d)"), i, strerror(errno));
				return -1;
			}				
		}
		else // block free --> skip
		{
			if (options->bEraseWithNull == true) // if free and we must write 0 bytes on unused clusters
			{
				errno = 0;
				nRes = fwrite(cZeroBuffer, m_dwClusterSize, 1, m_fDeviceFile);
				if (nRes != 1)
				{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write zero block %lu (%s) (%d)"), i, strerror(errno));
					return -1;
				}				
				qwSavedClusters++;
			}
			else // don't write zero bytes
			{
				nRes = fseek(m_fDeviceFile, m_dwClusterSize, SEEK_CUR);
				if (nRes == -1)
					return -1;	
			}			
		}
		
		// draw progress
		if (qwSavedClusters%20 == 0)
			m_guiRestore->setProgressValue(qwSavedClusters, m_qwUsedClustersCount, m_qwTotalClustersCount, options->bEraseWithNull);
	}
		
	m_guiRestore->setProgressValue(qwSavedClusters, m_qwUsedClustersCount, m_qwTotalClustersCount, options->bEraseWithNull);
	
	return 0;
}

// =======================================================
int CNtfsPart::readBitmap()
{
	int nRes;
	BYTE *cFileRecord; // 1024 in most cases
 	CNtfsRunList runlistBitmap, runlistMft;
 	QWORD qwBitmapSize;
 	QWORD qwAllocatedSize;
 	QWORD qwOffset;
 	QWORD i;

 	// used in experimental search
	char cBitmapNameUnicode[] = {0x24, 0, 0x42, 0, 0x69, 0, 0x74, 0, 0x6d, 0, 0x61, 0, 0x70, 0}; // len = 14

	// allocate memory
	cFileRecord = new BYTE[m_dwFileRecordSize+1];
	if (!cFileRecord)
	{	fprintf(m_fDebug, "Out of memory: need %lu bytes to read file record in readBitmap()\n", m_dwFileRecordSize); fflush(m_fDebug);
		newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Out of memory: need %lu bytes to read file record in readBitmap()"), m_dwFileRecordSize);
		return -1;
	}	
	
 	// 0. MFT Record: find $Bitmap file record
	/*qwOffset = ((QWORD) m_qwLCNOfMftDataAttrib * m_dwClusterSize);
	nRes = fseek(m_fDeviceFile, (long)qwOffset, SEEK_SET);
	if (nRes == -1)
		return -1;

	nRes = fread(cFileRecord, m_dwFileRecordSize, 1, m_fDeviceFile);
	if (nRes != 1)
		return -1;

	nRes = readFileRecord(cFileRecord, m_dwFileRecordSize, &runlistMft, &qwBitmapSize, &qwAllocatedSize);*/
	//runlistBitmap.show();
 	
	// read bitmap file record of the MFT (record number 6
	// TODO: bug if the begin of the data attrib of the MFT begins after 2 GB
	// TODO: bug if the MFT is fragmented, the $Bitmap record (6 th) may be in another location of the disk
	//       then we must read the MFT runlist, to know where the MFT is written. But we need to read the
	//       "attribute list" to do it... ==> very difficult
	
	// 1. read $Bitmap record
	qwOffset = ((QWORD) m_qwLCNOfMftDataAttrib * m_dwClusterSize) + ((QWORD) (6 * m_dwFileRecordSize));
	fprintf(m_fDebug, "qwOffset = %llu\n", qwOffset); fflush(m_fDebug);
	nRes = fseek(m_fDeviceFile, (long)qwOffset, SEEK_SET);
	if (nRes == -1)
	{	fprintf(m_fDebug, "readBitmap(): failed: cannot run fseek(to bitmap record)\n"); fflush(m_fDebug);
		delete []cFileRecord;
		return -1;
	}

	nRes = fread(cFileRecord, m_dwFileRecordSize, 1, m_fDeviceFile);
	if (nRes != 1)
	{	fprintf(m_fDebug, "readBitmap(): failed: cannot run fread(bitmap record)\n"); fflush(m_fDebug);
		delete []cFileRecord;
		return -1;
	}

	// check we read the $bitmap file record
	fprintf(m_fDebug, "begin of checkFilenameForRecordIs()\n"); fflush(m_fDebug);
	nRes = checkFilenameForRecordIs(cFileRecord, "$bitmap");
	fprintf(m_fDebug, "end of checkFilenameForRecordIs(), return = %d\n", nRes); fflush(m_fDebug);
	
	if (nRes == -1) // The file record we read was not the $bitmap one...
	{
		nRes = newtWinChoice(i18n("Error"), i18n("Yes"), i18n("No"), i18n("Unable to read the $Bitmap MFT record. "
		"Do you want to make an experimental search for it ? (it can take a long time)"));
		if (nRes == 2) // if "No"
		{	fprintf(m_fDebug, "readBitmap(): user doesn't want to search for $bitmap MFT record\n"); fflush(m_fDebug);
			delete []cFileRecord;
			return -1;
		}
			
		fprintf(m_fDebug, "begin of searchForRecordFilename()\n"); fflush(m_fDebug);
		nRes = searchForRecordFilename(cFileRecord, "$bitmap", cBitmapNameUnicode, 14);
		fprintf(m_fDebug, "end of searchForRecordFilename(), result = %d\n", nRes); fflush(m_fDebug);
		if (nRes == -1) // if found
		{	fprintf(m_fDebug, "readBitmap(): failed: Cannot find the $bitmap with experimental search\n"); fflush(m_fDebug);
			newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Cannot find the $bitmap with experimental search"));
			delete []cFileRecord;
			return -1;	
		}
	}
	
	// -------- decode the MFT File record
	fprintf(m_fDebug, "begin of readFileRecord()\n"); fflush(m_fDebug);  
	nRes = readFileRecord(cFileRecord, &runlistBitmap, &qwBitmapSize, &qwAllocatedSize);
	fprintf(m_fDebug, "end of readFileRecord(), return = %d\n", nRes); fflush(m_fDebug);
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Cannot use $bitmap file record"));
		fprintf(m_fDebug, "readBitmap(): failed: Cannot use $bitmap file record\n"); fflush(m_fDebug);	
	}
	
	//runlistBitmap.show();

	// If the bitmap returned cannot map all clusters of the partition (because bitmap too small)
	if ((runlistBitmap.clustersCount() * m_dwClusterSize * 8) < m_qwTotalClustersCount)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("NTFS $Bitmap is to small"));
		fprintf(m_fDebug, "readBitmap(): NTFS $Bitmap is to small\n"); fflush(m_fDebug);
		delete []cFileRecord;
		return -1;
	}
	
	// -------- read bitmap file from runlist
	m_dwBitmapSize = (DWORD) qwAllocatedSize;
	//fprintf(stderr,"m_dwBitmapSize=%lu\n", m_dwBitmapSize);
	//fprintf(stderr,"m_dwallocated=%llu\n", qwAllocatedSize);

	// allocate memory for bitmap
	nRes = m_bitmap.init(m_dwBitmapSize * 8);
	if (nRes == -1)
	{	fprintf(m_fDebug, "readBitmap(): failed: Out of memory in bitmap.init()\n"); fflush(m_fDebug);	
		newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Out of memory"));
		delete []cFileRecord;
		return -1;
	}

	// debug log file
	fprintf(m_fDebug, "readBitmap(): =========== begin of decoded run list ===============\n"); fflush(m_fDebug);
	for (i=0L; i < runlistBitmap.clustersCount(); i++)
		fprintf(m_fDebug, "i=%llu, cluster=%llu, offset=%llu\n", i, (runlistBitmap.offset(i)), (runlistBitmap.offset(i) * m_dwClusterSize));
	fprintf(m_fDebug, "readBitmap(): =========== end of decoded run list ===============\n"); fflush(m_fDebug);
	
	// read bitmap
	for (i=0L; i < runlistBitmap.clustersCount(); i++)
	{
		errno = 0;
		fprintf(m_fDebug, "begin of big_fseek()\n"); fflush(m_fDebug);	
		nRes = bigFseek(m_fDeviceFile, (QWORD)(runlistBitmap.offset(i) * ((QWORD)m_dwClusterSize)));
		fprintf(m_fDebug, "end of big_fseek() (%s) (error %d)\n", strerror(errno), errno); fflush(m_fDebug);	
		if (nRes == -1)
		{	fprintf(m_fDebug, "readBitmap(): failed: error in fseek() (%s) (error %d)\n", strerror(errno), errno); fflush(m_fDebug);	
			delete []cFileRecord;
			return -1;
		}

		errno = 0;
		nRes = fread(m_bitmap.data()+(i*m_dwClusterSize), m_dwClusterSize, 1, m_fDeviceFile);
		if (nRes != 1)
		{	fprintf(m_fDebug, "readBitmap(): failed: error in fread() (%s) (error %d)\n", strerror(errno), errno); fflush(m_fDebug);	
			delete []cFileRecord;
			return -1;
		}
	}

	m_qwUsedClustersCount = 0L;
	m_qwFreeClustersCount = 0L;
	
	// fill informations about free/used clusters count
	for (i=0; i < m_qwTotalClustersCount; i++)
	{
		if (m_bitmap.isBitSet(i))
			m_qwUsedClustersCount++;
		else
			m_qwFreeClustersCount++;
	}
	
	delete []cFileRecord;
	return 0;
}

// =======================================================
int CNtfsPart::saveToImage(CImage *image, COptions *options, CSaveWindow *gui)
{
	int nRes;

	image->m_dwCurrentVolumeSize = 0L;
	image->m_dwVolumeNumber = 0L;

	m_image = image;
	m_guiSave = gui;

	// 0. read informations using the boot sector (copy to m_ variables)
	newtPushHelpLine(i18n("reading the boot sector"));
	fprintf(m_fDebug, "begin of readBootSector()\n"); fflush(m_fDebug);  
	nRes = readBootSector();
	fprintf(m_fDebug, "end of readBootSector(), return = %d\n", nRes); fflush(m_fDebug);
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read boot sector"));
		return -1;
	}

	// 1. check the file system
	if (options->bCheckBeforeSaving == true)
	{
		newtPushHelpLine(i18n("checking the file system"));
		// TODO: check for NT file system errors
	}

	// ---- 2. read the bitmap
	newtPushHelpLine(i18n("reading NTFS bitmap file"));
	fprintf(m_fDebug, "begin of readBitmap()\n"); fflush(m_fDebug);
	nRes = readBitmap();
	fprintf(m_fDebug, "end of readBitmap(), return = %d\n", nRes); fflush(m_fDebug);
	if (nRes == -1)
	{
		newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read the bitmap file"));
		return -1;
	}

	// ---- 3. Write header
	CNtfsHeader headNtfs;
	memset(&headNtfs, 0, sizeof(CNtfsHeader));

	strcpy(headNtfs.szAppFormat, CURRENT_NTFS_FORMAT);
	strcpy(headNtfs.szLabel, m_szLabel);
	headNtfs.nBytesPerSector = m_nBytesPerSector;
	headNtfs.cSectorsPerCluster = m_cSectorsPerCluster;
	headNtfs.qwTotalSectorsCount = m_qwTotalSectorsCount;
	headNtfs.qwLCNOfMftDataAttrib = m_qwLCNOfMftDataAttrib;
	headNtfs.dwFileRecordSize = m_dwFileRecordSize;
	headNtfs.dwBitmapSize = m_dwBitmapSize;
	headNtfs.qwTotalClustersCount = m_qwTotalClustersCount;
	headNtfs.qwUsedClustersCount = m_qwUsedClustersCount;
	headNtfs.qwFreeClustersCount = m_qwFreeClustersCount;
	headNtfs.dwClusterSize = m_dwClusterSize;
	headNtfs.nNtfsVersion = m_nNtfsVersion;
	
	// write header to image file
	nRes = m_image->write(&headNtfs, sizeof(CNtfsHeader), *options, true);
	if (nRes != sizeof(CNtfsHeader))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write ntfs header"));
		return -1;
	}

	// read and print important informations
	printfInformations();

	// ---- 5. copy bitmap
	nRes = m_image->writeMagic(MAGIC_BEGIN_BITMAP, *options);
	if (nRes == -1)
		return -1;
	
	nRes = m_image->write(m_bitmap.data(), m_dwBitmapSize, *options, true);
	if (nRes != m_dwBitmapSize)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write bitmap"));
		return -1;
	}
	
	// ---- 6. Write all used clusters
	newtPushHelpLine(i18n("copying used data blocks"));
	
	nRes = m_image->writeMagic(MAGIC_BEGIN_DATABLOCKS, *options);
	if (nRes == -1)
		return -1;

	fprintf(m_fDebug, "begin of saveUsedBlocksToImageFile()\n"); fflush(m_fDebug);  
	nRes = saveUsedBlocksToImageFile(options);
	fprintf(m_fDebug, "end of saveUsedBlocksToImageFile(), return = %d\n", nRes); fflush(m_fDebug);
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Unable to copy used blocks to the image file"));
		return -1;
	}
	
	return 0;
}

// =======================================================
int CNtfsPart::restoreFromImage(CImage *image, COptions *options, CRestoreWindow *gui)
{
	CNtfsHeader headNtfs;
	int nRes;

	// init
	m_image = image;
	m_guiRestore = gui;

	newtPushHelpLine(i18n("reading partition informations"));

	// 1. ---- Read NTFS header	
	memset(&headNtfs, 0, sizeof(CNtfsHeader));
	nRes = m_image->read(&headNtfs, sizeof(CNtfsHeader), m_image->m_bIsSplitted, *options, true);
	if (nRes != sizeof(CNtfsHeader))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read ntfs header from the image file"));
		return -1;
	}
	
	// check the current format is supported
	if (strncmp(headNtfs.szAppFormat, CURRENT_NTFS_FORMAT, strlen(CURRENT_NTFS_FORMAT)) < 0)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("The partition was saved with a newer version of Partition Image. Please download the last version."));
		return -1;
	}	
	
	strcpy(m_szLabel, headNtfs.szLabel);
	m_nBytesPerSector = headNtfs.nBytesPerSector;
	m_cSectorsPerCluster = headNtfs.cSectorsPerCluster;
	m_qwTotalSectorsCount = headNtfs.qwTotalSectorsCount;
	m_qwLCNOfMftDataAttrib = headNtfs.qwLCNOfMftDataAttrib;
	m_dwFileRecordSize = headNtfs.dwFileRecordSize;
	m_dwBitmapSize = headNtfs.dwBitmapSize;
	m_qwTotalClustersCount = headNtfs.qwTotalClustersCount;
	m_qwUsedClustersCount = headNtfs.qwUsedClustersCount;
	m_qwFreeClustersCount = headNtfs.qwFreeClustersCount;
	m_dwClusterSize = headNtfs.dwClusterSize;
	m_nNtfsVersion = headNtfs.nNtfsVersion;

	
	// 2. ---- Print informations and ask confirmation to restore
	printfInformations();
	
	// ask confirmation
	nRes = newtWinChoice(i18n("Confirmation"), i18n("Yes"), i18n("No"), i18n("Do you really want to restore %s from the image file %s ?"), m_szDevice, image->m_szImageFilename);
	if (nRes == 2)
		return -1;
	
	// ---- 5. read bitmap
	newtPushHelpLine(i18n("reading bitmap..."));

	nRes = m_image->readAndCheckMagic(MAGIC_BEGIN_BITMAP, *options);
	if (nRes == -1)
		return -1;
	
	// init bitmap size
	m_bitmap.init(m_dwBitmapSize * 8);
	
	// copy bitmap blocks to bitmap in memory
	nRes = m_image->read(m_bitmap.data(), m_dwBitmapSize, m_image->m_bIsSplitted, *options, true);
	if (nRes != m_dwBitmapSize)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read bitmap from image file"));
		return -1;
	}

	// read and check magic string
	nRes = m_image->readAndCheckMagic(MAGIC_BEGIN_DATABLOCKS, *options);
	if (nRes == -1)
		return -1;
	
	// ---- 7. Restore all used blocks
	newtPushHelpLine(i18n("copying used data blocks"));
	errno = 0;
	fprintf(m_fDebug, "begin of restoreUsedBlocksToDisk()\n"); fflush(m_fDebug);  
	nRes = restoreUsedBlocksToDisk(options);
	fprintf(m_fDebug, "end of restoreUsedBlocksToDisk(), return = %d\n", nRes); fflush(m_fDebug);
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Unable to restore used blocks to disk. (%s)"), strerror(errno));
		return -1;
	}

	return 0;
}
// =======================================================
int CNtfsPart::printfInformations()
{
	char szTemp[2048];
	char szTemp2[2048];
	char szText[8192];
	
	sprintf(szText,  i18n("Total sectors count...........%llu\n"), m_qwTotalSectorsCount);
	sprintf(szTemp2, i18n("Total clusters count..........%llu\n"), m_qwTotalClustersCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Used clusters count...........%llu\n"), m_qwUsedClustersCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Free clusters count...........%llu\n"), m_qwFreeClustersCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Bytes per sector..............%u\n"), m_nBytesPerSector);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Sectors per cluster...........%u\n"), m_cSectorsPerCluster);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Bitmap size...................%s (%lu bytes)\n"), formatSize(m_dwBitmapSize, szTemp), m_dwBitmapSize);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("File record size..............%lu\n"), m_dwFileRecordSize);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Space usage:..................%lu %%\n"), (DWORD)(100 * m_qwUsedClustersCount)/m_qwTotalClustersCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Used space....................%s\n"), formatSize(m_qwUsedClustersCount * m_nBytesPerSector * m_cSectorsPerCluster, szTemp));
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Free space....................%s\n"), formatSize(m_qwFreeClustersCount * m_nBytesPerSector * m_cSectorsPerCluster, szTemp));
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Label:........................%s\n"), m_szLabel);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("LCN of MFT Data attrib........%llu\n"), m_qwLCNOfMftDataAttrib);
	strcat(szText, szTemp2);
	if (m_nNtfsVersion >= 0x0300)
	{	sprintf(szTemp2, i18n("NTFS version is Win2000 or greater\n"));
		strcat(szText, szTemp2);
	}

	fprintf(m_fDebug, "\n====================\n%s====================\n\n", szText); fflush(m_fDebug);
	newtWinMessage(i18n("NTFS informations"), i18n("Ok"), szText);
						
	return 0;
}

// =======================================================
int CNtfsPart::readVolumeLabel()
{
	int nRes;
	BYTE *cFileRecord; // 1024 in most cases
	QWORD qwOffset;
	UINT nOffsetSequenceAttribute;
	BYTE *cData;
	DWORD dwAttribType;
	DWORD dwAttribLen;
	bool bAttribResident;	
	BYTE *cDataResident;
	UINT nAttrSize;
 	QWORD i;

 	// init
 	memset(m_szLabel, 0, 127);
 	
	// allocate memory
	cFileRecord = new BYTE[m_dwFileRecordSize+1];
	if (!cFileRecord)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Out of memory: need %lu bytes to read file record in readVolumeLabel()"), m_dwFileRecordSize);
		return -1;
	}	
	
 	// TODO: read the MFT to have the $volume record location
 	
	// 1. read $Volume record
	qwOffset = ((QWORD) m_qwLCNOfMftDataAttrib * m_dwClusterSize) + ((QWORD) (3 * m_dwFileRecordSize));
	nRes = fseek(m_fDeviceFile, (long)qwOffset, SEEK_SET);
	if (nRes == -1)
	{	fprintf(m_fDebug, "readVolumeLabel(): fseek() failed\n"); fflush(m_fDebug);
		delete []cFileRecord;
		return -1;
	}

	nRes = fread(cFileRecord, m_dwFileRecordSize, 1, m_fDeviceFile);
	if (nRes != 1)
	{	fprintf(m_fDebug, "readVolumeLabel(): fread() failed\n"); fflush(m_fDebug);
		delete []cFileRecord;
		return -1;
	}

	// -------- decode the MFT File record
	nOffsetSequenceAttribute = NTFS_GETU16(cFileRecord+0x14);
	cData = cFileRecord + nOffsetSequenceAttribute;
	
	do
	{
		// szData points to the beginning of an attribute
		dwAttribType = NTFS_GETU32(cData);
		dwAttribLen = NTFS_GETU32(cData+4);
		bAttribResident = (NTFS_GETU8(cData+8)==0);
	
		if(dwAttribType == 0x60) // "volume_name"
		{
			if (bAttribResident == false)
			{	fprintf(m_fDebug, "readVolumeLabel(): failed: $volume_name attribute is not resident\n"); fflush(m_fDebug);
				delete []cFileRecord;
				return -1;
			}
				
			nAttrSize = NTFS_GETU16(cData+0x10);
			cDataResident = cData+NTFS_GETU16(cData+0x14);
	
			for (i=0; i < nAttrSize/2; i++)
				m_szLabel[i] = cDataResident[2*i];
		}
		if(dwAttribType == 0x70) // "volume_information"
		{	m_nNtfsVersion = 0;
			if (bAttribResident == true)
			{	
				nAttrSize = NTFS_GETU16(cData+0x10);
				cDataResident = cData+NTFS_GETU16(cData+0x14);
				m_nNtfsVersion = (((BYTE) cDataResident[8]) << 8) | ((BYTE) cDataResident[9]);
				fprintf(m_fDebug, "readVolumeLabel(): version is %d.%d\n", cDataResident[8], cDataResident[9]); fflush(m_fDebug);
			}
		}
		
		cData += dwAttribLen;
	} while(dwAttribType != -1); // attribute list ends with type -1
	
	delete []cFileRecord;
	return 0;
}
			
// =======================================================
int CNtfsPart::readFileRecord(BYTE *cRecordData, CNtfsRunList *runlist, QWORD *qwBitmapSize, QWORD *qwAllocatedSize)
{
	BYTE *cData;
	UINT nOffsetSequenceAttribute;
	DWORD dwAttribType;
	DWORD dwAttribLen;
	bool bAttribResident;
	int nRes;
	
	nOffsetSequenceAttribute = NTFS_GETU16(cRecordData+0x14);
	cData = cRecordData + nOffsetSequenceAttribute;
	
	fprintf(m_fDebug, "\n\n=============== begin FILE RECORD ================\n"); fflush(m_fDebug);
	fwrite(cRecordData, m_dwFileRecordSize, 1, m_fDebug);
	fprintf(m_fDebug, "\n=============== end FILE RECORD ================\n\n\n"); fflush(m_fDebug);
	
	do
	{
		// szData points to the beginning of an attribute
		dwAttribType = NTFS_GETU32(cData);
		dwAttribLen = NTFS_GETU32(cData+4);
		bAttribResident = (NTFS_GETU8(cData+8)==0);
	
		if(dwAttribType == 0x20) // "attribute list"
		{
			newtWinMessage(i18n("Error"), i18n("Ok"), i18n("NTFS: \"Attribute List\" attribute is not yet supported."));
			fprintf(m_fDebug, "readFileRecord(): NTFS: \"Attribute List\" attribute is not yet supported.\n"); fflush(m_fDebug);
			return -1;
		}
		else if(dwAttribType == 0x80) // "data"
		{
			fprintf(m_fDebug, "begin of readDataAttribute()\n"); fflush(m_fDebug);
			nRes = readDataAttribute(cData, runlist, qwBitmapSize, qwAllocatedSize);
			fprintf(m_fDebug, "end of readDataAttribute(), return = %d\n", nRes); fflush(m_fDebug);
			if (nRes == -1)
				return -1;
			
		}
		cData += dwAttribLen;
	} while(dwAttribType != -1); // attribute list ends with type -1

	return 0;
}

// =======================================================
int CNtfsPart::readDataAttribute(BYTE *cAttribData, CNtfsRunList *runlist, QWORD *qwBitmapSize, QWORD *qwAllocatedSize)
{
	bool bAttribResident;
	UINT nCompressed;
	int nRes;
	
	// Resident flag
	bAttribResident = NTFS_GETU8(cAttribData+8)==0;
	
	// Misc
	nCompressed = NTFS_GETU16(cAttribData+0xC);
	//fprintf(stderr,"Comp=%ld=%X with type=data\n", nCompressed, nCompressed);
	if (nCompressed)
	{
		newtWinMessage(i18n("Error"), i18n("Ok"), i18n("NTFS: Compression flag for DATA attribute not null: %ld"), nCompressed);
		fprintf(m_fDebug, "readDataAttribute(): NTFS: Compression flag for DATA attribute not null: %ld\n", nCompressed); fflush(m_fDebug);
		return -1;
	}

	if(bAttribResident)
	{
		// TODO: decode resident data
		UINT nAttrSize;
		BYTE *cDataResident;
		nAttrSize = NTFS_GETU16(cAttribData+0x10);
		cDataResident = cAttribData+NTFS_GETU16(cAttribData+0x14);
		/*UINT nOffsetOfStream;
		nOffsetOfStream=NTFS_GETU16(cAttribData+0x14);
		qwDataSize = (QWORD) NTFS_GETU16(cAttribData+0x10);*/
		newtWinMessage(i18n("Error"), i18n("Ok"), i18n("NTFS: Resident data attribute not supported"));
		fprintf(m_fDebug, "readDataAttribute(): NTFS: Resident data attribute not supported\n"); fflush(m_fDebug);
		return -1;
	}
	else
	{
		*qwAllocatedSize = NTFS_GETU32(cAttribData+0x28);
		*qwBitmapSize = NTFS_GETU32(cAttribData+0x30);
				
		fprintf(m_fDebug, "begin of processRunList()\n"); fflush(m_fDebug);  
		nRes = processRunList(cAttribData, runlist);
		fprintf(m_fDebug, "end of processRunList(), return = %d\n", nRes); fflush(m_fDebug);
		if (nRes == -1)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("NTFS: processRunList() returned -1"));
		}
		return nRes;
	}
		
	return -1;
}

// =======================================================
int CNtfsPart::readFilenameAttribute(BYTE *cAttribData, char *szFilename)
{
	bool bAttribResident;
	char szTemp[4096];
	UINT nAttrSize;
	BYTE *cDataResident;
	int i;
	
	// Resident flag
	bAttribResident = NTFS_GETU8(cAttribData+8)==0;
	if (bAttribResident == false)
	{	newtWinMessage(i18n("Error in readFilenameAttribute"), i18n("Ok"), i18n("The \"filename\" attribute must be resident !"));	
		fprintf(m_fDebug, "readFilenameAttribute(): NTFS: The \"filename\" attribute must be resident !\n"); fflush(m_fDebug);
		return -1;
	}
	
	nAttrSize = NTFS_GETU16(cAttribData+0x10)-0x42;
	cDataResident = cAttribData+NTFS_GETU16(cAttribData+0x14)+0x42;
		
	// decode the unicode filename
	memset(szTemp, 0, 4095);
	for (i=0; i < nAttrSize/2; i++)
		szTemp[i] = cDataResident[2*i];
	
	/*fprintf(stderr,"==============> Encoded file name: (");
	printBuf(stderr, (char*)cDataResident, 2*nAttrSize);
	fprintf(stderr,")\n==============> Decoded file name: (%s)\n", szTemp);*/
	
	// szFilename must be equal to szTemp now
	if (strcasecmp(szTemp, szFilename) != 0)
	{	//newtWinMessage(i18n("Error in readFilenameAttribute"), i18n("Ok"), i18n("Cannot check filename is \"%s\" (%s)"), szFilename, szTemp);
		fprintf(m_fDebug, "readFilenameAttribute(): Cannot check filename is \"%s\" (%s)\n", szFilename, szTemp); fflush(m_fDebug);	
		return -1;
	}
	else
	{
		fprintf(m_fDebug, "readFilenameAttribute(): success (%s)\n", szFilename); fflush(m_fDebug);		
	}
		
	return 0; // success
}

// =======================================================
int CNtfsPart::processRunList(BYTE *cAttribData, CNtfsRunList *runlist)
{			
	QWORD qwAllocatedSize;
	QWORD qwRealSize;
	QWORD qwCompressedSize;
	UINT nCompressionEngine;
	UINT nOffsetRunList;
	BYTE *cRunList;
	QWORD qwStartVcn;
	QWORD qwEndVcn;
	QWORD qwLastClusterOffset;
	QWORD qwClusterNumber;
		
	fprintf(m_fDebug, "\n\n\n================== begin processRunList =============\n"); fflush(m_fDebug);
	
	qwStartVcn = NTFS_GETU64(cAttribData+0x10);	
	qwEndVcn = NTFS_GETU64(cAttribData+0x18);
	qwAllocatedSize = NTFS_GETU32(cAttribData+0x28);	
	qwRealSize = NTFS_GETU32(cAttribData+0x30);
	qwCompressedSize = NTFS_GETU32(cAttribData+0x38);	
	nCompressionEngine = NTFS_GETU16(cAttribData+0x22);
	
	fprintf(m_fDebug, "qwStartVcn = %llu\n", qwStartVcn); fflush(m_fDebug);
	fprintf(m_fDebug, "qwEndVcn = %llu\n", qwEndVcn); fflush(m_fDebug);
	fprintf(m_fDebug, "qwAllocatedSize = %llu\n", qwAllocatedSize); fflush(m_fDebug);
	fprintf(m_fDebug, "qwRealSize = %llu\n", qwRealSize); fflush(m_fDebug);
	fprintf(m_fDebug, "qwCompressedSize = %llu\n", qwCompressedSize); fflush(m_fDebug);
	fprintf(m_fDebug, "nCompressionEngine = %d\n", nCompressionEngine); fflush(m_fDebug);
			
	//if(attr->compressed)
	//	attr->compsize=NTFS_GETU32(cAttribData+0x40);
		
	nOffsetRunList = NTFS_GETU16(cAttribData+0x20);
	cRunList = cAttribData + nOffsetRunList;
			
	BYTE cType;
	BYTE cSizeOfOffset;
	BYTE cSizeOfLength;
	BYTE *cData;
	QWORD qwLength;
	long long signed int llOffset;
	QWORD i;
	int nRes;
		
	nRes = runlist->init(qwEndVcn-qwStartVcn+1); // number of clusters used by data
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Out of memory"));
		fprintf(m_fDebug, "processRunList(): Out of memory.\n"); fflush(m_fDebug);
		return -1;
	}
			
	qwClusterNumber = qwStartVcn;		
	qwLastClusterOffset = 0L;
	while (qwClusterNumber < qwEndVcn)
	{
		cType = NTFS_GETU8(cRunList);
		cSizeOfLength = cType & 0xF;
		cSizeOfOffset = cType & 0xF0;
		cData = cRunList + 1;
		
		// 1: length
		switch(cSizeOfLength)
		{
			case 1: qwLength = NTFS_GETU8(cData); break;
			case 2: qwLength = NTFS_GETU16(cData); break;
			case 3: qwLength = NTFS_GETU24(cData); break;
      case 4: qwLength = NTFS_GETU32(cData); break;
      // Note: cases 5-8 are probably pointless to code,
      // since how many runs > 4GB of length are there?
      // at the most, cases 5 and 6 are probably necessary,
      // and would also require making length 64-bit
      // throughout
			default:
				//ntfs_error("Can't decode run type field %x\n",type);
				return -1;
		}
		cData += cSizeOfLength;
				
		// 2: offset
		switch(cSizeOfOffset)
		{
			//case 0:	   *ctype=2; break;
			case 0x10: llOffset = NTFS_GETS8(cData);break;
			case 0x20: llOffset = NTFS_GETS16(cData);break;
			case 0x30: llOffset = NTFS_GETS24(cData);break;
			case 0x40: llOffset = NTFS_GETS32(cData);break;
		#if 0 // Keep for future, in case ntfs_cluster_t ever becomes 64bit
			case 0x50: llOffset = NTFS_GETS40(cData);break;
			case 0x60: llOffset = NTFS_GETS48(cData);break;
			case 0x70: llOffset = NTFS_GETS56(cData);break;
			case 0x80: llOffset = NTFS_GETS64(cData);break;	
		#endif
			default:
				//ntfs_error("Can't decode run type field %x\n",type);
				return -1;
		}
		cData += cSizeOfOffset;
		
		for (i = 0; i < qwLength; i++)
		{	runlist->setCluster(qwClusterNumber, qwLastClusterOffset + llOffset + i);
			qwClusterNumber++;
		}
			
		qwLastClusterOffset += llOffset;
	}

	fprintf(m_fDebug, "\n================== end processRunList =============\n\n\n"); fflush(m_fDebug);

	return 0;
}

// ================================================
CNtfsRunList::CNtfsRunList()
{
	m_qwOffset = NULL; // no memory allocated
	m_qwClustersCount = 0L;
}

// ================================================
CNtfsRunList::~CNtfsRunList()
{
	freeMemory();
}

// ================================================
void CNtfsRunList::freeMemory()
{
	if (m_qwOffset)
	{
		delete []m_qwOffset;
		m_qwClustersCount = 0L;
	}
}

// ================================================
int CNtfsRunList::init(QWORD qwClustersCount)
{
	if (m_qwOffset)
		return -1; // mem alreay allocated
	
	m_qwOffset = new QWORD[qwClustersCount+1];
	if (!m_qwOffset)
		return -1;
		
	m_qwClustersCount = qwClustersCount;
	
	return 0; // success
}

// ================================================
int CNtfsRunList::setCluster(QWORD qwNumber, QWORD qwOffset) // ex: cluser 0 position is 12589
{
	if ((qwOffset) && (qwNumber <= m_qwClustersCount))
	{
		m_qwOffset[qwNumber] = qwOffset;
		return 0;
	}
	else
		return -1;
}

// ================================================
void CNtfsRunList::show()
{
	QWORD i;
	
	for (i=0L; i < m_qwClustersCount; i++)
		fprintf(stderr,"cluster[%llu] = %llu\n", i, m_qwOffset[i]);

}

// ================================================
QWORD CNtfsRunList::offset(QWORD qwCluster)
{
	if (m_qwOffset)
		return m_qwOffset[qwCluster];
	else
		return 0;
}

// ================================================
QWORD CNtfsRunList::clustersCount()
{
	return m_qwClustersCount;
}

// ================================================
int CNtfsPart::checkFilenameForRecordIs(BYTE *cRecordData, char *szFilename, bool bShowDebug)
{
	BYTE *cData;
	UINT nOffsetSequenceAttribute;
	DWORD dwAttribType;
	DWORD dwAttribLen;
	bool bAttribResident;
	int nRes;
	
	nOffsetSequenceAttribute = NTFS_GETU16(cRecordData+0x14);
	cData = cRecordData + nOffsetSequenceAttribute;
	
	// check this is a file record
	if (memcmp(cRecordData, "FILE", 4) != 0)
	{	if (bShowDebug)
			fprintf(m_fDebug, "checkFilenameForRecordIs(): error im magic string, not a valid file record\n"); fflush(m_fDebug);
		return -1;
	}	
	
	do
	{
		// szData points to the beginning of an attribute
		dwAttribType = NTFS_GETU32(cData);
		dwAttribLen = NTFS_GETU32(cData+4);
		bAttribResident = (NTFS_GETU8(cData+8)==0);
	
		if(dwAttribType == 0x30) // "filename"
		{
			fprintf(m_fDebug, "begin of readFilenameAttribute()\n"); fflush(m_fDebug);
			nRes = readFilenameAttribute(cData, szFilename);
			fprintf(m_fDebug, "end of readFilenameAttribute(), return = %d\n", nRes); fflush(m_fDebug);
			return nRes;
		}
		cData += dwAttribLen;
			
	} while(dwAttribType != -1); // attribute list ends with type -1

	return -1;
}

// ================================================
int CNtfsPart::searchForRecordFilename(BYTE *cFileRecord, char *szFilename, char *szUnicodeName, int nUnicodeNameLen)
{
	QWORD qwCurrentOffset = 0L;
	QWORD qwPartSize;
	char cBuffer[MAX_NTFS_BLOCK_SIZE+1];
	int nRes;
	int i;
	
	qwPartSize = getPartitionSize(m_szDevice);
	
	// go to the beginning of the partition
	nRes = fseek(m_fDeviceFile, 0L, SEEK_SET);
	if (nRes == -1)
	    return -1;
	
	while(qwCurrentOffset + m_dwClusterSize < qwPartSize)
	{
		nRes = fread(cBuffer, m_dwClusterSize, 1, m_fDeviceFile);
		if (nRes != 1)
		{	fprintf(m_fDebug, "searchForRecordFilename(): error in fread(1)\n"); fflush(m_fDebug);
			return -1;
		}	
		
		qwCurrentOffset += m_dwClusterSize;
			
		// if it can be the begin of the file record
		if (memcmp(cBuffer, "FILE", 4) == 0)
		{
			fprintf(m_fDebug, "searchForRecordFilename: begin FILE detected\n"); fflush(m_fDebug);	
		
			nRes = fseek(m_fDeviceFile, -m_dwClusterSize, SEEK_CUR);
			if (nRes == -1)
			{	fprintf(m_fDebug, "searchForRecordFilename(): error in fseek()\n"); fflush(m_fDebug);
				return -1;
			}	
		
			fprintf(m_fDebug, "searchForRecordFilename: 002 FILE detected\n"); fflush(m_fDebug);	
			
			qwCurrentOffset -= m_dwClusterSize;
			
			nRes = fread(cFileRecord, m_dwFileRecordSize, 1, m_fDeviceFile);
			if (nRes != 1)
			{	fprintf(m_fDebug, "searchForRecordFilename(): error in fread(2)\n"); fflush(m_fDebug);
				return -1;
			}	
			
			fprintf(m_fDebug, "searchForRecordFilename: 003 FILE detected\n"); fflush(m_fDebug);	
			
			qwCurrentOffset += m_dwFileRecordSize;

			// the "FILE" we read can be the magic string of the file record, but it can be a simple
			// data of the user. Then, check we can find what we search, before to decode the record
			
			// check the record contains the unicode name of the record we search
			for (i=0; i < m_dwFileRecordSize - nUnicodeNameLen; i++)
			{
				if (memcmp(cFileRecord+i, szUnicodeName, nUnicodeNameLen) == 0) // if record contains the name
				{
					fprintf(m_fDebug, "begin of checkFilenameForRecordIs()\n"); fflush(m_fDebug);
					nRes = checkFilenameForRecordIs(cFileRecord, szFilename, false);
					fprintf(m_fDebug, "end of checkFilenameForRecordIs(), return = %d\n", nRes); fflush(m_fDebug);
					if (nRes == 0)
					{	fprintf(m_fDebug, "searchForRecordFilename(): experimental search succeed. Offset = %llu\n", qwCurrentOffset); fflush(m_fDebug);
						newtWinMessage(i18n("$bitmap record search"), i18n("Ok"), i18n("experimental search succeed"));
						return 0; // success
					}
				}
			}
			
			fprintf(m_fDebug, "searchForRecordFilename: end FILE detected\n"); fflush(m_fDebug);	
			
		}
	}
	
	return -1;
}

