/***************************************************************************
                          common.cpp  -  description
                             -------------------
    begin                : Thu Jun 29 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "partimage.h"
#include "common.h"
#include "imagefile.h"
#include "misc.h"
#include "gui_text.h"

#include "cext2part.h"
#include "cfatpart.h"
#include "creiserpart.h"
#include "cntfspart.h"
#include "chpfspart.h"

#include <stdio.h>
#include <unistd.h>
#include <newt.h>

char *g_szFooterMagicString = "FOOTER-MAGIC";
	
// =======================================================
int systemChecks(char *szDevice, bool bFailIfMounted)
{
	char szMountPoint[1024];
	int nRes;
	
	// ----------- check for inode
	nRes = checkInodeForDevice(szDevice);
	if (nRes == -1)
		return -1;	
	
	// ----------- check the user is logged as root
	if (geteuid() != 0) // 0 is the UID of the root
	{	nRes = newtWinChoice(i18n("Warning"), i18n("Continue"), i18n("Cancel"), i18n("You are not logged as root. You may have \"access denied\" errors when working."));	
		if (nRes == 2)
			return -1;
	}		
	
	// ----------- check the partition is not mounted
	if ((bFailIfMounted == true) && (isMounted(szDevice, szMountPoint) != false))
	{	nRes = newtWinChoice(i18n("Error"), i18n("Continue"), i18n("Cancel"), i18n("The %s partition is mounted. Partition Image can't work on mounted partitions. "
		"Please, unmount it before. You can do it with \"umount %s\""), szDevice, szMountPoint);
		if (nRes == 2)
			return -1;
	}	

	return 0;
}

// =======================================================
int savePartition(char *szDevice, char *szImageName, char *szFilesystem, COptions *options)
{
	int nRes;
	time_t dt;
	FILE *fDeviceFile;
	QWORD qwFreeSpace;
	int nFs;
	CSaveWindow gui;
	QWORD qwPartSize;
	CMainHeader headMain;

	// structures to stock informations about the current partition
	CImage image;
	
	memset(&image, 0, sizeof(CImage));
	memset(&headMain, 0, sizeof(CMainHeader));
	
	// ----------- calculate filenames
	extractFilenameFromFullPath(szImageName, image.m_szOriginalFilename); // filename without path
	extractFilepathFromFullPath(szImageName, image.m_szPath); // filepath without filename
	
	if (options->dwSplitMode == SPLIT_NONE)
	{	sprintf(image.m_szImageFilename, "%s", szImageName);
	}
	else
	{	sprintf(image.m_szImageFilename, "%s.000", szImageName);
	}
			
	// initialize interface
	gui.create(szDevice, image.m_szImageFilename, *options);

	newtPushHelpLine(i18n("initializing the operation. Please wait..."));
	
	// ----------- system checks
	nRes = systemChecks(szDevice, options->bFailIfMounted);
	if (nRes == -1)
		return -1;

	// ----------- detect file system
	if (strcmp(szFilesystem, "detect") == 0)
	{	nFs = detectFileSystem(szDevice, szFilesystem);
		if (nFs == -1)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't detect file system of the %s partition, or unknown file system. (%s)"), szDevice, strerror(errno));
			return -1;
		}
		else
		{	gui.showDetectedFS(szFilesystem);
		}
	}
		
	// size of the partition
	qwPartSize = getPartitionSize(szDevice);
	gui.showPartitionInfo(szDevice, qwPartSize);
	
	getDiskFreeSpaceForFile(&qwFreeSpace, image.m_szImageFilename);
	gui.showAvaibleSpaceForImage(qwFreeSpace);
	
	// show operation window
	nRes = gui.runForm();
	if (nRes == -1)
		return -1;
	gui.getValues(options);
	
	// ask image description to the user
	if (options->bAskDesc)
	{
	    askDescription(headMain.szPartDescription, MAX_DESCRIPTION);
	}
	
	// ----------- open partition file
	errno = 0;
	fDeviceFile = fopen(szDevice, "rb");
	if (fDeviceFile == NULL)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't open partition %s. (%s)"), szDevice, strerror(errno));
		return -1;
	}
	
	// ----------- open the image file
	nRes = image.openWriting(*options);
	if (nRes == -1)
		return -1;
	
	image.setGuiSave(&gui);
	
	// ----------- write main header
	
	// fill header infos	
	strcpy(headMain.szAppName, "PartImage");
	strcpy(headMain.szFileSystem, szFilesystem);
	strcpy(headMain.szOriginalDevice, szDevice);
	
	extractFilenameFromFullPath(szImageName, headMain.szFirstImageFilepath); // filename without path
	headMain.dwCompression = options->dwCompression;
	headMain.dwMainFlags = 0;
	time(&dt);
	headMain.dateCreate = *localtime(&dt); // current time/date (date of image creation)

	// split mode
	if (options->dwSplitMode != SPLIT_NONE)
		headMain.dwMainFlags |= FLAG_SPLITTED;
	image.m_bIsSplitted = (options->dwSplitMode != SPLIT_NONE);
	
	headMain.qwPartSize = qwPartSize;
	
	nRes = image.write(&headMain, sizeof(CMainHeader), *options, true);
	if (nRes != sizeof(CMainHeader))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write main header"));
		closeFiles(*options, &image, fDeviceFile);
		return -1;
	}
	
	// copy options to debug file
	fprintf(m_fDebug, "\n\n============= begin OPTIONS =====================\n");
	fprintf(m_fDebug, "options.bOverwrite = %d\n", (int) options->bOverwrite);
	fprintf(m_fDebug, "options.dwSplitMode = %d\n", (int) options->dwSplitMode);
	fprintf(m_fDebug, "options.dwSplitSize = %d\n", (int) options->dwSplitSize);
	fprintf(m_fDebug, "options.bEraseWithNull = %d\n", (int) options->bEraseWithNull);
	fprintf(m_fDebug, "options.dwCompression = %d\n", (int) options->dwCompression);
	fprintf(m_fDebug, "options.bCheckBeforeSaving = %d\n", (int) options->bCheckBeforeSaving);
	fprintf(m_fDebug, "options.bFailIfMounted = %d\n", (int)options->bFailIfMounted);
	fprintf(m_fDebug, "options.bAskDesc = %d\n", (int) options->bAskDesc);
	fprintf(m_fDebug, "options.dwFinish = %d\n", (int) options->dwFinish);
	fprintf(m_fDebug, "============= end OPTIONS =====================\n\n\n");
	fflush(m_fDebug);
	
	// Write this in the help bar
	newtPushHelpLine(i18n("Saving partition to the image file..."));
		
	// -*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-
	if (strcmp(szFilesystem, "ext2fs") == 0) // if EXT2
	{	
		CExt2Part part(szDevice, fDeviceFile);
		fprintf(m_fDebug, "begin of CExt2Part::saveToImage()\n");
		nRes = part.saveToImage(&image, options, &gui);
		fprintf(m_fDebug, "end of CExt2Part::saveToImage(), return = %d\n", nRes);
	}
	else if (strcmp(szFilesystem, "reiserfs") == 0) // if REISER
	{	
		CReiserPart part(szDevice, fDeviceFile);
		fprintf(m_fDebug, "begin of CReiserPart::saveToImage()\n");
		nRes = part.saveToImage(&image, options, &gui);
		fprintf(m_fDebug, "end of CReiserPart::saveToImage(), return = %d\n", nRes);
	}
	else if (strcmp(szFilesystem, "fat") == 0) // if FAT
	{	
		CFatPart part(szDevice, fDeviceFile);
		fprintf(m_fDebug, "begin of CFatPart::saveToImage()\n");
		nRes = part.saveToImage(&image, options, &gui);
		fprintf(m_fDebug, "end of CFatPart::saveToImage(), return = %d\n", nRes);
	}
	else if (strcmp(szFilesystem, "ntfs") == 0) // if NTFS
	{
		newtWinMessage(i18n("Warning"), i18n("Ok"), i18n("The current NTFS support is experimental !"));
		
		CNtfsPart part(szDevice, fDeviceFile);
		fprintf(m_fDebug, "begin of CNtfsPart::saveToImage()\n");
		nRes = part.saveToImage(&image, options, &gui);
		fprintf(m_fDebug, "end of CNtfsPart::saveToImage(), return = %d\n", nRes);
	}
	else if (strcmp(szFilesystem, "hpfs") == 0) // if HPFS
	{
		CHpfsPart part(szDevice, fDeviceFile);
		fprintf(m_fDebug, "begin of CHpfsPart::saveToImage()\n");
		nRes = part.saveToImage(&image, options, &gui);
		fprintf(m_fDebug, "end of CHpfsPart::saveToImage(), return = %d\n", nRes);
	}
	else // if UNKNOWN
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("invalid file system \"%s\""), szFilesystem);
		nRes = -1;
	}
	// -*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-
		
	gui.destroyForm();

	if (nRes == -1) // if failed
	{
		closeFiles(*options, &image, fDeviceFile);
		return -1;	
	}
	
	// ---- write footer
	CMainFooter footerMain;
	memset(&footerMain, 0, sizeof(CMainFooter));
	strcpy(footerMain.szMagicString, g_szFooterMagicString);
	footerMain.qwCRC = image.m_qwCRC;
	footerMain.dwVolumeNumber = image.m_dwVolumeNumber;
	
	nRes = image.write(&footerMain, sizeof(CMainFooter), *options, false);
	if (nRes != sizeof(CMainFooter))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write main footer"));
		closeFiles(*options, &image, fDeviceFile);
		return -1;
	}
	
	// ---- Close files
	closeFiles(*options, &image, fDeviceFile);

	return 0; // success
}

// =======================================================
int restorePartition(char *szDevice, char *szImageName, COptions *options)
{
	int nRes;
	FILE *fDeviceFile;
	QWORD qwPartitionSize;
	CRestoreWindow gui;
	
	// structures to stock informations about the current partition
	CImage image;
	
	memset(&image, 0, sizeof(CImage));
	newtPushHelpLine(i18n("initializing the operation"));
	
	// ----------- system checks
	nRes = systemChecks(szDevice, options->bFailIfMounted);
	if (nRes == -1)
		return -1;

	// initialize interface
	gui.create(szDevice, szImageName, *options);
	
	// ----------- calculate filenames
	extractFilenameFromFullPath(szImageName, image.m_szOriginalFilename); // filename without path
	extractFilepathFromFullPath(szImageName, image.m_szPath); // filepath without filename
	strcpy(image.m_szImageFilename, szImageName);
		
	// ----------- check the image file is existing
	if (access(szImageName, F_OK | R_OK) == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("The %s image file doesn't exists or access denied."), szImageName);
		return -1;
	}
	
	// ----------- open partition file
	errno = 0;
	fDeviceFile = fopen(szDevice, "r+b");
	if (fDeviceFile == NULL)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't open partition %s. (%s)"), szDevice, strerror(errno));
		return -1;
	}	
	
	// ----------- get the image compression level
	options->dwCompression = (DWORD) getCompressionLevelForImage(image.m_szImageFilename);
	
	// ----------- open the image file
	nRes = image.openReading(*options);
	if (nRes == -1)
		return -1;

	image.setGuiRestore(&gui);
		
	// ----------- read header
	CMainHeader headMain;
		
	nRes = image.read(&headMain, sizeof(CMainHeader), false, *options, true); // false = maybe
	if (nRes != sizeof(CMainHeader))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read main header"));
		closeFiles(*options, &image, fDeviceFile);
		return -1;
	}
	
	// ----------- check header data
	if (strcmp(headMain.szAppName, "PartImage") != 0)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("%s is not a valid partition image file, or the compression option doesn't "
		"fit with the image file. Please, check this is the first volume and the good compression level is used."), szImageName);
		closeFiles(*options, &image, fDeviceFile);
		return -1;
	}
	
	image.m_bIsSplitted = !!(headMain.dwMainFlags & FLAG_SPLITTED);
	
	// ----------- image ?= splitted
	strncpy(image.m_szOriginalFilename, headMain.szFirstImageFilepath, MAXPATHLEN);
	image.m_bIsSplitted = (bool) (headMain.dwMainFlags & FLAG_SPLITTED);
	
	if (image.m_bIsSplitted)
		gui.showSplittedInfo(i18n("The image was splitted"));
	else
		gui.showSplittedInfo(i18n("The image was not splitted"));
		
	// check the destination partition is large enough to restore data
	qwPartitionSize = getPartitionSize(szDevice);
	if (qwPartitionSize < headMain.qwPartSize)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("The partition is too small to be restored:\n"
	               "Original partition size:........%llu bytes\n"
	               "Destination partition size:.....%llu bytes"), (QWORD) headMain.qwPartSize, qwPartitionSize);
		return -1;
	}
	
	gui.showPartitionInfo(szDevice, qwPartitionSize);

	// ----------- print informations about image file
	gui.showPartitionInfos(options->dwCompression, headMain.szOriginalDevice, headMain.szFileSystem, headMain.dateCreate, headMain.qwPartSize);
	
	// ----------- show the description
	headMain.szPartDescription[MAX_DESCRIPTION] = 0;
	if (*headMain.szPartDescription)
		newtWinMessage(i18n("Partition description"), i18n("Ok"), headMain.szPartDescription);
	
	// show operation window
	nRes = gui.runForm();
	if (nRes == -1)
		return -1;
	gui.getValues(options);

	// copy options to debug file
	fprintf(m_fDebug, "\n\n============= begin OPTIONS =====================\n");
	fprintf(m_fDebug, "options.bOverwrite = %d\n", (int) options->bOverwrite);
	fprintf(m_fDebug, "options.dwSplitMode = %d\n", (int) options->dwSplitMode);
	fprintf(m_fDebug, "options.dwSplitSize = %d\n", (int) options->dwSplitSize);
	fprintf(m_fDebug, "options.bEraseWithNull = %d\n", (int) options->bEraseWithNull);
	fprintf(m_fDebug, "options.dwCompression = %d\n", (int) options->dwCompression);
	fprintf(m_fDebug, "options.bCheckBeforeSaving = %d\n", (int) options->bCheckBeforeSaving);
	fprintf(m_fDebug, "options.bFailIfMounted = %d\n", (int)options->bFailIfMounted);
	fprintf(m_fDebug, "options.bAskDesc = %d\n", (int) options->bAskDesc);
	fprintf(m_fDebug, "options.dwFinish = %d\n", (int) options->dwFinish);
	fprintf(m_fDebug, "============= end OPTIONS =====================\n\n\n");
	fflush(m_fDebug);
	
	// -*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-
	if (strcmp(headMain.szFileSystem, "ext2fs") == 0) // if EXT2
	{	
		CExt2Part part(szDevice, fDeviceFile);
		fprintf(m_fDebug, "begin of CExt2Part::restoreFromImage()\n");
		nRes = part.restoreFromImage(&image, options, &gui);
		fprintf(m_fDebug, "end of CExt2Part::restoreFromImage(), retrun = %d\n", nRes);
	}
	else if (strcmp(headMain.szFileSystem, "reiserfs") == 0) // if REISER
	{	
		CReiserPart part(szDevice, fDeviceFile);
		fprintf(m_fDebug, "begin of CReiserPart::restoreFromImage()\n");
		nRes = part.restoreFromImage(&image, options, &gui);
		fprintf(m_fDebug, "end of CReiserPart::restoreFromImage(), retrun = %d\n", nRes);
	}
	else if (strcmp(headMain.szFileSystem, "fat") == 0) // if FAT
	{	
		CFatPart part(szDevice, fDeviceFile);
		fprintf(m_fDebug, "begin of CFatPart::restoreFromImage()\n");
		nRes = part.restoreFromImage(&image, options, &gui);
		fprintf(m_fDebug, "end of CFatPart::restoreFromImage(), retrun = %d\n", nRes);
	}
	else if (strcmp(headMain.szFileSystem, "ntfs") == 0) // if NTFS
	{	
		CNtfsPart part(szDevice, fDeviceFile);
		fprintf(m_fDebug, "begin of CNtfsPart::restoreFromImage()\n");
		nRes = part.restoreFromImage(&image, options, &gui);
		fprintf(m_fDebug, "end of CNtfsPart::restoreFromImage(), retrun = %d\n", nRes);
	}
	else if (strcmp(headMain.szFileSystem, "hpfs") == 0) // if HPFS
	{	
		CHpfsPart part(szDevice, fDeviceFile);
		fprintf(m_fDebug, "begin of CHpfsPart::restoreFromImage()\n");
		nRes = part.restoreFromImage(&image, options, &gui);
		fprintf(m_fDebug, "end of CHpfsPart::restoreFromImage(), retrun = %d\n", nRes);
	}
	else // if UNKNOWN
	{
		newtWinMessage(i18n("Error"), i18n("Ok"), i18n("The %s file system is not supported"), headMain.szFileSystem);
		nRes = -1;
		return -1;
	}
	// -*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-
	
	if (nRes == -1) // if failed
	{
		closeFiles(*options, &image, fDeviceFile);
		return -1;	
	}
	
	// ---- check footer
	CMainFooter footerMain;
	
	nRes = image.read(&footerMain, sizeof(CMainFooter), image.m_bIsSplitted, *options, false);
	if (nRes != sizeof(CMainFooter))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read main footer"));
		closeFiles(*options, &image, fDeviceFile);
		return -1;
	}
	
	// check the magic string
	if (strcmp(footerMain.szMagicString, g_szFooterMagicString) != 0)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Invalid main footer"));
		closeFiles(*options, &image, fDeviceFile);
		fprintf(stderr,"\n*****************\n%20s\n********instead of **************\n%20s\n**************\n", footerMain.szMagicString, g_szFooterMagicString);
		return -1;
	}
	
	// check CRC 64
	if (image.m_qwCRC != footerMain.qwCRC)
	{
		newtWinMessage(i18n("Error"), i18n("Ok"), i18n("There was an error while restoring partition:\n"
		"The image file was modified, or there was an error. CRC64 are differents:\n"
		"Original CRC64: %llu\nCurrent CRC64: %llu"), footerMain.qwCRC, image.m_qwCRC);
		closeFiles(*options, &image, fDeviceFile);
		return -1;
	}
	else
	{	//newtWinMessage(i18n("Success"), i18n("Ok"), i18n("CRC 64 is good (%llu)"), image.m_qwCRC);
	}
	
	// check volumes number
	if (image.m_dwVolumeNumber != footerMain.dwVolumeNumber)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Invalid volume number (%ld instead of %ld)."), image.m_dwVolumeNumber, footerMain.dwVolumeNumber);
		closeFiles(*options, &image, fDeviceFile);
		return -1;
	}

	// ---- close files
	closeFiles(*options, &image, fDeviceFile);
	
	return 0;
}

// =======================================================
void closeFiles(COptions options, CImage *image, FILE *fDeviceFile)
{
	int nRes = -1;
	
	if (fDeviceFile)
		fclose(fDeviceFile);
	
	nRes = image->close();
	
	if (nRes == -1)
		newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Error on close"), nRes);
}

// ========================================================================================
bool isFileSystemSupported(char *szFileSystem)
{
	return ((strcmp(optarg, "ext2fs") == 0) ||
					(strcmp(optarg, "reiserfs") == 0) ||
					(strcmp(optarg, "ntfs") == 0) ||
					(strcmp(optarg, "hpfs") == 0) ||
					(strcmp(optarg, "fat") == 0));
}






