/***************************************************************************
                          common.h  -  description
                             -------------------
    begin                : Thu Jun 29 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef GENERAL_H
#define GENERAL_H

#include "partimage.h"

// file systems
#include "cext2part.h"
#include "cfatpart.h"

#include <sys/vfs.h>
#include <stdio.h>

struct CImage;
struct CVolumeHeader;
struct CMainHeader;
struct CMainFooter;
struct COptions;

// =======================================================
int savePartition(char *szDevice, char *szImageName, char *szFilesystem, COptions *options);
int restorePartition(char *szDevice, char *szImageName, COptions *options);

// =======================================================
void closeFiles(COptions options, CImage *image, FILE *fDeviceFile);
int systemChecks(char *szDevice, bool bFailIfMounted);
bool isFileSystemSupported(char *szFileSystem);

#define GZIP_FREESPACE_NEED	65536 // else, zlib can't close the image file, and data error

// ================================================
struct CVolumeHeader // size must be 512 (adjust the reserved data)
{
	char szMagicString[32];
	DWORD dwVolumeNumber;
	
	unsigned char cReserved[476]; // Adjust to fit with total header size
};

// ================================================
struct CMainFooter // size must be 16384 (adjust the reserved data)
{
	char szMagicString[32];
	QWORD qwCRC;
	DWORD dwVolumeNumber;
	
	unsigned char cReserved[16340]; // Adjust to fit with total header size
};

// ================================================
struct CMainHeader // size must be 16384 (adjust the reserved data)
{
	char szAppName[32]; // "PartImage" // as MZ for EXE or PNG, ...
	char szFileSystem[512]; // ext2fs, ntfs, reiserfs
	char szPartDescription[MAX_DESCRIPTION]; // user description of the partition
	char szOriginalDevice[MAX_DEVICENAMELEN]; // original partition name
	char szFirstImageFilepath[MAXPATHLEN]; // for splitted image files
	DWORD dwCompression; // COMPRESS_XXXXXX
	DWORD dwMainFlags;
	
	// date of image creation
	struct tm dateCreate;
	QWORD qwPartSize; // size of the partition in bytes
	
	unsigned char cReserved[7076]; // Adjust to fit with total header size
};

// ================================================
inline int increaseCRC(QWORD *qwCRC, const char *cBlock, DWORD dwBlockLength)
{
	register DWORD i;
	for (i=0L; i < dwBlockLength; i++)
		*qwCRC += *(cBlock+i);
		
	return 0;
}

#endif // GENERAL_H
