/***************************************************************************
                          creiserpart.cpp  -  description
                             -------------------
    begin                : Thu Jul 13 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>

//#include <linux/fs.h>
//#include <linux/list.h>
//#include <linux/kdev_t.h>
#include <linux/types.h>

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//#include <linux/reiserfs_fs.h>
//#include <linux/reiserfs_fs_sb.h>

/*
I cannot include these two files: id I do it, I have two errors:
- kdev_t was net declared in this scope
- field 'list' has incomplete type
Then the only solution is to copy lines from reiser header files.
*/

#define REISERFS_SUPER_MAGIC_STRING "ReIsErFs"
#define REISERFS_DISK_OFFSET_IN_BYTES (64 * 1024)

struct reiserfs_super_block
{
  __u32 s_block_count;			/* blocks count         */
  __u32 s_free_blocks;                  /* free blocks count    */
  __u32 s_root_block;           	/* root block number    */
  __u32 s_journal_block;           	/* journal block number    */
  __u32 s_journal_dev;           	/* journal device number  */
  __u32 s_orig_journal_size; 		/* size of the journal on FS creation.  used to make sure they don't overflow it */
  __u32 s_journal_trans_max ;           /* max number of blocks in a transaction.  */
  __u32 s_journal_block_count ;         /* total size of the journal. can change over time  */
  __u32 s_journal_max_batch ;           /* max number of blocks to batch into a trans */
  __u32 s_journal_max_commit_age ;      /* in seconds, how old can an async commit be */
  __u32 s_journal_max_trans_age ;       /* in seconds, how old can a transaction be */
  __u16 s_blocksize;                   	/* block size           */
  __u16 s_oid_maxsize;			/* max size of object id array, see get_objectid() commentary  */
  __u16 s_oid_cursize;			/* current size of object id array */
  __u16 s_state;                       	/* valid or error       */
  char s_magic[12];                     /* reiserfs magic string indicates that file system is reiserfs */
  __u32 s_hash_function_code;		/* indicate, what hash fuction is being use to sort names in a directory*/
  __u16 s_tree_height;                  /* height of disk tree */
  __u16 s_bmap_nr;                      /* amount of bitmap blocks needed to address each block of file system */
  __u16 s_reserved;
};

#define SB_SIZE (sizeof(struct reiserfs_super_block))

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "creiserpart.h"
#include "partimage.h"
#include "imagefile.h"
#include "misc.h"
#include "gui_text.h"

// =======================================================
CReiserPart::CReiserPart(char *szDevice, FILE *fDeviceFile)
{
	fprintf(m_fDebug, "begin of CReiserPart::CReiserPart()\n"); fflush(m_fDebug);  
	m_fDeviceFile = fDeviceFile;
	strncpy(m_szDevice, szDevice, MAX_DEVICENAMELEN);
	fprintf(m_fDebug, "end of CReiserPart::CReiserPart()\n"); fflush(m_fDebug);  
}

// =======================================================
CReiserPart::~CReiserPart()
{
}

// =======================================================
int CReiserPart::saveUsedBlocksToImageFile(COptions *options)
{
	int nRes;
	DWORD i;
	char cBuffer[MAX_REISER_BLOCK_SIZE+1];
	
	DWORD dwSavedBlocks = 0L;

	// go to the beginning of the disk
	nRes = fseek(m_fDeviceFile, 0L, SEEK_SET);
	if (nRes == -1)
	    return -1;
	
	for (i=0; i < m_dwBlocksCount; i++)
	{
		// if block is used...
		if (m_bitmap.isBitSet(i)) // not damaged and not free
		{
			dwSavedBlocks++;
			//printf("%d used\n", i);
				
			errno = 0;
			nRes = fread(cBuffer, m_dwBlockSize, 1, m_fDeviceFile);
			if (nRes != 1)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read block %ld from partition (%s)"), i, strerror(errno));
				return -1;
			}
				
			errno = 0;
			nRes = m_image->write(cBuffer, m_dwBlockSize, *options, true);
			if (nRes != m_dwBlockSize)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write block %ld to image (%s)"), i, strerror(errno));
				return -1;
			}
		}
		else // block free --> skip
		{
			//printf("%d free\n", i);
			nRes = fseek(m_fDeviceFile, m_dwBlockSize, SEEK_CUR);
			if (nRes == -1)
				return -1;	
		}
		
		// draw progress
		if (dwSavedBlocks%20 == 0)
			m_guiSave->setProgressValue(dwSavedBlocks, m_dwUsedBlocks);
	}
		
	m_guiSave->setProgressValue(m_dwUsedBlocks, m_dwUsedBlocks);

	
	return 0;
}

// =======================================================
int CReiserPart::restoreUsedBlocksToDisk(COptions *options)
{
	int nRes;
	DWORD i;
	char cBuffer[MAX_REISER_BLOCK_SIZE+1];
	char cZeroBuffer[MAX_REISER_BLOCK_SIZE+1];
	
	DWORD dwSavedBlocks = 0L;
	memset(cZeroBuffer, 0, MAX_REISER_BLOCK_SIZE);

	// go to the beginning of the disk
	nRes = fseek(m_fDeviceFile, 0L, SEEK_SET);
	if (nRes == -1)
	    return -1;
	
	for (i=0; i < m_dwBlocksCount; i++)
	{	
		if (m_bitmap.isBitSet(i)) // block used
		{
			dwSavedBlocks++;
				
		  nRes = m_image->read(cBuffer, m_dwBlockSize, m_image->m_bIsSplitted, *options, true);
			if (nRes != m_dwBlockSize)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read data block %ld from image file"), i);
				return -1;
			}	
				
			errno = 0;
			nRes = fwrite(cBuffer, m_dwBlockSize, 1, m_fDeviceFile);
			if (nRes != 1)
			{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't restore data block %ld from image file. (%s) (%d)"), i, strerror(errno));
				return -1;
			}				
		}
		else // block free --> skip
		{
			if (options->bEraseWithNull == true) // if free and we must write 0 bytes on unused clusters
			{
				errno = 0;
				nRes = fwrite(cZeroBuffer, m_dwBlockSize, 1, m_fDeviceFile);
				if (nRes != 1)
				{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write zero block %lu (%s) (%d)"), i, strerror(errno));
					return -1;
				}				
				dwSavedBlocks++;
			}
			else // don't write zero bytes
			{
				nRes = fseek(m_fDeviceFile, m_dwBlockSize, SEEK_CUR);
				if (nRes == -1)
					return -1;	
			}			
		}
		
		// draw progress
		if (dwSavedBlocks%20 == 0)
			m_guiRestore->setProgressValue(dwSavedBlocks, m_dwUsedBlocks, m_dwBlocksCount, options->bEraseWithNull);
	}
		
	m_guiRestore->setProgressValue(dwSavedBlocks, m_dwUsedBlocks, m_dwBlocksCount, options->bEraseWithNull);
	
	return 0;
}

// =======================================================
int CReiserPart::copyBitmap(COptions *options)
{
	DWORD i;
	DWORD dwBlock;
	BYTE cBuffer[MAX_REISER_BLOCK_SIZE+1];
	int nRes;
	DWORD dwOffset;

	// init bitmap size
	m_bitmap.init(m_dwBitmapBlocksCount * m_dwBlockSize * 8);
	
	//printf ("\nBitmap blocks are: ");

	// copy bitmap blocks to bitmap in memory
	for (i = 0; i < m_dwBitmapBlocksCount; i ++)
	{	
		// We don't use fssek(dwCurrentBlock, SEEK_SET) after bitmap block 1, because fseek is
		// limited to 2 GB (because offset is long int). Then, we try to use an ossfet which can
		// fit in this long int. Then we only seek from the last bitmap block
	
		if (i == 0) // TODO: Check it for bitmap block 0
		{
			dwBlock = (REISERFS_DISK_OFFSET_IN_BYTES / m_dwBlockSize) + 1; // Skipped blocks + super block
			nRes = fseek(m_fDeviceFile, (long) (dwBlock * m_dwBlockSize), SEEK_SET);
		}
		else if (i == 1)
		{
			dwBlock = i * m_dwBlockSize * 8;
			nRes = fseek(m_fDeviceFile, (long) (dwBlock * m_dwBlockSize), SEEK_SET);
		}
		else // (i > 1)
		{
			dwBlock = i * m_dwBlockSize * 8;
			dwOffset = ((m_dwBlockSize * 8) - 1) * m_dwBlockSize;		
			nRes = fseek(m_fDeviceFile, (long) dwOffset, SEEK_CUR);
		}
		
		//printf ("%ld ", dwBlock);
	
		// go to the beginning of the block
		if (nRes == -1) // result of the fseek()
			return -1;
		
		// read bitmap from disk
		errno = 0;
		nRes = fread(cBuffer, m_dwBlockSize, 1, m_fDeviceFile);
		if (nRes != 1)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read bitmap block %ld (%ld): %s"), i, dwBlock, strerror(errno));
			return -1;
		}
				
		// write bitmap to image file
		errno = 0;
		nRes = m_image->write(cBuffer, m_dwBlockSize, *options, true);
		if (nRes != m_dwBlockSize)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write bitmap block %ld (%ld) to image. (%s)\n"), i, dwBlock, strerror(errno));
			return -1;
		}
	
		// write bitmap in memory
		memcpy(m_bitmap.data()+(i*m_dwBlockSize), cBuffer, m_dwBlockSize);
	}
	
  return 0;
}			

// =======================================================
int CReiserPart::printfInformations()
{
	char szTemp[2048];
	char szTemp2[2048];
	char szText[8192];
	
	sprintf(szText,  i18n("Block size....................%lu bytes\n"), m_dwBlockSize);
	sprintf(szTemp2, i18n("Total blocks count............%lu\n"), m_dwBlocksCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Used blocks count.............%lu\n"), m_dwUsedBlocks);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Free blocks count.............%lu\n"), m_dwBlocksCount-m_dwUsedBlocks);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Bitmap blocks count...........%lu\n"), m_dwBitmapBlocksCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Space usage:..................%lu %%\n"), (100 * m_dwUsedBlocks)/m_dwBlocksCount);
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Used size.....................%s\n"), formatSize(m_dwUsedBlocks * m_dwBlockSize, szTemp));
	strcat(szText, szTemp2);
	sprintf(szTemp2, i18n("Free size.....................%s\n"), formatSize((m_dwBlocksCount-m_dwUsedBlocks) * m_dwBlockSize, szTemp));
	strcat(szText, szTemp2);

	fprintf(m_fDebug, "\n====================\n%s====================\n\n", szText); fflush(m_fDebug);
	newtWinMessage(i18n("ReiserFS informations"), i18n("Ok"), szText);
						
  return 0;
}			

// =======================================================
int CReiserPart::readInformations()
{
	reiserfs_super_block sb;
	int nRes;
	
	// 0. go to the beginning of the super block
	nRes = fseek(m_fDeviceFile, REISERFS_DISK_OFFSET_IN_BYTES, SEEK_SET);
	if (nRes == -1)
		return -1;
	
	// 1. read and print important informations
	nRes = fread(&sb, SB_SIZE, 1, m_fDeviceFile);
	if (nRes != 1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read super block"));
		return -1;
	}
		
	// check the magic number is good
	if (strncmp(sb.s_magic, REISERFS_SUPER_MAGIC_STRING, strlen(REISERFS_SUPER_MAGIC_STRING)) != 0)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Not a reiserfs partition. Check the file system type."));
		return -1;
	}
	
	// 2. calculate infos
	m_dwBlockSize = sb.s_blocksize;
	m_dwUsedBlocks = sb.s_block_count - sb.s_free_blocks;
	m_dwBlocksCount = sb.s_block_count;
	m_dwBitmapBlocksCount = sb.s_bmap_nr;
		
	// check data
	if ((m_dwBlockSize % 512 != 0) || (m_dwBlockSize < 512))
		return -1;
	if (m_dwUsedBlocks > m_dwBlocksCount)
		return -1;
	
	// Label: TODO
	
	return 0;
}

// =======================================================
int CReiserPart::saveToImage(CImage *image, COptions *options, CSaveWindow *gui)
{
	image->m_dwCurrentVolumeSize = 0L;
	image->m_dwVolumeNumber = 0L;
	m_guiSave = gui;
	int nRes;	
	
	m_image = image;

	// ---- 0. Initializations
	
	// A. read informations
	newtPushHelpLine(i18n("reading partition properties"));
	fprintf(m_fDebug, "begin of readInformations()\n"); fflush(m_fDebug);  
	nRes = readInformations();
	fprintf(m_fDebug, "end of readInformations()\n"); fflush(m_fDebug);  
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read partition super block."));
		return -1;
	}	

	// B. print informations	
	newtPushHelpLine(i18n("reading partition informations"));
	printfInformations();
	
	// ---- 1. Write Reiser header
	newtPushHelpLine(i18n("writing header"));
	CReiserHeader headReiser;
	memset(&headReiser, 0, sizeof(CReiserHeader));
	
	strcpy(headReiser.szAppFormat, CURRENT_REISER_FORMAT);
	
	headReiser.dwBlockSize = m_dwBlockSize;
	headReiser.dwUsedBlocks = m_dwUsedBlocks;
	headReiser.dwBlocksCount = m_dwBlocksCount;
	headReiser.dwBitmapBlocksCount = m_dwBitmapBlocksCount;
	
	errno = 0;
	nRes = m_image->write(&headReiser, sizeof(CReiserHeader), *options, true);
	if (nRes != sizeof(CReiserHeader))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't write reiser header: (%s)"), strerror(errno));
		return -1;
	}

	// B. Copy bitmap to memory and to image file
	nRes = m_image->writeMagic(MAGIC_BEGIN_BITMAP, *options);
	if (nRes == -1)
		return -1;
	
	fprintf(m_fDebug, "begin of copyBitmap()\n"); fflush(m_fDebug);
	copyBitmap(options);
	fprintf(m_fDebug, "end of copyBitmap()\n"); fflush(m_fDebug);

	// write magic string
	nRes = m_image->writeMagic(MAGIC_BEGIN_DATABLOCKS, *options);
	if (nRes == -1)
		return -1;

	// ---- 5. Write all used blocks
	newtPushHelpLine(i18n("copying used data blocks"));
	fprintf(m_fDebug, "begin of saveUsedBlocksToImageFile()\n"); fflush(m_fDebug);  
	nRes = saveUsedBlocksToImageFile(options);
	fprintf(m_fDebug, "end of saveUsedBlocksToImageFile()\n"); fflush(m_fDebug);  
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Unable to copy used blocks to the image file"));
		return -1;
	}
	
	return 0;
}

// =======================================================
int CReiserPart::restoreFromImage(CImage *image, COptions *options, CRestoreWindow *gui)
{
	CReiserHeader headReiser;
	int nRes;

	// init
	m_image = image;
	m_guiRestore = gui;

	newtPushHelpLine(i18n("reading partition informations"));
	
	// 1. ---- Read Reiser header	
	memset(&headReiser, 0, sizeof(CReiserHeader));
	nRes = m_image->read(&headReiser, sizeof(CReiserHeader), m_image->m_bIsSplitted, *options, true);
	if (nRes != sizeof(CReiserHeader))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read reiser header from the image file"));
		return -1;
	}
	
	// check the current format is supported
	if (strncmp(headReiser.szAppFormat, CURRENT_REISER_FORMAT, strlen(CURRENT_REISER_FORMAT)) < 0)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("The partition was saved with a newer version of Partition Image. Please download the last version."));
		return -1;
	}	
	
	m_dwBlockSize = headReiser.dwBlockSize;
	m_dwUsedBlocks = headReiser.dwUsedBlocks;
	m_dwBlocksCount = headReiser.dwBlocksCount;
	m_dwBitmapBlocksCount = headReiser.dwBitmapBlocksCount;
	
	// 2. ---- Print informations and ask confirmation to restore
	printfInformations();
	
	// ask confirmation
	nRes = newtWinChoice(i18n("Confirmation"), i18n("Yes"), i18n("No"), i18n("Do you really want to restore %s from the image file %s ?"), m_szDevice, image->m_szImageFilename);
	if (nRes == 2)
		return -1;
	
	// ---- 5. read bitmap
	nRes = m_image->readAndCheckMagic(MAGIC_BEGIN_BITMAP, *options);
	if (nRes == -1)
		return -1;
	
	// init bitmap size
	m_bitmap.init(m_dwBitmapBlocksCount * m_dwBlockSize * 8);

	// copy bitmap blocks to bitmap in memory
	nRes = m_image->read(m_bitmap.data(), (long)(m_dwBlockSize * m_dwBitmapBlocksCount) , m_image->m_bIsSplitted, *options, true);
	if (nRes != (m_dwBlockSize * m_dwBitmapBlocksCount))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read bitmap from image file"));
		return -1;
	}
	
	// read and check magic string
	nRes = m_image->readAndCheckMagic(MAGIC_BEGIN_DATABLOCKS, *options);
	if (nRes == -1)
		return -1;
	
	// ---- 7. Restore all used blocks
	errno = 0;
	newtPushHelpLine(i18n("copying used data blocks"));
	fprintf(m_fDebug, "begin of restoreUsedBlocksToDisk()\n"); fflush(m_fDebug);  
	nRes = restoreUsedBlocksToDisk(options);
	fprintf(m_fDebug, "end of restoreUsedBlocksToDisk()\n"); fflush(m_fDebug);  
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Unable to restore used blocks to disk. (%s)"), strerror(errno));
		return -1;
	}

	return 0;
}

