/***************************************************************************
                          gui_text.cpp  -  description
                             -------------------
    begin                : Tue Aug 22 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "gui_text.h"
#include "partimage.h"
#include "misc.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <newt.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

// =======================================================
int fillPartitionList(newtComponent editPartition)
{
	FILE *fPart;
	int nMajor, nMinor, nBlocks;
	char szDevice[128];
	char szFullDevice[128];
	char cBuffer[32768];
	char cFormat[1024];
	char szTemp[1024];
	char *cPtr;
	int nLines;
	int nSize;
	QWORD qwSize;
	int i;
	int nRes;
	char *szDeviceName;
	int nPartNum;
	
	errno = 0;
	fPart = fopen("/proc/partitions", "rb");
	if (!fPart)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Cannot read \"/proc/partitions\" (%s). Then, you must use the command line to run Partition Image. Type \"partimage --help\" for help."), strerror(errno));
		return -1;	
	}

	nSize = 0;
	nLines = 0;
	memset(cBuffer, 0, 32767);
	while (!feof(fPart))
	{
		cBuffer[nSize] = fgetc(fPart);
		if (cBuffer[nSize] == '\n')
			nLines++;		
		nSize++;
	}
	
	cPtr = cBuffer;

	// skip two first lines
	nLines -= 2;
	for (i=0; i < 2; i++)
	{
		while (*cPtr != '\n')
			cPtr++;
		cPtr++;
	}
	
	for (i=0; i < nLines; i++)
	{
		cPtr = decodePartitionEntry(cPtr, &nMajor, &nMinor, &nBlocks, &nPartNum, szDevice);
		if (/*nMinor && */nPartNum) // if not an hard drive (as hda, hdc)
		{
			memset(cFormat, ' ', 21);
			memset(cFormat+21, 0, 50);
			
			memcpy(cFormat, szDevice, strlen(szDevice)); // Device
			
			sprintf(szFullDevice, "/dev/%s", szDevice);
			strcpy(szTemp, "unknown");
			nRes = detectFileSystem(szFullDevice, szTemp);
			memcpy(cFormat+8, szTemp, strlen(szTemp)); // File system
			
			if (nRes != -1) // if file system known
			{
				qwSize = getPartitionSize(szFullDevice);
				formatSize(qwSize, szTemp);	
				memcpy(cFormat+21, szTemp, strlen(szTemp)); // Size
			}
			
			szDeviceName = new char[strlen(szFullDevice)+1]; // TODO: never deleted from memory
			strcpy(szDeviceName, szFullDevice);
			newtListboxAppendEntry(editPartition, cFormat, (void*)szDeviceName);
		}
	}

	fclose(fPart);
	return 0;
}

// =======================================================
int guiInitMainWindow(char *szDevice, char *szImageFile)
{
	newtComponent formMain, btnContinue, btnExit, btnAbout;
	newtComponent labelPartition, labelImage, editPartition, editImage;
	newtComponent widgetTemp, labelAction, radioSave, radioRestore;
	char szTemp[1024];
	int nAction;
	int nRes;
	
	sprintf(szTemp, "Partition Image %s", VERSION);
	newtCenteredWindow(67, 20, szTemp);

	labelPartition = newtLabel(1, 1, i18n("* Partition to save/restore"));
	editPartition = newtListbox(3, 2, 9, NEWT_FLAG_SCROLL);
	labelImage = newtLabel(1, 12, i18n("* Image file to create/use"));
	editImage = newtEntry(3, 13, szImageFile, 60, NULL, 0);
	labelAction = newtLabel(1, 15, i18n("Action to be done:"));
	radioSave = newtRadiobutton(1, 16, i18n("Save partition into a new image file"), true, NULL);
	radioRestore = newtRadiobutton(1, 17, i18n("Restore partition from an existing image file"), false, radioSave);
	btnContinue = newtCompactButton(5, 19, i18n("Next >"));
	btnAbout = newtCompactButton(27, 19, i18n("About"));
	btnExit = newtCompactButton(50, 19, i18n("Exit"));
	
	nRes = fillPartitionList(editPartition);
	if (nRes == -1)
		return -1;
	
	formMain = newtForm(NULL, NULL, 0);
	newtFormAddComponents(formMain, labelPartition, labelImage, editPartition, editImage,
	labelAction, radioSave, radioRestore, btnContinue,  btnAbout, btnExit,NULL);

begin:
	widgetTemp = newtRunForm(formMain);
	
 	if(widgetTemp == btnExit)
		return OPERATION_EXIT;
	if (widgetTemp == btnAbout)
	{	showAboutDialog();
		goto begin;
	}

	// get and check values
	/*if (strcmp(newtEntryGetValue(editPartition), "") == 0) // if "partition" field empty
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("The \"Partition\" filed is empty. Cannot continue"));
		goto begin;
	}*/
	if (strcmp(newtEntryGetValue(editImage), "") == 0) // if "image" field empty
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("The \"Image\" filed is empty. Cannot continue"));
		goto begin;
	}
	
	// get device
	strcpy(szDevice, (char*)newtListboxGetCurrent(editPartition));
	
	// image file
	strcpy(szImageFile, newtEntryGetValue(editImage));
	
	nAction = 0;
		
	if (newtRadioGetCurrent(radioRestore) == radioSave)
	{	
	  nAction = OPERATION_SAVE;
	}
	else if(newtRadioGetCurrent(radioRestore) == radioRestore)
	{
		nAction = OPERATION_RESTORE;
	}

	newtFormDestroy(formMain);
	newtPopWindow();
			
	return nAction;
}

// =======================================================
int CSaveWindow::create(char *szDevice, char *szImageFile, COptions options)
{
	char szTemp[1024];

	sprintf(szTemp, "save partition to image file");
	newtCenteredWindow(78, 20, szTemp);

	sprintf(szTemp, i18n("Partition to save:.........%s"), szDevice);
	m_labelPartition = newtLabel(1, 0, szTemp);
	sprintf(szTemp, i18n("Image file to create.......%s"), szImageFile);
	m_labelImageFile = newtLabel(1, 1, szTemp);
	m_labelFreeSpace = newtLabel(1, 2, "");
	m_labelFS = newtLabel(1, 3, "");
	
	m_labelCompression = newtLabel(1, 5, i18n("Compression level"));
	m_labelOptions = newtLabel(25, 5, i18n("Options"));
	m_labelSplit = newtLabel(1, 10, i18n("Image split mode"));
	m_labelFinish = newtLabel(43, 10, i18n("If finished successfully:"));
	m_radioCompNone = newtRadiobutton(1, 6, i18n("None"), options.dwCompression == COMPRESS_NONE, NULL);
	m_radioCompGzip = newtRadiobutton(1, 7, "Gzip (.gz)", options.dwCompression == COMPRESS_GZIP, m_radioCompNone);
	m_radioCompBzip2 = newtRadiobutton(1, 8, "Bzip2 (.bz2)", options.dwCompression == COMPRESS_BZIP2, m_radioCompGzip);
	m_checkCheckBeforeSaving = newtCheckbox(25, 6, i18n("Check partition before saving"), (!!options.bCheckBeforeSaving ? '*' : ' '), " *", NULL);
	m_checkAskDesc = newtCheckbox(25, 7, i18n("Enter description"), (!!options.bAskDesc ? '*' : ' '), " *", NULL);
	m_checkOverwrite = newtCheckbox(25, 8, i18n("Overwrite without prompt"), (!!options.bOverwrite ? '*' : ' '), " *", NULL);
	m_radioSplitNone = newtRadiobutton(1, 11, i18n("No split"), options.dwSplitMode == SPLIT_NONE, NULL);
	m_radioSplitSize = newtRadiobutton(1, 12, i18n("Into files whose size is:"), options.dwSplitMode == SPLIT_SIZE, m_radioSplitNone);
	m_radioSplitSize2 = newtLabel(14, 13, i18n("Kilo-Bytes"));
	m_radioSplitAuto = newtRadiobutton(1, 14, i18n("Automatic (when no space left)"), options.dwSplitMode == SPLIT_AUTO, m_radioSplitSize);
	sprintf(szTemp, "%lu", options.dwSplitSize/1024);
	m_editSplitSize = newtEntry(5, 13, szTemp, 8, NULL, 0);
	m_radioFinishWait = newtRadiobutton(43, 11, i18n("Wait"), options.dwFinish == FINISH_WAIT, NULL);
	m_radioFinishHalt = newtRadiobutton(43, 12, i18n("Halt"), options.dwFinish == FINISH_HALT, m_radioFinishWait);
	m_radioFinishReboot = newtRadiobutton(43, 13, i18n("Reboot"), options.dwFinish == FINISH_REBOOT, m_radioFinishHalt);
	m_progressSaving = newtScale(1, 17, 70, 100);
	m_labelPercent = newtLabel(72, 17, "");
	m_btnSave = newtCompactButton(20, 19, i18n("Save"));
	m_btnExit = newtCompactButton(42, 19, i18n("Exit"));
	
	m_formMain = newtForm(NULL, NULL, 0);
	newtFormAddComponents(m_formMain, m_labelPartition, m_labelImageFile, m_labelFreeSpace, m_labelFS, m_labelCompression, m_labelOptions, m_labelSplit, m_btnSave, m_btnExit, NULL);
	newtFormAddComponents(m_formMain, m_radioCompNone, m_radioCompGzip, m_radioCompBzip2, m_checkCheckBeforeSaving, m_checkAskDesc, m_checkOverwrite, NULL);
	newtFormAddComponents(m_formMain, m_labelFinish, m_radioFinishWait, m_radioFinishHalt, m_radioFinishReboot, NULL);	
	newtFormAddComponents(m_formMain, m_radioSplitNone, m_radioSplitSize, m_radioSplitSize2, m_radioSplitAuto, m_editSplitSize, m_progressSaving, m_labelPercent, NULL);

	return 0;	
}

// =======================================================
int CSaveWindow::getValues(COptions *options)
{
	// get options
	options->bAskDesc = (newtCheckboxGetValue(m_checkAskDesc) == '*');
	options->bCheckBeforeSaving = (newtCheckboxGetValue(m_checkCheckBeforeSaving) == '*');
	options->bOverwrite = (newtCheckboxGetValue(m_checkOverwrite) == '*');

	// get compression level
	if (newtRadioGetCurrent(m_radioCompNone) == m_radioCompNone)
		options->dwCompression = COMPRESS_NONE;
	else if (newtRadioGetCurrent(m_radioCompNone) == m_radioCompGzip)
		options->dwCompression = COMPRESS_GZIP;
	else if (newtRadioGetCurrent(m_radioCompNone) == m_radioCompBzip2)
		options->dwCompression = COMPRESS_BZIP2;
		
	// get finish level
	if (newtRadioGetCurrent(m_radioFinishWait) == m_radioFinishWait)
		options->dwFinish = FINISH_WAIT;
	else if (newtRadioGetCurrent(m_radioFinishWait) == m_radioFinishHalt)
		options->dwFinish = FINISH_HALT;
	else if (newtRadioGetCurrent(m_radioFinishWait) == m_radioFinishReboot)
		options->dwFinish = FINISH_REBOOT;
		
	// get split mode
	if (newtRadioGetCurrent(m_radioSplitNone) == m_radioSplitNone)
		options->dwSplitMode = SPLIT_NONE;
	else if (newtRadioGetCurrent(m_radioSplitNone) == m_radioSplitSize)
	{	
		options->dwSplitMode = SPLIT_SIZE;
		options->dwSplitSize = atoll(newtEntryGetValue(m_editSplitSize)) * 1024;
	}
	else if (newtRadioGetCurrent(m_radioSplitNone) == m_radioSplitAuto)
	{
		options->dwSplitMode = SPLIT_AUTO;
		options->dwSplitSize = 0;
	}
	
	return 0;
}

// =======================================================
void CSaveWindow::showDetectedFS(char *szFs)
{
	char szTemp[1024];
	
	sprintf(szTemp, i18n("Detected file system:......%s"), szFs);
	newtLabelSetText(m_labelFS, szTemp);
}

// =======================================================
void CSaveWindow::showPartitionInfo(char *szDevice, QWORD qwSize)
{
	char szTemp[1024];
	char szTemp2[1024];
	
	sprintf(szTemp, i18n("Partition to save:.........%s (%s = %llu bytes)"), szDevice, formatSize(qwSize, szTemp2), qwSize);
	newtLabelSetText(m_labelPartition, szTemp);
}

// =======================================================
void CSaveWindow::showAvaibleSpaceForImage(QWORD qwFreeSpace)
{
	char szTemp[1024];
	char szTemp2[1024];
	
	sprintf (szTemp, i18n("Avaible space for image:...%s (%llu bytes)"), formatSize(qwFreeSpace, szTemp2), qwFreeSpace);
	newtLabelSetText(m_labelFreeSpace, szTemp);
}

// =======================================================
void CSaveWindow::showImageFileInfo(char *szImageFile, QWORD qwFreeSpace, QWORD qwImageSize)
{
	char szTemp[2048];
	char szTemp2[2048];
	
	if (qwImageSize)
		sprintf(szTemp, i18n("Current image file:........%s (%s)"), szImageFile, formatSize(qwImageSize, szTemp2));
	else
		sprintf(szTemp, i18n("Current image file:........%s"), szImageFile);
		
	newtLabelSetText(m_labelImageFile, szTemp);
	
	showAvaibleSpaceForImage(qwFreeSpace);
}

// =======================================================
int CSaveWindow::runForm()
{
	newtComponent widgetTemp;
	widgetTemp = newtRunForm(m_formMain);
	if (widgetTemp == m_btnSave)
		return 0;
	else
		return -1;
}

// =======================================================
void CSaveWindow::destroyForm()
{
	newtFormDestroy(m_formMain);
	newtPopWindow();
}

// =======================================================
void CSaveWindow::setProgressValue(QWORD qwSaved, QWORD qwTotal)
{
	QWORD qwPercent;
	char szTemp[1024];
	
	qwPercent = (100 * qwSaved) / qwTotal;
	sprintf(szTemp, "%d %%", (int)qwPercent);
	newtScaleSet(m_progressSaving, (int)qwPercent);
	newtLabelSetText(m_labelPercent, szTemp);
	newtRefresh();
}





// =======================================================
int CRestoreWindow::create(char *szDevice, char *szImageFile, COptions options)
{
	char szTemp[1024];
	
	sprintf(szTemp, "restore partition from image file");
	newtCenteredWindow(78, 20, szTemp);

	sprintf(szTemp, i18n("Partition to restore:.............%s"), szDevice);
	m_labelPartition = newtLabel(1, 0, szTemp);
	sprintf(szTemp, i18n("Image file to use.................%s"), szImageFile);
	m_labelImageFile = newtLabel(1, 1, szTemp);
	m_labelFS = newtLabel(1, 2, "");
	m_labelCompression = newtLabel(1, 3, "");
	m_labelOldDevice = newtLabel(1, 4, "");
	m_labelDate = newtLabel(1, 5, "");
	m_labelOriginalSize = newtLabel(1, 6, "");
	m_labelSplit = newtLabel(1, 7, "");
	
	m_labelOptions = newtLabel(1, 9, i18n("Options"));
	m_checkEraseWithNull = newtCheckbox(1, 10, i18n("Erase free blocks with zero values"), (!!options.bEraseWithNull ? '*' : ' '), " *", NULL);
	m_labelFinish = newtLabel(1, 12, i18n("If finished successfully:"));
	m_radioFinishWait = newtRadiobutton(1, 13, i18n("Wait"), options.dwFinish == FINISH_WAIT, NULL);
	m_radioFinishHalt = newtRadiobutton(1, 14, i18n("Halt"), options.dwFinish == FINISH_HALT, m_radioFinishWait);
	m_radioFinishReboot = newtRadiobutton(1, 15, i18n("Reboot"), options.dwFinish == FINISH_REBOOT, m_radioFinishHalt);
	m_progressRestoring = newtScale(1, 17, 70, 100);
	m_labelPercent = newtLabel(72, 17, "");
	m_btnRestore = newtCompactButton(20, 19, i18n("Restore"));
	m_btnExit = newtCompactButton(42, 19, i18n("Exit"));

	m_formMain = newtForm(NULL, NULL, 0);
	newtFormAddComponents(m_formMain, m_labelPartition, m_labelImageFile, m_labelFS, m_labelCompression, m_labelSplit, NULL);
	newtFormAddComponents(m_formMain, m_labelOldDevice, m_labelDate, m_labelOriginalSize, m_labelOptions, m_checkEraseWithNull, NULL);
	newtFormAddComponents(m_formMain, m_labelFinish, m_radioFinishWait, m_radioFinishHalt, m_radioFinishReboot, NULL);	
	newtFormAddComponents(m_formMain, m_progressRestoring, m_labelPercent, m_btnRestore, m_btnExit, NULL);

	return 0;	
}

// =======================================================
void CRestoreWindow::showFS(char *szFileSystem)
{
	char szTemp[1024];
	
	sprintf(szTemp, i18n("File system:......................%s"), szFileSystem);
	newtLabelSetText(m_labelFS, szTemp);
}

// =======================================================
int CRestoreWindow::getValues(COptions *options)
{
	// get options
	options->bEraseWithNull = (newtCheckboxGetValue(m_checkEraseWithNull) == '*');
	
	// get finish level
	if (newtRadioGetCurrent(m_radioFinishWait) == m_radioFinishWait)
		options->dwFinish = FINISH_WAIT;
	else if (newtRadioGetCurrent(m_radioFinishWait) == m_radioFinishHalt)
		options->dwFinish = FINISH_HALT;
	else if (newtRadioGetCurrent(m_radioFinishWait) == m_radioFinishReboot)
		options->dwFinish = FINISH_REBOOT;
		
	return 0;
}

// =======================================================
void CRestoreWindow::showPartitionInfo(char *szDevice, QWORD qwSize)
{
	char szTemp[1024];
	char szTemp2[1024];
	
	sprintf(szTemp, i18n("Partition to restore:.............%s (%s = %llu bytes)"), szDevice, formatSize(qwSize, szTemp2), qwSize);
	newtLabelSetText(m_labelPartition, szTemp);
}

// =======================================================
int CRestoreWindow::runForm()
{
	newtComponent widgetTemp;
	widgetTemp = newtRunForm(m_formMain);
	if (widgetTemp == m_btnRestore)
		return 0;
	else
		return -1;
}

// =======================================================
void CRestoreWindow::destroyForm()
{
	newtFormDestroy(m_formMain);
	newtPopWindow();
}

// =======================================================
void CRestoreWindow::showSplittedInfo(char *szState)
{
	newtLabelSetText(m_labelSplit, szState);
}	

// =======================================================
void CRestoreWindow::setProgressValue(QWORD qwRestoredBlocks, QWORD qwUsedBlocks, QWORD qwTotalBlocks, bool bEraseWithNull)
{
	QWORD qwPercent;
	char szTemp[1024];
	
	if (bEraseWithNull == false)
		qwPercent = (100 * qwRestoredBlocks) / qwUsedBlocks;
	else
		qwPercent = (100 * qwRestoredBlocks) / qwTotalBlocks;
		
	sprintf(szTemp, "%d %%", (int)qwPercent);

	newtScaleSet(m_progressRestoring, (int)qwPercent);
	newtLabelSetText(m_labelPercent, szTemp);
	newtRefresh();
}

// =======================================================
void CRestoreWindow::showPartitionInfos(DWORD dwCompression, char *szOriginalDevice, char *szFileSystem, tm dateCreate, QWORD qwPartSize)
{
	char szTemp[1024];
	char szTemp2[1024];
	
	sprintf(szTemp, i18n("File system:......................%s"), szFileSystem);
	newtLabelSetText(m_labelFS, szTemp);
	
	switch (dwCompression)
	{
		case COMPRESS_NONE:
			sprintf(szTemp, i18n("Compression level:................None\n"));
			break;
		case COMPRESS_GZIP:
			sprintf(szTemp, i18n("Compression level:................gzip"));
			break;
		case COMPRESS_BZIP2:
			sprintf(szTemp, i18n("Compression level:................bzip2"));
			break;
	}
	newtLabelSetText(m_labelCompression, szTemp);
	
	sprintf(szTemp, i18n("Partition was on device:..........%s\n"), szOriginalDevice);
	newtLabelSetText(m_labelOldDevice, szTemp);
	
	sprintf(szTemp, i18n("Image created on:.................%.4d/%.2d/%.2d\n"), dateCreate.tm_year+1900, dateCreate.tm_mon+1, dateCreate.tm_mday);
	newtLabelSetText(m_labelDate, szTemp);
	
	sprintf(szTemp, i18n("Size of the original partition:...%llu bytes = %s\n\n"), qwPartSize, formatSize(qwPartSize, szTemp2));
	newtLabelSetText(m_labelOriginalSize, szTemp);
	
}

// =======================================================
void CRestoreWindow::showImageFileInfo(char *szImageFile)
{
	char szTemp[2048];
	
	sprintf(szTemp, i18n("Current image file:...............%s"), szImageFile);	
	newtLabelSetText(m_labelImageFile, szTemp);
}











// ========================================================================================
void askDescription(char *szDescription, DWORD dwMaxLen)
{
	newtComponent editDesc;
	newtComponent labelText;
	newtComponent form;
	newtComponent btnOk;
	
	memset(szDescription, 0, dwMaxLen);

	newtCenteredWindow(65, 18, i18n("Partition description"));
	
	labelText = newtLabel(1, 1, i18n("You can enter a description of the saved partition:"));
	editDesc = newtEntry(1, 3, "", 60, NULL, NEWT_ENTRY_SCROLL);
	btnOk = newtButton(27, 13, i18n("Ok"));
	
	form = newtForm(NULL, NULL, 0);
	newtFormAddComponents(form, labelText, editDesc, btnOk, NULL);
	
	newtRunForm(form);
	newtPopWindow();	
	
	strncpy(szDescription, newtEntryGetValue(editDesc), dwMaxLen);
	newtFormDestroy(form);
}

// ========================================================================================
int askText(char *szMessage, char *szTitle, char *szDestText)
{
	newtComponent editText;
	newtComponent labelText;
	newtComponent form;
	newtComponent btnOk;
	newtComponent btnCancel;
	newtComponent temp;
	
	*szDestText = 0;

	while (!*szDestText)
	{
		newtCenteredWindow(65, 20, szTitle);
		
		labelText = newtTextbox(1, 1, 60, 7, 0/*NEWT_TEXTBOX_SCROLL*/);
		newtTextboxSetText(labelText, szMessage);
		
		editText = newtEntry(1, 9, "", 55, NULL, NEWT_ENTRY_SCROLL);
		btnOk = newtButton(15, 14, i18n("Ok"));
		btnCancel = newtButton(40, 14, i18n("Cancel"));
	
		form = newtForm(NULL, NULL, 0);
		newtFormAddComponents(form, labelText, editText, btnOk, btnCancel, NULL);
	
		temp = newtRunForm(form);
		newtPopWindow();	
		if (temp == btnCancel)
		{	newtFormDestroy(form);
			return -1;
		}
	
		strcpy(szDestText, newtEntryGetValue(editText));
	
		if (*szDestText == 0)
			newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Text cannot be empty"));

		newtFormDestroy(form);
	}
	
	return 0;
}

// ========================================================================================
void showAboutDialog()
{
	char szText[128];
	
	sprintf(szText, i18n("About Partition Image %s"), VERSION);

	newtWinMessage(szText, i18n("Ok"), i18n("Partition Image is distributed under the GNU GPL 2 (General Public License)\n\n"
	"web: http://partimage.sourceforge.net/\nAuthor: Franois Dupoux\nContact: fdupoux@free.fr"));
}

