/***************************************************************************
                          imagefile.cpp  -  description
                             -------------------
    begin                : Mon Aug 14 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "imagefile.h"
#include "common.h"
#include "gui_text.h"

#include <newt.h>

#define GZIP_FREESPACE_NEED	65536 // else, zlib can't close the image file, and data error
char *g_szVolumeMagicString = "Partimage-volume";

// =======================================================
DWORD CImage::write(void *buf, DWORD dwLength, COptions options, bool bUpdateCRC) // return 0 if error, or written bytes
{
	if (m_bIsOpened == false)
		return 0;
	
	/* bUpdateCRC is false when m_image->write call m_image->write itself: do to update the CRC twice. It is also useful
	   when we don't want the CRC to be updated: when we write the CRC itself to the image
	*/

	DWORD dwRes;
	int nRes;
	QWORD qwFreeSpace;
	DWORD i;
	static QWORD qwCount = 0LL;
	qwCount += (QWORD) dwLength;
	
	// update the CRC control
	if (bUpdateCRC)
	{
		BYTE *cBuffer = (BYTE *) buf;
		for (i=0; i < dwLength; i++)
			m_qwCRC += cBuffer[i];
	}
	
	getDiskFreeSpaceForFile(&qwFreeSpace, m_szImageFilename);

	// Change volume if need
	if ((options.dwSplitMode == SPLIT_SIZE) && (m_dwCurrentVolumeSize + dwLength >= options.dwSplitSize))
	{	
		m_dwVolumeNumber++;
		m_dwCurrentVolumeSize = 0L;	
				
		fprintf(m_fDebug, "begin of splitChangeImageFileWrite()\n"); fflush(m_fDebug);  
		nRes = splitChangeImageFileWrite(options, false); // change file
		fprintf(m_fDebug, "end of splitChangeImageFileWrite(), return = %d\n", nRes); fflush(m_fDebug);  
		if (nRes == -1)
			return 0;
		
		m_guiSave->showImageFileInfo(m_szImageFilename, qwFreeSpace, 0);
	}
	
	// check there is enough free space
	if (qwFreeSpace < (QWORD) (dwLength + GZIP_FREESPACE_NEED))
	{
		// If automatic change in split mode
		if (options.dwSplitMode != SPLIT_NONE)
		{	
			m_dwVolumeNumber++;
			m_dwCurrentVolumeSize = 0L;	
		
			fprintf(m_fDebug, "begin of splitChangeImageFileWrite()\n"); fflush(m_fDebug);  
			nRes = splitChangeImageFileWrite(options, true); // change file
			fprintf(m_fDebug, "end of splitChangeImageFileWrite(), return = %d\n", nRes); fflush(m_fDebug);  
			if (nRes == -1)
				return 0;
				
			m_guiSave->showImageFileInfo(m_szImageFilename, qwFreeSpace, 0);
			return write(buf, dwLength, options, false); // restart the write operation
		}
		else // if not in the auto split mode
		{
			newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Error while writing image file: disk full !"));
			return 0;
		}
	
	}
	
	errno = 0;
	
	// only (-va) in SPLIT AUTO
	
	// write core
	if (m_dwCompression == COMPRESS_NONE)
		dwRes = fwrite(buf, 1, dwLength, m_fImageFile);
	else if (m_dwCompression == COMPRESS_GZIP)
		dwRes = gzwrite(m_gzImageFile, buf, dwLength);
	else if (m_dwCompression == COMPRESS_BZIP2)
		dwRes = BZ2_bzwrite(m_bzImageFile, buf, dwLength);
	else
		return 0;
	
	// if success
	if (dwRes == dwLength)
	{	
		if (options.dwCompression == COMPRESS_NONE)
			m_dwCurrentVolumeSize += dwLength;
		else if	(options.dwCompression == COMPRESS_GZIP)
			m_dwCurrentVolumeSize = getFileSize(m_szImageFilename);
		else if (options.dwCompression == COMPRESS_BZIP2)
			m_dwCurrentVolumeSize = getFileSize(m_szImageFilename);
		else
			return 0;
			
		if (qwCount > (1024*1024))
		{	//QWORD qwSize;
			//qwSize = getFileSize(m_szImageFilename);
			m_guiSave->showImageFileInfo(m_szImageFilename, qwFreeSpace, /*qwSize*/ m_dwCurrentVolumeSize);
			qwCount = 0LL;
		}
		
		return dwRes;
	}

	// if error
	if ((dwRes == -1 && errno == ENOSPC) || (dwRes < dwLength && dwRes != -1))
	{
	}
	
	return dwRes;
}

// =======================================================
int CImage::writeMagic(char *szMagicString, COptions options)
{
	if (m_bIsOpened == false)
		return -1;
	
	fprintf(m_fDebug, "begin of CImage::writeMagic()\n"); fflush(m_fDebug);
	DWORD dwRes;
	
	dwRes = write(szMagicString, strlen(szMagicString), options, true);
	if (dwRes != strlen(szMagicString))
	{	fprintf(m_fDebug, "end of CImage::writeMagic(), return = -1\n"); fflush(m_fDebug);  
		return -1;
	}
		
	fprintf(m_fDebug, "end of CImage::writeMagic(), return = 0\n"); fflush(m_fDebug);  
	return 0;
}

// =======================================================
int CImage::readAndCheckMagic(char *szMagicString, COptions options)
{
	if (m_bIsOpened == false)
		return -1;
	
	fprintf(m_fDebug, "begin of CImage::readAndCheckMagic()\n"); fflush(m_fDebug);
	DWORD dwRes;
	char cBuffer[512];
	
	// read string
	dwRes = read(cBuffer, strlen(szMagicString), m_bIsSplitted, options, true);
	if (dwRes != strlen(szMagicString))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read magic string from image file"));
		fprintf(m_fDebug, "end of CImage::readAndCheckMagic(), return = -1 (a)\n"); fflush(m_fDebug);  
		return -1;
	}
	
	// compare string
	if (memcmp(cBuffer, szMagicString, strlen(szMagicString)) != 0)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Invalid magic string"));
		fprintf(m_fDebug, "end of CImage::readAndCheckMagic(), return = -1 (b)\n"); fflush(m_fDebug);  
		return -1;
	}

	fprintf(m_fDebug, "end of CImage::readAndCheckMagic(), return = 0\n"); fflush(m_fDebug);  
	return 0;
}

// =======================================================
DWORD CImage::read(void *buf, DWORD dwLength, bool bSplitted, COptions options, bool bUpdateCRC) // return 0 if error, or read bytes
{
	if (m_bIsOpened == false)
		return 0;
	
	DWORD dwRes;
	int nRes;
	DWORD i;
	
	/* bUpdateCRC is false when write() call write() itself: do to update the CRC twice. It is also useful
	   when we don't want the CRC to be updated: when we write the CRC itself to the image
	*/
	
	// ****************************************
	if (m_dwCompression == COMPRESS_NONE)
		dwRes = fread(buf, 1, dwLength, m_fImageFile);
	else if (m_dwCompression == COMPRESS_GZIP)
		dwRes = gzread(m_gzImageFile, buf, dwLength);
	else if (m_dwCompression == COMPRESS_BZIP2)
	{	dwRes = BZ2_bzread(m_bzImageFile, buf, dwLength);
	}
	else
	{	return 0;
	}
	
	// end of file
	if (dwRes < dwLength && dwRes >= 0)
	{
		if (bSplitted)
		{
			m_dwVolumeNumber++;
		
			fprintf(m_fDebug, "begin of CImage::splitChangeImageFileRead()\n"); fflush(m_fDebug);  
			nRes = splitChangeImageFileRead(options); // change file
			fprintf(m_fDebug, "end of CImage::splitChangeImageFileRead(), return = %d\n", nRes); fflush(m_fDebug);  
			if (nRes == -1)
			{	return 0;
			}
				
			return read(buf, dwLength, false, options, true); // restart the read
		}
		else
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Unexpected end of image file"));
			return 0;
		}	
	}
	
	// update the CRC control if the read operation is whole successful
	if (bUpdateCRC)
	{
		BYTE *cBuffer = (BYTE *) buf;
		for (i=0; i < dwLength; i++)
			m_qwCRC += cBuffer[i];
	}

	return dwRes;
}

// =======================================================
int CImage::splitChangeImageFileRead(COptions options)
{
	if (m_bIsOpened == false)
		return -1;
	
	int nRes;
	DWORD dwRes;
	
	// close the current image file
	close();
	
	// calculate new file path
	sprintf(m_szImageFilename, "%s/%s.%.3ld", m_szPath, m_szOriginalFilename, m_dwVolumeNumber);
	
	// open new file
	nRes = openReading(options);
	if (nRes == -1)
		return -1;
	
	//printf (" - new volume: %s \n", m_szImageFilename);
	// Check volume header
	CVolumeHeader headVolume;
	
	dwRes = read(&headVolume, sizeof(CVolumeHeader), true,  options, false);
	if (dwRes != sizeof(CVolumeHeader))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't read volume header of %s"), m_szImageFilename);
		return -1;
	}
	
	if (strcmp(headVolume.szMagicString, g_szVolumeMagicString) != 0)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("%s is not a Partition Image volume."), m_szImageFilename);
		return -1;
	}

	if (headVolume.dwVolumeNumber != m_dwVolumeNumber)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Invalid volume number (%ld instead of %ld)."), headVolume.dwVolumeNumber, m_dwVolumeNumber);
		return -1;
	}
	
	m_guiRestore->showImageFileInfo(m_szImageFilename);
	
	return 0;
}

// =======================================================
int CImage::splitChangeImageFileWrite(COptions options, bool bAskPath)
{
	if (m_bIsOpened == false)
		return -1;
	
	int nRes;
	char szMess[2048];
	
	// close the current image file
	close();
	
	// confirm new path
	if (bAskPath)
	{	
		sprintf(szMess, i18n("Cannot write next volume file (%s.%.3d) to directory\n%s.\nPlease, enter another directory path (without filename):"), m_szOriginalFilename, m_dwVolumeNumber, m_szPath);
		nRes = askText(szMess, i18n("New volume location"), m_szPath);
		if (nRes == -1)
			return -1;

		if (m_szPath[strlen(m_szPath)] == '/')
			m_szPath[strlen(m_szPath)] = 0;

	}
	
	sprintf(m_szImageFilename, "%s/%s.%.3ld", m_szPath, m_szOriginalFilename, m_dwVolumeNumber);
	//printf (i18n("\n==> new volume: %s \n"), m_szImageFilename);

	// open new file
	nRes = openWriting(options);
	if (nRes == -1)
		return -1;
		
	// Write volume header
	CVolumeHeader headVolume;
	memset(&headVolume, 0, sizeof(CVolumeHeader));
	
	strcpy(headVolume.szMagicString, g_szVolumeMagicString);
	headVolume.dwVolumeNumber = m_dwVolumeNumber;
	
	write(&headVolume, sizeof(CVolumeHeader), options, false);
	
	return 0;
}

// =======================================================
int CImage::close()
{
	fprintf(m_fDebug, "begin of CImage::close()\n"); fflush(m_fDebug);
	
	if (m_bIsOpened == false)
		return 0;
	
	int nRes = 0;
	
	if (m_dwCompression == COMPRESS_NONE) // No compression
		nRes = fclose(m_fImageFile);	
	else if (m_dwCompression == COMPRESS_GZIP) // Gzip compression
		nRes = gzclose(m_gzImageFile);
	else if (m_dwCompression == COMPRESS_BZIP2) // Bzip2 compression
		BZ2_bzclose(m_bzImageFile);
		
	m_bIsOpened = false;
		
	fprintf(m_fDebug, "end of CImage::close(), return = %lu\n", nRes); fflush(m_fDebug);
	return nRes;
}

// =======================================================
int CImage::openWriting(COptions options)
{
	if (m_bIsOpened == true)
		return -1;
	
	int nRes;
	
	fprintf(m_fDebug, "begin of CImage::openWriting()\n"); fflush(m_fDebug);  
	
	m_dwCompression = options.dwCompression;

	// asks user to overwrite the existing image file if need
	if ((access(m_szImageFilename, F_OK) != -1) && (options.bOverwrite != true)) // if the file already exists
	{	nRes = newtWinChoice(i18n("Overwrite image ?"), i18n("Yes"), i18n("No"), i18n("The %s image file already exists. Do you want to overwrite it ?"), m_szImageFilename);
		if (nRes == 2)
		{	fprintf(m_fDebug, "end of CImage::openWriting(), return=-1 (a)\n"); fflush(m_fDebug);  
			return -1;
		}
	}
	
	// check the user has image file write rights
	if ( (access(m_szImageFilename, F_OK) != -1) && (access(m_szImageFilename, W_OK) == -1)) // if existing and no rights
	{	newtWinMessage(i18n("Access denied"), i18n("Cancel"), i18n("Access denied: you don't have the write rights on %s"), m_szImageFilename);
		fprintf(m_fDebug, "end of CImage::openWriting(), return=-1 (b)\n"); fflush(m_fDebug);  
		return -1;
	}

	if (options.dwCompression == COMPRESS_NONE) // No compression
	{
		m_fImageFile = fopen(m_szImageFilename, "wb");
		if (m_fImageFile == NULL)
		{	fprintf(m_fDebug, "end of CImage::openWriting(), return=-1 (c)\n"); fflush(m_fDebug);  
			newtWinMessage(i18n("Error"), i18n("Cancel"), i18n("Can't open image %s for writing"), m_szImageFilename);
			return -1;
		}
	}
	else if (options.dwCompression == COMPRESS_GZIP) // Gzip compression
	{
		m_gzImageFile = gzopen(m_szImageFilename, "wb");
		if (m_gzImageFile == NULL)
		{	fprintf(m_fDebug, "end of CImage::openWriting(), return=-1 (d)\n"); fflush(m_fDebug);  
			newtWinMessage(i18n("Error"), i18n("Cancel"), i18n("Can't open image %s for writing"), m_szImageFilename);
			return -1;
		}	
	}
	else if (options.dwCompression == COMPRESS_BZIP2) // Bzip2 compression
	{
		m_bzImageFile = BZ2_bzopen(m_szImageFilename, "wb");
		if (m_bzImageFile == NULL)
		{	newtWinMessage(i18n("Error"), i18n("Cancel"), i18n("Can't open image %s for writing"), m_szImageFilename);
			return -1;
		}	
	}
	else
	{
		fprintf(m_fDebug, "end of CImage::openWriting(), return=-1 (e)\n"); fflush(m_fDebug);  
		newtWinMessage(i18n("Error"), i18n("Cancel"), i18n("Invalid compression level"), m_szImageFilename);
		return -1;
	}	

	m_bIsOpened = true;
	
	fprintf(m_fDebug, "end of CImage::openWriting(), return=0\n"); fflush(m_fDebug);
	return 0;
}


// =======================================================
int CImage::openReading(COptions options)
{
	if (m_bIsOpened == true)
		return -1;
	
	char szMess[2048];
	int nRes;
	
	m_dwCompression = options.dwCompression;
	
	// check the file exists
	errno = 0;
	while (access(m_szImageFilename, F_OK) == -1) // if the file doesn't exists
	{	sprintf (szMess, i18n("Can't read the following volume file:\n%s\nEnter another full path (directory and name)."), m_szImageFilename);
		nRes = askText (szMess, i18n("Next volume not found"), m_szImageFilename);
		if (nRes == -1)
			return -1;

		extractFilepathFromFullPath(m_szImageFilename, m_szPath);
		errno = 0;
	}
	
	if (options.dwCompression == COMPRESS_NONE) // No compression
	{
		m_fImageFile = fopen(m_szImageFilename, "rb");
		if (m_fImageFile == NULL)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't open image %s for reading"), m_szImageFilename);
			return -1;
		}
	}
	else if (options.dwCompression == COMPRESS_GZIP) // Gzip compression
	{
		m_gzImageFile = gzopen(m_szImageFilename, "rb");
		if (m_gzImageFile == NULL)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't open gzipped image %s for reading"), m_szImageFilename);
			return -1;
		}	
	}
	else if (options.dwCompression == COMPRESS_BZIP2) // Bzip2 compression
	{
		m_bzImageFile = BZ2_bzopen(m_szImageFilename, "rb");
		if (m_bzImageFile == NULL)
		{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't open bzipped image %s for reading"), m_szImageFilename);
			return -1;
		}	
	}
	else
	{
		newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Invalid compression level"));
		return -1;
	}
	
	m_bIsOpened = true;
	return 0;
}	

// =======================================================
void CImage::setGuiSave(CSaveWindow *gui)
{
	m_guiSave = gui;
}

// =======================================================
void CImage::setGuiRestore(CRestoreWindow *gui)
{
	m_guiRestore = gui;
}

// =======================================================
QWORD CImage::ftell() // used for debug
{
	if (m_bIsOpened == false)
		return 0;
	
	if (m_dwCompression == COMPRESS_NONE)
		return ::ftell(m_fImageFile);
	else if (m_dwCompression == COMPRESS_GZIP)
		return gztell(m_gzImageFile);
	else if (m_dwCompression == COMPRESS_BZIP2)
		return 0; //BZ2_bztell(m_bzImageFile);
	else
		return 0;
}
