/***************************************************************************
                          main.cpp  -  description
                             -------------------
    begin                : lun mai 22 18:04:54 CEST 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <unistd.h>
#include <newt.h>

#include "cext2part.h"
#include "cfatpart.h"
#include "creiserpart.h"
#include "cntfspart.h"
#include "chpfspart.h"

#include "partimage.h"
#include "common.h"
#include "gui_text.h"
#include "parttable.h"

extern char *optarg;
extern int optind;
extern int opterr;

static struct option const long_options[] =
{
  {"compress", required_argument, NULL, 'z'},
  {"overwrite", no_argument, NULL, 'o'},
  {"volume", required_argument, NULL, 'v'},
  {"erase", no_argument, NULL, 'e'},
  {"nocheck", no_argument, NULL, 'c'},
  {"forcefs", required_argument, NULL, 't'},
  {"allowmounted", no_argument, NULL, 'm'},
  {"nodesc", no_argument, NULL, 'd'},
  {"help", no_argument, NULL, 'h'},
  {"version", no_argument, NULL, 'i'},
  {"finish", required_argument, NULL, 'f'},
  {NULL, 0, NULL, 0}
};

static char optstring[]="z:ov:ect:mdhif:";
FILE *m_fDebug; // debug file

// =======================================================
int main(int argc, char *argv[])
{
	int nRes;
	int nOptCh;
	COptions options;
	int nChoice;
	char szFileSystem[128];
	char szDevice[4096];
	char szImageFile[4096];

	// initialize options with defaults values
	options.bOverwrite = OPTIONS_DEFAULT_OVERWRITE;
	options.dwSplitMode = OPTIONS_DEFAULT_SPLIT_MODE;
	options.dwSplitSize = OPTIONS_DEFAULT_SPLIT_SIZE * 1024;
	options.bEraseWithNull = OPTIONS_DEFAULT_ERASE_EMPTY;
	options.dwCompression = OPTIONS_DEFAULT_COMPRESS;
	options.bCheckBeforeSaving = OPTIONS_DEFAULT_CHECK;
	options.bFailIfMounted = OPTIONS_DEFAULT_FAIL_IF_MOUNTED;
	options.bAskDesc = OPTIONS_DEFAULT_ASK_DESC;
	options.dwFinish = OPTIONS_DEFAULT_FINISH;
	strcpy(szFileSystem, "detect"); // default file system of the saved partition
	
	// initialize language for i18n
	setlocale(LC_ALL, "");
	bindtextdomain (PACKAGE, "/usr/share/locale");
  textdomain (PACKAGE);
	
	// check main header size is 16384
	if (sizeof(CMainHeader) != 16384)
	{	printf ("Error: main header size != 16384 (%ld)\n", sizeof(CMainHeader));
		return EXIT_FAILURE;
	}
	
	// check ext2 header size is 16384
	if (sizeof(CExt2Header) != 16384)
	{	printf ("Error: ext2 header size != 16384 (%ld)\n", sizeof(CExt2Header));
		return EXIT_FAILURE;
	}
	
	// check fat header size is 16384
	if (sizeof(CFatHeader) != 16384)
	{	printf ("Error: fat header size != 16384 (%ld)\n", sizeof(CFatHeader));
		return EXIT_FAILURE;
	}
	
	// check reiser header size is 16384
	if (sizeof(CReiserHeader) != 16384)
	{	printf ("Error: reiser header size != 16384 (%ld)\n", sizeof(CReiserHeader));
		return EXIT_FAILURE;
	}
	
	// check ntfs header size is 16384
	if (sizeof(CNtfsHeader) != 16384)
	{	printf ("Error: ntfs header size != 16384 (%ld)\n", sizeof(CNtfsHeader));
		return EXIT_FAILURE;
	}
	
	// check hpfs header size is 16384
	if (sizeof(CHpfsHeader) != 16384)
	{	printf ("Error: hpfs header size != 16384 (%ld)\n", sizeof(CHpfsHeader));
		return EXIT_FAILURE;
	}
	
	// check ext2 footer size is 16384
	if (sizeof(CMainFooter) != 16384)
	{	printf ("Error: main footer size != 16384 (%ld)\n", sizeof(CMainFooter));
		return EXIT_FAILURE;
	}
	
	// check volume header size is 512
	if (sizeof(CVolumeHeader) != 512)
	{	printf ("Error: volume hedaer size != 512 (%ld)\n", sizeof(CVolumeHeader));
		return EXIT_FAILURE;
	}
	
	// DEVELOP:
	//printPartTable("/dev/hda");
	//return 0;
	
	// read command line
	while ((nOptCh = getopt_long (argc, argv, optstring, long_options, NULL)) != EOF)
	{
		switch(nOptCh)
		{
			case 'z': // compression level
				options.dwCompression = atol(optarg);
				if ((options.dwCompression < 0) || (options.dwCompression > 2))
				{	printf(i18n("Compression mode must be 0 (none), 1(gzip) or 2(bzip2)\n"));
					return EXIT_FAILURE;
				}
				break;
			
			case 'f': // what to do if finished successfully ?
				options.dwFinish = atol(optarg);
				if ((options.dwFinish < 0) || (options.dwFinish > 2))
				{	printf(i18n("Finish mode must be 0 (wait), 1(halt) or 2(reboot)\n"));
					return EXIT_FAILURE;
				}
				break;
			
			case 't': // file system to use to save the partition
				strcpy(szFileSystem, optarg);
				if (!isFileSystemSupported(optarg))
				{	printf(i18n("The %s file system is not supported.\n"), optarg);
					return EXIT_FAILURE;
				}
				break;
			
			case 'v': // split image into multiple files
				
				if (*optarg == 'a')
				{	options.dwSplitMode = SPLIT_AUTO;
				}
				else
				{	options.dwSplitMode = SPLIT_SIZE;
					options.dwSplitSize = atol(optarg) * 1024;
					printf (i18n("Volume size: %ld bytes (%ld KB)\n"), options.dwSplitSize, atol(optarg));
					if (atol(optarg) < 1024)
					{	printf(i18n("Split volumes size must be an integer greater than 1024 (size >= 1024 KB)\n"));
						return EXIT_FAILURE;
					}
				}
				break;
			
			case 'o': // overwrite existring image
				options.bOverwrite = true;
				break;
			
			case 'e': // erase empty blocks with zero bytes
				options.bEraseWithNull = true;
				break;
				
			case 'm': // don't fail if mounted
				options.bFailIfMounted = false;
				break;
				
			case 'c': // don't check partition before saving
				options.bCheckBeforeSaving = false;
				break;

			case 'd': // don't ask any description
				options.bAskDesc = false;
				break;

			case 'h': // help
				usage();
				return EXIT_SUCCESS;
				break;

			case 'i': // version
				printf(i18n("Partition Image version %s (distributed under the GNU GPL2).\n"), VERSION);
				return EXIT_SUCCESS;
				break;
		}
	}

	// debug file
	m_fDebug = fopen("partimage-debug.log", "w+");
	if (!m_fDebug)
	{	printf(i18n("Cannot open debug file.\n"));
		//return EXIT_FAILURE;
                m_fDebug = fopen("/dev/null", "w+");
                if (!m_fDebug)
                  m_fDebug = stderr;
	}
	else
	{
		fprintf(m_fDebug, "partimage-debug.log: Partition Image version %s\n", VERSION); fflush(m_fDebug);
	}
	
	// init interface
	newtInit();
	newtCls();

	// ----------- check the user is logged as root
	if (geteuid() != 0) // 0 is the UID of the root
	{	nRes = newtWinChoice(i18n("Warning"), i18n("Continue"), i18n("Cancel"), i18n("You are not logged as root. You may have \"access denied\" errors when working."));	
		if (nRes == 2)
		{	newtFinished();
			return EXIT_FAILURE;
		}
	}		

	// run operations if the command line is full
	if (argc - optind == 3)
	{
		strcpy(szDevice, argv[optind+1]);
		strcpy(szImageFile, argv[optind+2]);
		
		if (strcmp(argv[optind], "save")==0) // save
		{
			nRes = savePartition(szDevice, szImageFile, szFileSystem, &options);	
		}
		else if (strcmp(argv[optind], "restore")==0) // restore
		{
			nRes = restorePartition(szDevice, szImageFile, &options);		
		}
		else
		{	newtFinished();	
			newtWinMessage(i18n("Error"), i18n("Ok"), i18n("%s is not a valid action."), argv[optind]);
			usage();
			fclose(m_fDebug);
			return EXIT_FAILURE;
		}
	}
	else
	{
		strcpy(szDevice, "/dev/");
		*szImageFile = 0;		
		
		// draw main window
		nChoice = guiInitMainWindow(szDevice, szImageFile);
		switch(nChoice)
		{
			case OPERATION_SAVE:
				nRes = savePartition(szDevice, szImageFile, szFileSystem, &options);
				break;
			
			case OPERATION_RESTORE:
				nRes = restorePartition(szDevice, szImageFile, &options);
				break;	
			
			default: // exit
				sync();
				newtFinished();	
				fclose(m_fDebug);
				return EXIT_SUCCESS;
				break;
		}
	}
	
	// update the disk IO
	sync();
	
	// show result message
	if (nRes == -1) // FAILED
	{	fprintf(m_fDebug, "End of operation: FAILED\n"); fflush(m_fDebug); fclose(m_fDebug);
		newtWinMessage(i18n("End of operation"), i18n("Cancel"), i18n("Operation failed"));
		newtFinished();
	}
	else // SUCCESS
	{	
		fprintf(m_fDebug, "End of operation: SUCCESS\n"); fflush(m_fDebug); fclose(m_fDebug);
		
		if (options.dwFinish == FINISH_WAIT)
		{
			newtWinMessage(i18n("End of operation"), i18n("Ok"), i18n("Operation successfully finished"));
			newtFinished();
		}
		else if (options.dwFinish == FINISH_HALT)
		{
			newtFinished();	
			nRes = system("/sbin/shutdown -h now");
			if (nRes != 0)
			{	printf(i18n("Error: Cannot run halt to power off the computer"));
				return EXIT_SUCCESS;
			}
		}
		else if (options.dwFinish == FINISH_REBOOT)
		{
			newtFinished();
			nRes = system("/sbin/shutdown -r now");
			if (nRes != 0)
			{	printf(i18n("Error: Cannot run reboot to power off the computer"));
				return EXIT_SUCCESS;
			}
		}
		
	}
	
		
	return EXIT_SUCCESS;
}

// =======================================================
void usage()
{
	cout << "===============================================================================" << endl;
	cout << "Partition Image (http://partimage.sourceforge.net/) version " << VERSION << endl;
	cout << "---- distributed under the GPL 2 license (GNU General Public License) ----" << endl << endl;
	cout << "Supported file systems:....ext2fs, reiserfs, fat16, fat32" << endl << endl;
	cout << "usage: partimage [options] <action> <device> <image_file>" << endl << endl;
	cout << "ex: partimage -z1 -o -d save /dev/hda12 /mnt/backup/redhat-6.2.partimg.gz" << endl;
	cout << "ex: partimage restore /dev/hda13 /mnt/backup/suse-6.4.partimg" << endl;
	cout << "ex: partimage -z1 -om save /dev/hda9 /mnt/backup/win95-osr2.partimg.gz" << endl << endl;
	cout << "Arguments:" << endl;
	cout << "* <action>:" << endl;
	cout << "  - save: to save the partition datas in an image file" << endl;
	cout << "  - restore: to restore the partition from an image file" << endl;	
	cout << "* <device>: partition to save/restore (example: /dev/hda1)" << endl;
	cout << "* <image_file>: file where data will be read/written. can be very big" << endl;
	cout << endl;
	cout << "Options:" << endl;
	cout << "* -z,  --compress     (image file compression level):" << endl;
	cout << "  -z0, --compress=0   don't compress: very fast but very big image file" << endl;
	cout << "  -z1, --compress=1   compress using gzip: fast and small image file (default)" << endl;
	cout << "  -z2, --compress=2   (compress using bzip2: very slow and very small image file):" << endl;
	cout << "* -c,  --nocheck      don't check the partition before saving" << endl;
	cout << "* -o,  --overwrite    overwrite the existing image file without confirmation" << endl;
	cout << "* -d,  --nodesc       don't ask any description for the image file" << endl;
	cout << "* -v,  --volume       (split image into multiple volumes files)" << endl;
	cout << "  -va, --volume=a     change volume when disk is full, and ask new path" << endl;
	cout << "  -vX, --volume=X     create volumes with a size of X KB" << endl;
	cout << "* -e,  --erase        erase empty blocks on restore with zero bytes" << endl;
	cout << "* -tX, --forcefs=X    force file system to be in the X file format (X can be ext2fs, ntfs, fat, reiserfs)" << endl;
	cout << "* -m,  --allowmounted don't fail if the partition is mounted. Dangerous !" << endl;
	cout << "* -h,  --help         show help" << endl;
	cout << "* -i,  --version      show version" << endl;
	cout << "* -f,  --finish       (action to do if finished successfully):" << endl;
	cout << "  -f0, --finish=0     wait: don't make anything" << endl;
	cout << "  -f1, --finish=1     halt (power off) the computer" << endl;
	cout << "  -f2, --finish=2     reboot (restart the computer):" << endl;
	cout << "===============================================================================" << endl;
}
