/***************************************************************************
                          misc.cpp  -  description
                             -------------------
    begin                : Sun Jul 16 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <mntent.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <sys/vfs.h>
#include <newt.h>
#include <ctype.h>

#include "misc.h"
#include "common.h"

DWORD g_dwBit[] = {1,2,4,8,16,32,64,128}; // Do not go from (128 to 1), reiser would fail
//DWORD g_dwBit[] = {128,64,32,16,8,4,2,1}; // Do not go from (128 to 1), reiser would fail

// =======================================================
int isMounted(const char *szDevice, char *szMountPoint)
{
	FILE * f;
	struct mntent * mnt;
	bool bMounted = false;

	if ((f = setmntent (MOUNTED, "r")) == 0)
		return -1;
	
	while ((mnt = getmntent (f)) != 0)
	{	if (strcmp(szDevice, mnt->mnt_fsname) == 0)
		{	bMounted = true;
			strcpy(szMountPoint, mnt->mnt_dir);
		}
	}
	endmntent (f);

	return bMounted;
}

// =======================================================
QWORD getPartitionSize(const char *szDevice)
{
	QWORD qwSize;
	long lSize;
	int fdDevice;

	qwSize = 0;
	
	fdDevice = open(szDevice, O_RDONLY);
	if (fdDevice < 0)
		return 0LL;

	if (ioctl(fdDevice, BLKGETSIZE, &lSize) >= 0)
	{	qwSize = ((QWORD) 512) * ((QWORD) lSize);
	}

	close(fdDevice);
	return qwSize;
}

// ========================================================================================
// Create the text with a size in Bytes, KiloBytes, MegaBytes, GigaBytes, TeraBytes from a 64 bits number
// Parameters::.....* qwSize: 64 bits number of the size in bytes
// Return values:...* formatted text size
// ========================================================================================
char *formatSize(QWORD qwSize, char *szText)
{	
    double dSize;

    QWORD llKiloB = ((QWORD) 1024);
    QWORD llMegaB = ((QWORD) 1024) * ((QWORD) 1024);
    QWORD llGigaB = ((QWORD) 1024) * ((QWORD) 1024) * ((QWORD) 1024);
    QWORD llTeraB = ((QWORD) 1024) * ((QWORD) 1024) * ((QWORD) 1024) * ((QWORD) 1024);
	
    if (qwSize < llKiloB) // In Bytes
    {	sprintf(szText, i18n("%ld bytes"), qwSize);
    }
    else if (qwSize < llMegaB) // In KiloBytes
    {	dSize = ((double) qwSize) / ((double) 1024.0);
    	sprintf(szText, i18n("%.2f KB"), (float) dSize);
    }
    else if (qwSize < llGigaB) // In MegaBytes
    {	dSize = ((double) qwSize) / ((double) (1024.0 * 1024.0));
    	sprintf(szText, i18n("%.2f MB"), (float) dSize);
    }
    else if (qwSize < llTeraB)// In GigaBytes
    {	dSize = ((double) qwSize) / ((double) (1024.0 * 1024.0 * 1024.0));
    	sprintf(szText, i18n("%.2f GB"), (float) dSize);
    }
    else // In TeraBytes
    {	dSize = ((double) qwSize) / ((double) (1024.0 * 1024.0 * 1024.0 * 1024.0));
    	sprintf(szText, i18n("%.2f TB"), (float) dSize);
    }

    return szText;
}

// =======================================================
CBlocksUseBitmap::CBlocksUseBitmap()
{
    m_cBitmap = NULL;
}

// =======================================================
CBlocksUseBitmap::~CBlocksUseBitmap()
{
    freeBitmap();
}

// =======================================================
int CBlocksUseBitmap::setBit(DWORD dwBit, bool bState)
{
	if (m_cBitmap == NULL)
		return -1;
	
	DWORD dwLocalBit;
	DWORD dwByte;
    
	dwLocalBit = dwBit % 8;
	dwByte = (dwBit - dwLocalBit) / 8;
	
	//printf ("dwBit=%ld ==> (byte,bit)=(%ld,%ld)\n", dwByte, dwLocalBit);

	// if we must set the bit
	if (bState == true)
	{
		m_cBitmap[dwByte] |= g_dwBit[dwLocalBit];	
	}
	else // if we must unset the bit
	{
		m_cBitmap[dwByte] &= ~g_dwBit[dwLocalBit];
	}

	return 0;
}

// =======================================================
int CBlocksUseBitmap::isBitSet(DWORD dwBit)
{
	if (m_cBitmap == NULL)
		return -1;

	DWORD dwLocalBit;
	DWORD dwByte;
    
	dwLocalBit = dwBit % 8;
	dwByte = (dwBit - dwLocalBit) / 8;
	
	return (m_cBitmap[dwByte] & g_dwBit[dwLocalBit]);
}

// =======================================================
int CBlocksUseBitmap::init(DWORD dwBlocksCount)
{
	// if memory already allocated
	if (m_cBitmap)
		return -1;
    
	m_dwBitmapSize = ((int)(dwBlocksCount / 8)) + 1;

	//printf(i18n("Creating the bitmap in memory (%s)"), formatSize(m_dwBitmapSize, szTemp));
	
	m_cBitmap = new BYTE[m_dwBitmapSize+1];
	if (m_cBitmap == NULL)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Out of memory"));
		return -1;
	}
	
	memset(m_cBitmap, 0, m_dwBitmapSize);
	return 0;
}

// =======================================================
int CBlocksUseBitmap::freeBitmap()
{
	if (m_cBitmap) // if memory was allocated
	{
		delete m_cBitmap;
		m_cBitmap = NULL;
	}
	return 0;
}

// =======================================================
int CBlocksUseBitmap::fill(BYTE cValue)
{
	DWORD i;
	DWORD dwMemSize;
    
	if (!m_cBitmap) // if memory was not allocated
		return -1;
	
	dwMemSize = m_dwBitmapSize / 8;
	
	for (i=0; i < dwMemSize; i++)
		m_cBitmap[i] = cValue;
    
	return 0;
}

// =======================================================
BYTE* CBlocksUseBitmap::data()
{
	return m_cBitmap;
}


// =======================================================
DWORD CBlocksUseBitmap::dataSize()
{
	return m_dwBitmapSize;
}

// =======================================================
DWORD CBlocksUseBitmap::crc(QWORD qwClustersCount)
{
	DWORD dwCRC = 0;
	DWORD i;
	DWORD qwLimit;
	
	if (!m_cBitmap)
		return 0;
	
	if (qwClustersCount == 0)
		qwLimit = m_dwBitmapSize;
	else
		qwLimit = qwClustersCount;
		
	for (i=0; i < qwLimit; i++)
		dwCRC += m_cBitmap[i];
	
	return dwCRC;
}

// Taken from reiserfs_fs.h
#define REISERFS_SUPER_MAGIC_STRING "ReIsErFs"
#define REISERFS_DISK_OFFSET_IN_BYTES (64 * 1024)

// =======================================================
int detectFileSystem(const char *szDevice, char *szFileSys)
{
	BYTE cBuffer[8192];
	FILE *fDeviceFile;
	UINT nTemp;
	int nRes;

	// 0. Open device
	fDeviceFile = fopen(szDevice, "rb");
	if (fDeviceFile == NULL)
	{	//newtWinMessage(i18n("Error"), i18n("Ok"), i18n("Can't open partition %s. (%s)"), szDevice, strerror(errno));
		return -1;
	}
	
	// 1. Read boot sector
	nRes = fseek(fDeviceFile, 0L, SEEK_SET);
	if (nRes == -1)
		return -1;
	
	nRes = fread(cBuffer, sizeof(cBuffer), 1, fDeviceFile);
	if (nRes != 1)
		return -1;
		
	// A. **** check for EXT2FS file system
	memcpy(&nTemp, cBuffer+0x0438, sizeof(UINT));
	if (nTemp == EXT2_SUPER_MAGIC)
	{	fclose(fDeviceFile);
		strcpy(szFileSys, "ext2fs");
		return FS_EXT2;
	}

	// B. **** check for NTFS file system
	if (memcmp(cBuffer+3, "NTFS", 4) == 0)	
	{	fclose(fDeviceFile);
		strcpy(szFileSys, "ntfs");
		return FS_NTFS;
	}
	
	// B. **** check for HPFS file system
	if (memcmp(cBuffer+54, "HPFS", 4) == 0)	
	{	fclose(fDeviceFile);
		strcpy(szFileSys, "hpfs");
		return FS_HPFS;
	}
	
	// C. **** check for FAT file system
	if (cBuffer[510] == 0x55 && cBuffer[511] == 0xAA)
	// NTFS can be detected here, then NTFS must be before here
	{	fclose(fDeviceFile);
		strcpy(szFileSys, "fat");
		return FS_FAT;
	}
	
	// D. **** check for REISER file system
	nRes = fseek(fDeviceFile, REISERFS_DISK_OFFSET_IN_BYTES, SEEK_SET);
	if (nRes == -1)
		return -1;
	
	nRes = fread(cBuffer, sizeof(cBuffer), 1, fDeviceFile);
	if (nRes != 1)
		return -1;
	
	if (memcmp(cBuffer+52, REISERFS_SUPER_MAGIC_STRING, strlen(REISERFS_SUPER_MAGIC_STRING)) == 0)	
	{	fclose(fDeviceFile);
		strcpy(szFileSys, "reiserfs");
		return FS_REISER;
	}
	
	// end on error
	fclose(fDeviceFile);
	return -1;
}

// =======================================================
int checkInodeForDevice(char *szDevice)
{
	int nRes;
	struct stat fStat;

	// check the inode exists (for example, hda17 can not de present in /dev)
	nRes = access(szDevice, F_OK);
	if (nRes == -1)
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("%s inode doesn't exists. You can create it with the following command:\n"
		"\"mknod %s b <major> <minor>\"\n"
		"The minor number is the number of the partition, and you can get the major number by typing "
		"the command \"file /dev/<drive>\", where <drive> is a partition inode of the same hard drive"),
		szDevice, szDevice);
		return -1;
	}	
	
	// check the device is a block one
	nRes = stat(szDevice, &fStat);
	if (nRes == -1 || !S_ISBLK(fStat.st_mode))
	{	newtWinMessage(i18n("Error"), i18n("Ok"), i18n("%s is not a block device."), szDevice);
		return -1;
	}	
	
	return 0;
}

// =======================================================
int extractFilenameFromFullPath(char *szFullpath, char *szFilename)
{
	// if no '/' present
	if (strchr(szFullpath, '/') == 0)
	{	strcpy(szFilename, szFullpath);
		return 0;	
	}

	// if '/' is present
	strcpy(szFilename, strrchr(szFullpath, '/')+1);
	return 0;
}

// =======================================================
int extractFilepathFromFullPath(char *szFullpath, char *szFilepath)
{
	int i;
	int nLen;
	
	// if no '/' present
	if (strchr(szFullpath, '/') == 0)
	{	strcpy(szFilepath, ".");
		return 0;	
	}

	// if '/' is present
	nLen = strrchr(szFullpath, '/') - szFullpath;
	for (i=0; i < nLen; i++)
		szFilepath[i] = szFullpath[i];
	szFilepath[i] = 0;
	
	return 0;
}

// =======================================================
QWORD getFileSize(char *szFilename)
{
	struct stat st;
	stat(szFilename, &st);	
	return (QWORD) st.st_size;
}

// ========================================================================================
// retrun the free space on disk where the file szFilename is written
// Parameters::.....* szFilename: path of the file
// .................* qwAvailDiskSpace: pointer to the 64 bits number where the free disk space must be written
// Return values:...*  0 if success
//                  * -1 if error
// ========================================================================================
int getDiskFreeSpaceForFile(QWORD *qwAvailDiskSpace, const char *szFilename)
{
	int nRes;
	struct statfs fsInfo;

	*qwAvailDiskSpace = 0LL;

	errno = 0;
	nRes = statfs(szFilename, &fsInfo);
	if (nRes == -1) // error
	{
		if (errno == ENOENT) // file doesn't exists
		{	// create an empty file
			FILE *fFile;
			fFile = fopen(szFilename, "w");
			if (fFile == NULL)
				return -1;
			fclose(fFile);
			nRes = getDiskFreeSpaceForFile(qwAvailDiskSpace, szFilename);
			unlink(szFilename);
			return nRes;
		}
		else // another error
		{	return -1;
		}
	
	}	
	
  *qwAvailDiskSpace = ((QWORD) fsInfo.f_bavail) * ((QWORD)fsInfo.f_bsize);

	return 0;
}

// ========================================================================================
int getCompressionLevelForImage(char *szFilename)
{
	gzFile gzImageFile;
	FILE *fImageFile;
	BZFILE *bzImageFile;
	DWORD dwRes;
	CMainHeader headMain;
	
		// ------ 0. Check for a bzip2 compression
checkBzip2:
	bzImageFile = BZ2_bzopen(szFilename, "rb");
	if (bzImageFile == NULL)
		goto checkNone;
	dwRes = BZ2_bzread(bzImageFile, &headMain, sizeof(CMainHeader));
	BZ2_bzclose(bzImageFile);
	if (dwRes != sizeof(CMainHeader))
		goto checkNone;
	if (strcmp(headMain.szAppName, "PartImage") == 0)
		return COMPRESS_BZIP2;
	
	// ------ 2. Check for no compression
checkNone:
	fImageFile = fopen(szFilename, "rb");
	if (fImageFile == NULL)
		goto checkGzip;
	dwRes = fread(&headMain, 1, sizeof(CMainHeader), fImageFile);
	fclose(fImageFile);	
	if (dwRes != sizeof(CMainHeader))
		goto checkGzip;
	if (strcmp(headMain.szAppName, "PartImage") == 0)
		return COMPRESS_NONE;

checkGzip:
	// ------ 3. Check for a gzip compression
	gzImageFile = gzopen(szFilename, "rb");
	if (gzImageFile == NULL)
		return -1; // error
	dwRes = gzread(gzImageFile, &headMain, sizeof(CMainHeader));
	gzclose(gzImageFile);
	if (dwRes != sizeof(CMainHeader))
		return -1; // error
	if (strcmp(headMain.szAppName, "PartImage") == 0)
		return COMPRESS_GZIP;
	
	return -1; // error
}

// =======================================================
int bigFseek(FILE *fStream, QWORD qwOffset) // SEEK_SET forced
{
	const long lGigaByte = (QWORD) (1024L * 1024L * 1024L);
	int nRes;

	// go to the beginning
	nRes = fseek(fStream, 0L, SEEK_SET);
	
	while (qwOffset > 0)
	{
		fprintf(m_fDebug, "big_fseek: qwOffset=%llu\n", qwOffset); fflush(m_fDebug);
		
		if (qwOffset >= (QWORD)lGigaByte)
		{
			nRes = fseek(fStream, (long)lGigaByte, SEEK_CUR);
			fprintf(m_fDebug, "big_fseek: big seek\n"); fflush(m_fDebug);
			if (nRes == -1)
				return -1;
			qwOffset -= ((QWORD)lGigaByte);	
		}
		else
		{
			fprintf(m_fDebug, "big_fseek: small seek: %llu\n", qwOffset); fflush(m_fDebug);
			nRes = fseek(fStream, (long)qwOffset, SEEK_CUR);
			return nRes;
		}
	}
	
	return 0;
}

// =======================================================
char *decodePartitionEntry(char *cPtr, int *nMajor, int *nMinor, int *nBlocks, int *nPartNum, char *szDevice)
{
	char cTemp[128];
	char *cCur;
	char *cDevPtr1;
	char *cDevPtr2;
		
	// Major number
	while (!isalnum(*cPtr))
		cPtr++;
	memset(cTemp, 0, 127);
	cCur = cTemp;
	while(isalnum(*cPtr))
	{	*cCur = *cPtr;
		cCur++;
		cPtr++;
	}
	*nMajor = atoi(cTemp);

	// Minor number
	while (!isalnum(*cPtr))
		cPtr++;
	memset(cTemp, 0, 127);
	cCur = cTemp;
	while(isalnum(*cPtr))
	{	*cCur = *cPtr;
		cCur++;
		cPtr++;
	}
	*nMinor = atoi(cTemp);

	// Block number
	while (!isalnum(*cPtr))
		cPtr++;
	memset(cTemp, 0, 127);
	cCur = cTemp;
	while(isalnum(*cPtr))
	{	*cCur = *cPtr;
		cCur++;
		cPtr++;
	}
	*nBlocks = atoi(cTemp);

	// Device
	while (!isalnum(*cPtr))
		cPtr++;
	memset(szDevice, 0, 127);
	cCur = szDevice;
	while(isalnum(*cPtr))
	{	*cCur = *cPtr;
		cCur++;
		cPtr++;
	}
	
	// go to the next line
	while((*cPtr != '\n') && (*cPtr != 0))
		cPtr++;
		
	// extract number from device
	cDevPtr1 = szDevice;
	cDevPtr2 = cTemp;
	memset(cTemp, 0, 127);
	strcpy(cTemp, "0");
	while (isalpha(*cDevPtr1)) // skip letters
	{	cDevPtr1++;
	}
	while (isdigit(*cDevPtr1))
	{	*cDevPtr2 = *cDevPtr1;
		cDevPtr1++;
		cDevPtr2++;
	}
	*nPartNum = atoi(cTemp);
	
	return cPtr;
}
