/***************************************************************************
                        partimage.h  -  description
                             -------------------
    begin                : Mon May 22 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef PARTIMAGE_H
#define PARTIMAGE_H

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <linux/ext2_fs.h>
#include <libintl.h> // transaltion
#include <zlib.h> // gzip compression
#include <bzlib.h> // bzip2 compression
#include <sys/stat.h>
#include <sys/param.h>
#include <errno.h>
#include <sys/statvfs.h>

// debug
extern FILE *m_fDebug;
extern char m_szDebug[];

#define i18n(X)	gettext(X)

typedef unsigned short int UINT;
typedef unsigned long int DWORD;
typedef unsigned long long int QWORD;
typedef unsigned char BYTE;

#define MAX_DEVICENAMELEN 512

#define OPERATION_SAVE			1
#define OPERATION_RESTORE		2
#define OPERATION_EXIT			3

#define FS_UNKNOWN   0
#define FS_EXT2      1
#define FS_FAT       2 // can be FAT12, FAT16, FAT32
#define FS_FAT12     3
#define FS_FAT16     4
#define FS_FAT32     5
#define FS_NTFS      6
#define FS_REISER    7
#define FS_HPFS      8

#define MAX_DESCRIPTION		    4096

// COMPRESSION
#define COMPRESS_NONE    0
#define COMPRESS_GZIP    1
#define COMPRESS_BZIP2   2	

// SPLIT MODE
#define SPLIT_NONE       0
#define SPLIT_AUTO	     1
#define SPLIT_SIZE       2

// FINISH MODE
#define FINISH_WAIT			 0
#define FINISH_HALT      1
#define FINISH_REBOOT    2

// FLAGS
#define FLAG_SPLITTED    1

#define NB_BOOT_COPIES_IN_IMAGE        2

// Magic string used to check data
#define MAGIC_BEGIN_BOOT			         "MAGIC-BEGIN-BOOT"
#define MAGIC_BEGIN_DATABLOCKS         "MAGIC-BEGIN-DATABLOCKS"
#define MAGIC_BEGIN_BITMAP             "MAGIC-BEGIN-BITMAP"
#define MAGIC_BEGIN_RESERVED           "MAGIC-BEGIN-RESERVED"
#define MAGIC_BEGIN_FAT		             "MAGIC-BEGIN-FAT"
#define MAGIC_BEGIN_ROOTDIR            "MAGIC-BEGIN-ROOTDIR"

// ======================== OPTIONS ================================
#define OPTIONS_DEFAULT_COMPRESS              COMPRESS_GZIP
#define OPTIONS_DEFAULT_ERASE_EMPTY           false
#define OPTIONS_DEFAULT_SPLIT_MODE            SPLIT_NONE
#define OPTIONS_DEFAULT_SPLIT_SIZE            1440
#define OPTIONS_DEFAULT_OVERWRITE             false
#define OPTIONS_DEFAULT_CHECK                 true
#define OPTIONS_DEFAULT_FAIL_IF_MOUNTED       true
#define OPTIONS_DEFAULT_ASK_DESC	      			true
#define OPTIONS_DEFAULT_FINISH	      				FINISH_WAIT

struct COptions
{
	bool bOverwrite;
	DWORD dwSplitMode;
	DWORD dwSplitSize; // 0LL = auto, !0 = bytes size
	bool bEraseWithNull;
	DWORD dwCompression;
	bool bCheckBeforeSaving;
	bool bFailIfMounted; // Fail if file system is mounted
	bool bAskDesc;
	DWORD dwFinish;
};

// ======================== FUNCTIONS ================================
void usage();


#endif // PARTIMAGE_H
